#include "VimanTags.h"
#include <iostream>
#include <fstream>
#include <string>

using namespace std;

bool VimanTags::_showText = true;
std::string VimanTags::_viman_path = "undef";

unsigned int VimanTags::readTag(Robot *rob){

    return readTag(rob,rob->getObjectRob());

}

unsigned int VimanTags::readTag(Robot* rob,ObjectRob *obj_rob){
    string cfg_path;
    string viman_path;
    string name = rob->getName();
    unsigned int nb=0;
    bool ok=true;

    viman_path = getVimanPath();
    ok = viman_path.size() != 0;
    if(ok){
        cfg_path = getCfgPath(viman_path,rob->getName());
        ok = cfg_path.size() != 0;
    }
    if(ok){
        vector<Eigen::Affine3d> tfs= getTransforms(cfg_path);
        if(tfs.size()){
            obj_rob->clearTags();
        }
        for(unsigned int i=0;i<tfs.size();i++){
            obj_rob->addTag(tfs[i]);
        }
        nb=tfs.size();

    }

    return nb;

}


string const & VimanTags::getVimanPath(){
    string m3dhome;

    if (_viman_path == "undef"){
        m3dhome = getenv("HOME_MOVE3D");
        if (m3dhome.size()){
            m3dhome+="/viman_tags/";
        }else if(_showText){
            cout << "VimanTags WARNING: viman directory not found, expected at $HOME_MOVE3D/viman_tags/"<<endl;
        }
        _viman_path = m3dhome;
    }else{
        m3dhome = _viman_path;
    }
    return _viman_path;
}

string VimanTags::getCfgPath(string const & viman_path, string const & robot_name){
    string cfg_list_path=viman_path;
    string cfg_path; //what we are looking for
    cfg_list_path+="/objectList_cfg";
    ifstream cfg;
    cfg.open(cfg_list_path.c_str());
    string search("object ");
    search+=robot_name;

    while(cfg && !cfg_path.size()){
        string line;
        getline(cfg,line);
        size_t pos=line.find(search);
        if( pos != string::npos){
            //the line of this object
            pos = line.find_first_of(' ',pos+search.size());
            if (pos != string::npos){
                cfg_path=viman_path + line.substr(pos+1); // append the path found in the file to the viman path
            }
        }
    }

    return cfg_path;
}

vector<Eigen::Affine3d> VimanTags::getTransforms(const string &cfg_path){
    vector<Eigen::Affine3d> tfs;
    ifstream cfg;
    cfg.open(cfg_path.c_str());
    typedef enum{NONE=0,NB_TAG,ID_TAG,SIZE_TAG,DUMMY,LINE1,LINE2,LINE3} STATES;
    int current_state=(int)NONE;
    double scale;
    double tf[12];
    while(cfg){
        string line;
        getline (cfg,line);
        size_t pos = line.find_first_not_of(' ');
        if(pos!=string::npos){
            line=line.substr(pos);
            istringstream iss(line);
            if (line.at(0) != '#'){
                //read
                current_state+=1;
                int i=-1;
                switch((STATES)current_state){
                case SIZE_TAG:
                    iss>>scale;
                    scale = scale/1000; //from mm to meters
                case LINE1:
                    i=0;
                    break;
                case LINE2:
                    i=1;
                    break;
                case LINE3:
                    i=2;
                    current_state=NB_TAG;
                    break;
                default:
                    break;
                }
                if(i>=0){
                    for(unsigned int j=0;j<4;j++){
                        iss >> tf[i*4+j];
                    }
                }
                if(i==2){
                    //add the  transform
                    Eigen::Affine3d mat;
                    // if(_showText){
                    //     cout<<"parsed:"<<endl;
                    // }
                    for(unsigned int r=0;r<3;r++){
                        for(unsigned int c=0;c<4;c++){
                            mat(r,c) = tf[r*4+c];
                            if(c==3){
                                mat(r,c)=mat(r,c)/1000; // from mm to meters
                            }
                            if(_showText)
                                cout << tf[r*4+c]<<" ";
                        }
                        // if(_showText)
                        //     cout<<endl;
                    }
                    // if(_showText){
                    //     cout<< "mat:"<<endl;
                    //     cout << mat.matrix()<<endl;
                    // }
                    Eigen::Affine3d t= mat ;//* Eigen::Scaling(scale) ;
                    t.scale(scale);
                    // if(_showText){
                    //     cout << "add transform "<<endl<<t.matrix()<<endl;
                    // }
                    tfs.push_back(t);
                }
            }

        }
    }
    return tfs;
}

std::vector<Eigen::Vector3d> VimanTags::getRectangleFromTransform(Eigen::Affine3d const &t, Eigen::Vector3d &norm){
    vector<Eigen::Vector3d> corners;
    corners.reserve(4);
    Eigen::Vector3d x(1,0,0),y(0,1,0),z(0,0,1),o(0,0,0);
    x=t.linear()*x;y=t.linear()*y;z=t.linear()*z;o=t*o;

    norm = t.linear() * z;

    corners.push_back(o-x/2-y/2);
    corners.push_back(o+x/2-y/2);
    corners.push_back(o+x/2+y/2);
    corners.push_back(o-x/2+y/2);

    return corners;
}

Eigen::Vector3d VimanTags::getTagNormalScaled(Eigen::Affine3d const &t){
    Eigen::Vector3d z(0,0,1);
    return t.linear() *z;
}

Eigen::Vector3d VimanTags::getTagNormal(Eigen::Affine3d const &t){
    Eigen::Vector3d n=getTagNormalScaled(t);
    n.normalize();
    return n;
}


