#include "utils_quadspline.h"
#include "Localpath-pkg.h"

void print_data(quad_data *data){
    printf("---- DATA -------------------\n");
    printf("| v0=%f\n",data->v0);
    printf("| a0=%f\n",data->a0);
    printf("| amax=%f\n",data->amax);
    printf("| jmax=%f\n",data->jmax);
    printf("| smax=%f\n",data->smax);
    printf("| durations=[%f %f %f %f %f]\n",data->durations[A1]
                                           ,data->durations[A2]
                                           ,data->durations[B]
                                           ,data->durations[C1]
                                           ,data->durations[C2]);
    printf("| signs=[%f %f]\n",data->signs[A],data->signs[C]);
    printf("-----------------------------\n");
}

int get_case(double aB,quad_data *data){
    double a0=data->a0;
    double jmax=data->jmax;
    double smax=data->smax;
    int sA=(int)sign(aB-a0);
    int sC=-(int)sign(aB);
    int cond_A=(fabs(a0-aB)>SQR(jmax)/smax);
    int cond_C=(fabs(aB)>SQR(jmax)/smax);
    return 1000*(sA+1)+100*(sC+1)+10*cond_A+cond_C;
}

void durations_and_signs(double aB,quad_data *data){
    double a0=data->a0;
    double jmax=data->jmax;
    double smax=data->smax;
    data->signs[A]=sign(aB-a0);
    data->signs[C]=-sign(aB);
    if(fabs(aB-a0)>SQR(jmax)/smax){
        data->durations[A1]=jmax/smax;
        data->durations[A2]=fabs(aB-a0)/jmax-jmax/smax;
    }
    else{
        data->durations[A1]=sqrt(fabs(a0-aB)/smax);
        data->durations[A2]=0;
    }
    if(fabs(aB)>SQR(jmax)/smax){
        data->durations[C1]=jmax/smax;
        data->durations[C2]=fabs(aB)/jmax-jmax/smax;
    }
    else{
        data->durations[C1]=sqrt(fabs(aB)/smax);
        data->durations[C2]=0;
    }
}

double v_c(double aB,double tB,quad_data *data){
    double v0=data->v0;
    double a0=data->a0;
    double smax=data->smax;
    data->durations[B]=tB;
    durations_and_signs(aB,data);
    double tA1=data->durations[A1];
    double tA2=data->durations[A2];
    double tC1=data->durations[C1];
    double tC2=data->durations[C2];
    double sA=data->signs[A];
    double sC=data->signs[C];
    double tA1_2=tA1*tA1;
    double tA1_3=tA1_2*tA1;
    double tA2_2=tA2*tA2;
    double tC1_2=tC1*tC1;
    double tC1_3=tC1_2*tC1;
    double tC2_2=tC2*tC2;
    return sA*smax*tA1_3+(3*sA*smax*tA1_2*tA2)/2+2*sA*smax*tA1_2*tC1+sA*smax*tA1_2*tC2+sA*smax*tB*tA1_2+(sA*smax*tA1*tA2_2)/2+2*sA*smax*tA1*tA2*tC1+sA*smax*tA1*tA2*tC2+sA*smax*tB*tA1*tA2+2*a0*tA1+a0*tA2+sC*smax*tC1_3+(3*sC*smax*tC1_2*tC2)/2+(sC*smax*tC1*tC2_2)/2+2*a0*tC1+a0*tC2+v0+a0*tB;
}

quad_data *intervals(double v0,double a0,double amax,double jmax,double smax){
    quad_data *data=MY_ALLOC(quad_data,1);
    data->v0=v0;
    data->a0=a0;
    data->amax=amax;
    data->jmax=jmax;
    data->smax=smax;
    if(!data){
        printf("error intervals: cannot allocate data structure\n");
        return NULL;
    }
    int i,k,l,n=1;
    double a1,a2;
    double aj=jmax*jmax/smax;
    double Ac[7]={amax,0,aj,-aj,a0,a0+aj,a0-aj};
    data->int_a[0]=-amax;
    for(int k=0;k<7;k++){
        if((Ac[k]*a0<=0 || fabs(Ac[k])>=fabs(a0)) && fabs(Ac[k])<=amax){
            l=0;
            while(l<n && Ac[k]>data->int_a[l]){
                l++;
            }
            if(l==n || fabs(Ac[k]-data->int_a[l])>EPSI){
                for(i=n+1;i>l;i--){
                    data->int_a[i]=data->int_a[i-1];
                }
                data->int_a[l]=Ac[k];
                n++;
            }
        }
    }
    n--;
    data->cases[0]=n;
    a1=data->int_a[0];
    data->int_v[0]=v_c(a1,0,data);
    for(k=1;k<=n;k++){
        a1=data->int_a[k-1];
        a2=data->int_a[k];
        data->int_v[k]=v_c(a2,0,data);
        data->cases[k]=get_case((a1+a2)/2,data);
    }
    return data;
}

double a_b(double vC,quad_data *data){

    double v0=data->v0;
    double a0=data->a0;
    double amax=data->amax;
    double jmax=data->jmax;
    double smax=data->smax;
    double a0_2=a0*a0;
    double a0_3=a0_2*a0;
    int nb_int=data->cases[0];
    double jmax_2=jmax*jmax;
    double jmax_4=jmax_2*jmax_2;
    double smax_2=smax*smax;

    if(vC<data->int_v[0]){
        data->durations[B]=(data->int_v[0]-vC)/amax;
        return -amax;
    }

    if(vC>data->int_v[nb_int]){
        data->durations[B]=(vC-data->int_v[nb_int])/amax;
        return amax;
    }

    data->durations[B]=0;
    int k=1;
    while(k<=nb_int && vC>data->int_v[k]){
        k++;
    }
    int index_int=k;
    if(index_int>nb_int){
        printf("error: a_b: vC not found\n");
        return a0;
    }
    int case_ABC=data->cases[index_int];
    double a,b,c,d;
    double *sol;
    switch(case_ABC){
    case 210:
        return -SQR(jmax-sqrt(4*sqrt(smax*(a0*jmax_2+a0_2*smax+2*jmax*smax*(v0-vC)))+jmax_2))/(4*smax);
    case 211:
        return (jmax_2-sqrt(jmax_4+2*a0_2*smax_2+4*jmax*smax_2*(v0-vC)+2*a0*jmax_2*smax))/(2*smax);
    case 2010:
        return SQR(jmax-sqrt(4*sqrt(smax*(a0_2*smax-a0*jmax_2+2*jmax*smax*(vC-v0)))+jmax_2))/(4*smax);
    case 2011:
        return (sqrt(jmax_4+2*a0_2*smax_2+4*jmax*smax_2*(vC-v0)-2*a0*jmax_2*smax)-jmax_2)/(2*smax);
    case 200:
        a=2*sqrt(smax)*(v0-vC)/a0;
        b=-a0;
        c=-4*a0*sqrt(smax)*(v0-vC)/a0;
        d=-(smax*SQR(v0-vC)+a0_3)/a0;
        sol=poly_root_4(1,a,b,c,d);
        for(k=1;k<=(int)sol[0];k++){
            sol[k]=-SQR(sol[k])+a0;
        }
        break;
    case 201:
        a=2*jmax/sqrt(smax);
        b=(jmax_2-2*a0*smax)/smax;
        c=-4*a0*jmax/sqrt(smax);
        d=-(a0*jmax_2-a0_2*smax+2*jmax*smax*(v0-vC))/smax;
        sol=poly_root_4(1,a,b,c,d);
        for(k=1;k<=(int)sol[0];k++){
            sol[k]=-SQR(sol[k])+a0;
        }
        break;
    case 2000:
        a=2*sqrt(smax)*(vC-v0)/a0;
        b=-a0;
        c=0;
        d=-(smax*SQR(vC-v0)+a0_3)/a0;
        sol=poly_root_4(1,a,b,c,d);
        for(k=1;k<=(int)sol[0];k++){
            sol[k]=SQR(sol[k]);
        }
        break;
    case 2001:
        a=2*jmax/sqrt(smax);
        b=(2*a0*smax+jmax_2)/smax;
        c=4*a0*jmax/sqrt(smax);
        d=(a0*jmax_2+a0_2*smax+2*jmax*smax*(v0-vC))/smax;
        sol=poly_root_4(1,a,b,c,d);
        for(k=1;k<=(int)sol[0];k++){
            sol[k]=SQR(sol[k])+a0;
        }
        break;
    default:
        free(sol);
        return data->int_a[index_int];
    }
    int N=(int)sol[0];
    double dVmin,aB,dV,mina,maxa;
    if(N>0){
        dVmin=fabs(vC-v_c(sol[1],0,data));
        aB=sol[1];
        for(k=2;k<=N;k++){
            dV=fabs(vC-v_c(sol[k],0,data));
            if(dV<dVmin){
                dVmin=dV;
                aB=sol[k];
            }
        }
        dV=dVmin;
    }
    else{
        dV=EPSI+1;
    }
    free(sol);
    if(dV>EPSI){
        mina=data->int_a[index_int-1];
        maxa=data->int_a[index_int];
        aB=(mina+maxa)/2;
        dV=vC-v_c(aB,0,data);
        while(fabs(dV)>EPSI && fabs(mina-maxa)>EPSI){
            if(dV>0){
                mina=aB;
            }
            else{
                maxa=aB;
            }
            aB=(mina+maxa)/2;
            dV=vC-v_c(aB,0,data);
        }
    }

    return aB;
}

double get_v_root(quad_data *data){
    double Times[8];
    Times[0]=0;
    Times[1]=data->durations[A1];
    Times[2]=Times[1]+data->durations[A2];
    Times[3]=Times[2]+data->durations[A1];
    Times[4]=Times[3]+data->durations[B];
    Times[5]=Times[4]+data->durations[C1];
    Times[6]=Times[5]+data->durations[C2];
    Times[7]=Times[6]+data->durations[C1];
    double S[7]={ data->smax*data->signs[A],0,
                  -data->smax*data->signs[A],0,
                  data->smax*data->signs[C],0,
                  -data->smax*data->signs[C]};
    double j_cur=0;
    double a_cur=data->a0;
    double v_cur=data->v0;
    for(int i=0;i<7;i++){
        double tmin=Times[i];
        double tmax=Times[i+1];
        double *sol=poly_root_3(S[i]/6,j_cur/2,a_cur,v_cur);
        for(int j=1;j<=(int)sol[0];j++){
            double root=sol[j]+tmin;
            if(tmin<=root && root<=tmax){
                free(sol);
                return root;
            }
        }
        double trel=tmax-tmin;
        double trel_2=trel*trel;
        double trel_3=trel*trel_2;
        v_cur=S[i]*trel_3/6+j_cur*trel_2/2+a_cur*trel+v_cur;
        a_cur=S[i]*trel_2/2+j_cur*trel+a_cur;
        j_cur=S[i]*trel+j_cur;
        free(sol);
    }

    return -1;
}

double spline_at(double t,double x0,quad_data *data,double *v){
    if(t<0)
        return x0;
    double Times[8];
    Times[0]=0;
    Times[1]=data->durations[A1];
    Times[2]=Times[1]+data->durations[A2];
    Times[3]=Times[2]+data->durations[A1];
    Times[4]=Times[3]+data->durations[B];
    Times[5]=Times[4]+data->durations[C1];
    Times[6]=Times[5]+data->durations[C2];
    Times[7]=Times[6]+data->durations[C1];
    double S[7]={ data->smax*data->signs[A],0,
                  -data->smax*data->signs[A],0,
                  data->smax*data->signs[C],0,
                  -data->smax*data->signs[C]};
    double j_cur=0;
    double a_cur=data->a0;
    double v_cur=data->v0;
    double x_cur=x0;
    for(int i=0;i<7;i++){
        double tmin=Times[i];
        double tmax=Times[i+1];
        if(tmin<=t && t<=tmax){
            double trel=t-tmin;
            double trel_2=trel*trel;
            double trel_3=trel*trel_2;
            double trel_4=trel*trel_3;
            *v=S[i]*trel_3/6+j_cur*trel_2/2+a_cur*trel+v_cur;
            return S[i]*trel_4/24+j_cur*trel_3/6+a_cur*trel_2/2+v_cur*trel+x_cur;
        }
        else{
            double trel=tmax-tmin;
            double trel_2=trel*trel;
            double trel_3=trel*trel_2;
            double trel_4=trel*trel_3;
            x_cur=S[i]*trel_4/24+j_cur*trel_3/6+a_cur*trel_2/2+v_cur*trel+x_cur;
            v_cur=S[i]*trel_3/6+j_cur*trel_2/2+a_cur*trel+v_cur;
            a_cur=S[i]*trel_2/2+j_cur*trel+a_cur;
            j_cur=S[i]*trel+j_cur;
        }
    }
    *v=v_cur;
    return x_cur;
}

double x_bound(double v0,double a0,double xmin,double xmax,double vmax,double amax,double jmax,double smax,bool backward){
    double bound=0;
    quad_data *data=intervals(v0,a0,amax,jmax,smax);
    double vC=-sign(v0)*vmax;
    double aB=a_b(vC,data);
    durations_and_signs(aB,data);
    double root=get_v_root(data);
    if(root>0){
        double v;
        double x=spline_at(root,0,data,&v);
        if(backward)
            x=-x;
        if(fabs(v)>1e-4)
            printf("error: x_bound: v(root)=%f\n",v);
        if(backward){
            if(v0>0)
                bound=xmin-x;
            else
                bound=xmax-x;
        }
        else{
            if(v0<0)
                bound=xmin-x;
            else
                bound=xmax-x;
        }
    }
    else
        printf("error: x_bound: v_root not found\n");
    MY_FREE(data,quad_data,1);
    return bound;
}

double get_a_root(double a0,double aB,double smax,double t1,double t2){
    double dB0=sign(aB-a0);
    double delta=-2*dB0*smax*a0;
    double root;
    if(delta>=0){
        root=sqrt(delta)/(dB0*smax);
        if(root<0)
            root=-root;
        if(root<=t1)
            return root;
    }
    root=-(-dB0*smax*t1*t1+2*a0)/(2*dB0*smax*t1);
    if(t1<=root && root<=t1+t2)
        return root;
    delta=2*dB0*smax*(dB0*smax*t1*(t1+t2)+a0);
    if(delta>=0){
        double b=dB0*smax*(2*t1+t2);
        root=(-b-sqrt(delta))/(dB0*smax);
        if(t1+t2<=root && root<=2*t1+t2)
            return root;
        root=(-b+sqrt(delta))/(dB0*smax);
        if(t1+t2<=root && root<=2*t1+t2)
            return root;
    }
    printf("error: get_a_root: cannot find root :(\n");
    return -1;
}

double get_velocity(double t,double v0,double a0,double aB,double smax,double t1,double t2){
   double dB0=sign(aB-a0);
   if(t<t1)
      return dB0*smax*t*t*t/6+a0*t+v0;
   if(t<t1+t2)
      return dB0*smax*t1*(t-t1)*(t-t1)/2+(dB0*smax*t1*t1/2+a0)*(t-t1)+dB0*smax*t1*t1*t1/6+a0*t1+v0;
   return -dB0*smax*(t-t1-t2)*(t-t1-t2)*(t-t1-t2)/6+dB0*smax*t1*(t-t1-t2)*(t-t1-t2)/2+(dB0*smax*t1*t2+dB0*smax*t1*t1/2+a0)*(t-t1-t2)+dB0*smax*t1*t2*t2/2+(dB0*smax*t1*t1/2+a0)*t2+dB0*smax*t1*t1*t1/6+a0*t1+v0;
}
