#include "API/Roadmap/graphupdater.hpp"
#include "API/Roadmap/edge.hpp"
#include "API/ConfigSpace/configuration.hpp"
#include <time.h>
GraphUpdater* API_activeGraphUpdater = NULL;

using namespace std;
using namespace API;

GraphUpdater::GraphUpdater(Graph * graph){
	this->graph = graph;

	Configuration startConf(this->graph->getRobot(), this->graph->getRobot()->getRobotStruct()->ROBOT_POS);
	Configuration goalConf(this->graph->getRobot(), this->graph->getRobot()->getRobotStruct()->ROBOT_GOTO);

	this->start = this->graph->searchConf(startConf);
	if(this->start == NULL)
	{
		cout<<"Warning, start is not in graph, using first node in graph"<<endl;
		this->start = graph->getNodes()[0];
	}

	this->goal = this->graph->searchConf(goalConf);
	if(this->goal == NULL)
	{
		cout<<"Warning, goal is not in graph, using last node in graph"<<endl;
		this->goal = graph->getLastNode();
	}

    this->trajectories.push_back(this->graph->extractAStarShortestNodePaths(this->start,this->goal));
	
}

vector<Node*> GraphUpdater::getChildrenOf(Graph* graph, vector<Node*> nodes)
{
    set<Node*> resultSet;

    for(unsigned int i = 0; i < nodes.size(); i++)
    {
        vector<Edge*> edges = graph->getEdges(nodes[i],true);

        for(unsigned int j = 0; j < edges.size(); j++)
        {
            resultSet.insert(edges[j]->getTarget());
        }
    }


    vector<Node*> result(resultSet.begin(),resultSet.end());

    return result;
}

vector<Node*> GraphUpdater::getParentOf(Graph* graph, vector<Node*> nodes)
{
	set<Node*> resultSet;

	for(unsigned int i = 0; i < nodes.size(); i++)
	{
		vector<Edge*> edges = graph->getEdges(nodes[i],true);

		for(unsigned int j = 0; j < edges.size(); j++)
		{
            resultSet.insert(edges[j]->getSource());
		}
	}


	vector<Node*> result(resultSet.begin(),resultSet.end());

	return result;
}

unsigned int GraphUpdater::connectNodes(vector<Node*> nodes)
{
    unsigned int allIsValid = 0;
    for(unsigned int i = 0; i < (nodes.size() - 1); i++)
    {
        localPathPtr_t path(new LocalPath(nodes.at(i)->getConfiguration(), nodes.at(i+1)->getConfiguration()));
        if (path->isValid())
        {
            this->graph->addEdges(nodes.at(i),nodes.at(i+1),path);
            allIsValid++;
        }
    }

    if(allIsValid == nodes.size() - 1)
    {
        nodes.insert(nodes.begin(), this->start);
        nodes.push_back(this->goal);
        vector<Node *> newPath;
        newPath = this->graph->extractAStarShortestNodePaths(nodes);
        if(newPath.size() > 0)
            this->trajectories.push_back(newPath);
    }

    return allIsValid;
}

double GraphUpdater::reduction(unsigned int param){
	vector<Node*> nodes = this->graph->getNodes();
    double result = 1.0;
    result /= (double)(nodes.size());

    if(this->trajectories.size() == 0)
	{
		cout<<"Error, no valid trajectory available"<<endl;
		return 0;
	}

    vector<Node*> compareNodes;

    for(unsigned int i = 0; i < this->trajectories.size(); i++)
    {
        compareNodes.insert(compareNodes.end(), this->trajectories.at(i).begin(), this->trajectories.at(i).end());
    }

	bool isIn = false;

	Node* currentNode;
    set<Node*> nodesToRemove(nodes.begin(),nodes.end());
    for(unsigned int pi = 0; pi <= param; pi++)
    {
        for (set<Node*>::iterator it=nodesToRemove.begin(); it!=nodesToRemove.end(); ++it)
        {
            currentNode = *it;

            isIn = false;
            for(unsigned int ci = 0; ci < compareNodes.size(); ci++)
            {
                if(currentNode == compareNodes[ci])
                {
                    isIn = true;
                    break;
                }
            }

            if(isIn)
            {
                nodesToRemove.erase(currentNode);
            }

        }

        compareNodes = this->getChildrenOf(this->graph, compareNodes);

    }

    for (set<Node*>::iterator it=nodesToRemove.begin(); it!=nodesToRemove.end(); ++it)
    {
        this->graph->removeNode(*it);
    }
    nodes = this->graph->getNodes();
    result *= (double)(nodes.size());
	return result;
}

bool GraphUpdater::connexity(unsigned int param){

    Node* connectStart, *connectGoal;

    set<Node *> nodePaths;

    for(unsigned int i = 0; i < trajectories.size(); i++)
    {
        copy( trajectories.at(i).begin(), trajectories.at(i).end(), inserter( nodePaths, nodePaths.end() ) );
    }

    vector<Node *> nodePathsVector(nodePaths.begin(),nodePaths.end());

    if(param == 1)
    {
        connectStart = nodePathsVector.at(rand() % nodePathsVector.size());
        do{
            connectGoal = nodePathsVector.at(rand() % nodePathsVector.size());
        }while(connectStart == connectGoal && this->graph->isEdgeInGraph(connectStart, connectGoal));
    }
    else
    {
        vector<Node *> nodeNeighborsVector = this->getChildrenOf(this->graph, nodePathsVector);
        unsigned int randStart = rand() % 100;
        unsigned int randGoal = rand() % 100;
        if(randStart < param)
        {
            connectStart = nodePathsVector.at(rand() % nodePathsVector.size());
        }
        else
        {
            connectStart = nodeNeighborsVector.at(rand() % nodeNeighborsVector.size());
        }

        do{
            if(randGoal < param)
            {
                connectGoal = nodePathsVector.at(rand() % nodePathsVector.size());
            }
            else
            {
                connectGoal = nodeNeighborsVector.at(rand() % nodeNeighborsVector.size());
            }
        }while(connectStart == connectGoal && this->graph->isEdgeInGraph(connectStart, connectGoal));
    }

    localPathPtr_t path(new LocalPath(connectStart->getConfiguration(), connectGoal->getConfiguration()));
    if (path->isValid())
    {
        this->graph->addEdges(connectStart,connectGoal,path);
        vector<Node *> newPath;
        vector<Node *> passingBy;

        passingBy.push_back(this->start);
        passingBy.push_back(connectStart);
        passingBy.push_back(connectGoal);
        passingBy.push_back(this->goal);

        newPath = this->graph->extractAStarShortestNodePaths(passingBy);


        if(newPath.size() > 0)
            this->trajectories.push_back(newPath);

        return true;
    }

	return false;	
}
bool GraphUpdater::coverage(unsigned int param){

    Node* connectStart, *connectGoal;

    set<Node *> nodePaths;

    for(unsigned int i = 0; i < trajectories.size(); i++)
    {
        copy( trajectories.at(i).begin(), trajectories.at(i).end(), inserter( nodePaths, nodePaths.end() ) );
    }

    vector<Node *> nodePathsVector(nodePaths.begin(),nodePaths.end());

    if(param == 1)
    {
        connectStart = nodePathsVector.at(rand() % nodePathsVector.size());
        do{
            connectGoal = nodePathsVector.at(rand() % nodePathsVector.size());
        }while(connectStart == connectGoal && this->graph->isEdgeInGraph(connectStart, connectGoal));
    }
    else
    {
        vector<Node *> nodeNeighborsVector = this->getChildrenOf(this->graph, nodePathsVector);
        unsigned int randStart = rand() % 100;
        unsigned int randGoal = rand() % 100;
        if(randStart < param)
        {
            connectStart = nodePathsVector.at(rand() % nodePathsVector.size());
        }
        else
        {
            connectStart = nodeNeighborsVector.at(rand() % nodeNeighborsVector.size());
        }

        do{
            if(randGoal < param)
            {
                connectGoal = nodePathsVector.at(rand() % nodePathsVector.size());
            }
            else
            {
                connectGoal = nodeNeighborsVector.at(rand() % nodeNeighborsVector.size());
            }
        }while(connectStart == connectGoal && this->graph->isEdgeInGraph(connectStart, connectGoal));
    }

    confPtr_t start = connectStart->getConfiguration();
    confPtr_t goal = connectGoal->getConfiguration();

    double distance = start->distance(goal.get());
    double interdistance = numeric_limits<double>::max();
    confPtr_t inter(new Configuration(this->graph->getRobot()));
    do{
        p3d_shoot ( this->graph->getRobot()->getRobotStruct(), inter->getConfigStruct() , FALSE);

        interdistance = max(start->distance(inter.get()), goal->distance(inter.get()));
        inter->setAsNotTested();
    }while(inter->isInCollision() || interdistance > distance);

    Node* interNode = new Node(this->graph,inter);
    this->graph->addNode(interNode);

    localPathPtr_t path(new LocalPath(connectStart->getConfiguration(), interNode->getConfiguration()));
    if (path->isValid())
    {
        this->graph->addEdges(connectStart,interNode,path);
    }

    localPathPtr_t path1(new LocalPath(interNode->getConfiguration(), connectGoal->getConfiguration()));
    if (path1->isValid())
    {
        this->graph->addEdges(interNode,connectGoal,path);
    }

    if((!path->isValid()) && (!path1->isValid()))
    {
        this->graph->removeNode(interNode);
        return false;
    }

    if(path->isValid() && path1->isValid())
    {

    vector<Node *> passingBy;
    passingBy.push_back(this->start);
    passingBy.push_back(connectStart);
    passingBy.push_back(interNode);
    passingBy.push_back(connectGoal);
    passingBy.push_back(this->goal);

    vector<Node *> newPath = this->graph->extractAStarShortestNodePaths(passingBy);

    if(newPath.size() > 0)
        this->trajectories.push_back(newPath);
    }

    return true;
}

bool GraphUpdater::correction(unsigned int param, double timelimit){

    confPtr_t fromQ;
    for(unsigned int i = 0; i < trajectories.size(); i++)
    {
        for(unsigned int j = 0; j < trajectories.at(i).size(); j++)
        {
            trajectories.at(i).at(j)->getConfiguration()->setAsNotTested();
            if(trajectories.at(i).at(j)->getConfiguration()->isInCollision())
            {
                fromQ = trajectories.at(i).at(j)->getConfiguration();
                return this->correction(fromQ,param,timelimit);
            }
        }
    }

    return true;
}

bool GraphUpdater::correction(Node* from, unsigned int param, double timelimit){

    confPtr_t fromQ;
    fromQ = from->getConfiguration();

    return this->correction(fromQ,param,timelimit);
}

bool GraphUpdater::correction(confPtr_t from, unsigned int param, double timelimit){

    if(timelimit <= 0)
    {
        timelimit = numeric_limits<double>::max();
    }

    double timer,timer1;
    vector<Node *> closestNode;
    closestNode = this->graph->kNearestNeighbours(1,from);

    vector<vector<Node*> > trajectoriesTocorrect;

    trajectoriesTocorrect = this->getAvailableTrajectories(closestNode.at(0),false);

    bool corrected = false;

    if(trajectoriesTocorrect.size() > 0)
    {
        vector<Node*> nodesBefore;
        vector<Node*> nodesAfter;

        for(unsigned int i = 0; i < trajectoriesTocorrect.size(); i++)
        {
            for(unsigned int j = 0; j < trajectoriesTocorrect.at(i).size(); j++)
            {
                if(trajectoriesTocorrect.at(i).at(j) == closestNode.at(0))
                {
                    if(j == 0)
                    {
                        nodesAfter.push_back(trajectoriesTocorrect.at(i).at(j+1));
                    }
                    else if(j == (trajectoriesTocorrect.at(i).size() - 1))
                    {
                        nodesBefore.push_back(trajectoriesTocorrect.at(i).at(j-1));
                    }
                    else
                    {
                        nodesBefore.push_back(trajectoriesTocorrect.at(i).at(j-1));
                        nodesAfter.push_back(trajectoriesTocorrect.at(i).at(j+1));
                    }
                }
            }
        }
        timer = (double) clock()/CLOCKS_PER_SEC;
        unsigned int p = 0;
        do{


            for(unsigned int i = 0; i < nodesBefore.size(); i++)
            {
                for(unsigned int j = 0; j < nodesAfter.size(); j++)
                {
                    confPtr_t start = nodesBefore.at(i)->getConfiguration();
                    confPtr_t goal = nodesAfter.at(j)->getConfiguration();

                    double distance = start->distance(goal.get());
                    double interdistance = numeric_limits<double>::max();
                    confPtr_t inter(new Configuration(this->graph->getRobot()));
                    do{
                        p3d_shoot ( this->graph->getRobot()->getRobotStruct(), inter->getConfigStruct() , FALSE);

                        interdistance = max(start->distance(inter.get()), goal->distance(inter.get()));

                    }while(inter->isInCollision() || interdistance > distance);

                    Node* interNode = new Node(this->graph,inter);
                    this->graph->addNode(interNode);

                    vector<Node *> attempt;

                    attempt.push_back(nodesBefore.at(i));
                    attempt.push_back(interNode);
                    attempt.push_back(nodesAfter.at(j));
                    unsigned int attemptResult = this->connectNodes(attempt);
                    if(attemptResult == 0)
                    {
                        this->graph->removeNode(interNode);
                    }
                    else if(attemptResult == 2)
                    {
                        corrected = true;
                    }
                    timer1 = (double) clock()/CLOCKS_PER_SEC;
                    if(timelimit > (timer - timer1))
                    {
                        return corrected;
                    }
                }
            }

            nodesBefore = this->getParentOf(this->graph, nodesBefore);
            nodesAfter = this->getChildrenOf(this->graph, nodesAfter);

            p++;
            timer1 = (double) clock()/CLOCKS_PER_SEC;
        }while(!corrected && p < param && timelimit > (timer - timer1));

    }

    return corrected;
}

vector<vector<Node*> > GraphUpdater::getAvailableTrajectories(confPtr_t from, bool computeNew){
    Node* fromNode;

    fromNode = this->graph->searchConf(*from);
    if(fromNode == NULL && computeNew)
    {
        fromNode = new Node(this->graph,from);
        this->graph->addNode(fromNode);
    }
    else
    {
        vector<vector<Node*> > result;
        return result;
    }


    return this->getAvailableTrajectories(fromNode,computeNew);
}

vector<vector<Node*> > GraphUpdater::getAvailableTrajectories(Node* from, bool computeNew ){
    vector<vector<Node*> > result;

    for(unsigned int i = 0; i < trajectories.size(); i++)
    {
        for(unsigned int j = 0; j < trajectories.at(i).size(); j++)
        {
            if(from == trajectories.at(i).at(j))
            {
                result.push_back(trajectories.at(i));
                break;
            }
        }
    }

	return result;
}

Trajectory* GraphUpdater::getBestTrajectory(confPtr_t from, bool updateEnv){
    Node* fromNode;

    fromNode = this->graph->searchConf(*from);

    return this->getBestTrajectory(fromNode,updateEnv);
}

Trajectory* GraphUpdater::getBestTrajectory(Node* from, bool updateEnv){
	Trajectory* result;

	return result;
}

void GraphUpdater::updateTrajectories(){

}

vector<Trajectory*> GraphUpdater::getTrajectories(){
    this->updateTrajectories();

    vector<Trajectory*> result;

    for(unsigned int i = 0; i < trajectories.size(); i++)
    {
        // build the trajectory
        bool toAdd = true;
        Trajectory * trajectory = new Trajectory(this->graph->getRobot());
        for (unsigned j(trajectories.at(i).size() - 1); j > 0; --j)
        {
            // TODO DD Get the edges going out of nodes[i], for better performance
            Edge * edge = this->graph->isEdgeInGraph(trajectories.at(i)[j], trajectories.at(i)[j - 1]);
            if (!edge)
            {
                toAdd = false;
                break;
            }
            else
            {
                trajectory->push_back(edge->getLocalPath());
            }
        }

        if(toAdd)
        {
            result.push_back(trajectory);
        }

    }
    return result;
}

void GraphUpdater::setCurrentTrajectory(Trajectory* trajectory)
{
    currentTrajectory = trajectory;
}

Trajectory* GraphUpdater::getCurrentTrajectory()
{
    return currentTrajectory;
}

void GraphUpdater::setCurrentConfiguration(confPtr_t configuration)
{
    currentConfiguration = configuration;

    LocalPath* path = currentTrajectory->getLocalPathPtrOf(currentConfiguration);

    if(path == NULL)
    {
        cout<<"Error config not found on current trajectory"<<endl;
    }
    else
    {
        prevNode = this->graph->searchConf(*path->getBegin());
        nextNode = this->graph->searchConf(*path->getEnd());
    }
}

Node* GraphUpdater::getPrevNode()
{
    return prevNode;
}

Node* GraphUpdater::getNextNode()
{
    return nextNode;
}
