#include "rectangleForm.hpp"
#include "geometricComputations.hpp"
using namespace std;

RectangleForm::RectangleForm(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_vector3 norm)
{
    _a = a;
    _b = b;
    _c = c;
    _d = d;
    p3d_vectCopy(norm,_norm);
    type = "rectangle";
    isSupport = true;
}

bool RectangleForm::isPointInForm(p3d_point p)
{
    GeometricComputations GC = GeometricComputations();

    return GC.point_in_rect2(_a,_b,_c,_d,p);
}

double RectangleForm::distHToPoint(p3d_point p)
{
    //TODO: compute the real distance
    double r = (_a.z + _b.z +_c.z +_d.z)/4;
    return p.z - r;
}

double RectangleForm::getDistanceFromPoint(p3d_point p)
{
    double dist = 1000;

    double tmp;

    tmp = sqrt(pow(p.x - _a.x,2) + pow(p.y - _a.y,2) + pow(p.z - _a.z,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _b.x,2) + pow(p.y - _b.y,2) + pow(p.z - _b.z,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _c.x,2) + pow(p.y - _c.y,2) + pow(p.z - _c.z,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _d.x,2) + pow(p.y - _d.y,2) + pow(p.z - _d.z,2));
    if (tmp < dist)
        dist = tmp;

    return dist;
}

double RectangleForm::getPlannarDistanceFromPoint(p3d_point p)
{

    double dist = 1000;

    double tmp;

    tmp = sqrt(pow(p.x - _a.x,2) + pow(p.y - _a.y,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _b.x,2) + pow(p.y - _b.y,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _c.x,2) + pow(p.y - _c.y,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _d.x,2) + pow(p.y - _d.y,2));
    if (tmp < dist)
        dist = tmp;

    return dist;
}

p3d_point RectangleForm::getRandomPointInForm()
{
    GeometricComputations GC = GeometricComputations();
    return GC.getRandomPointInRectangle(_a,_b,_c,_d);
}

vector<p3d_point> RectangleForm::getFormGrid(double sampleRate)
{
    vector<p3d_point> t;
    GeometricComputations GC = GeometricComputations();
    int xlenght = GC.distBetweenPoints(_a,_b)/sampleRate;
    int ylenght = GC.distBetweenPoints(_a,_c)/sampleRate;
    double normeAB = GC.distBetweenPoints2D(_a,_b);
    double normeAC = GC.distBetweenPoints2D(_a,_c);

    p3d_point vectAB;
    vectAB.x = (_b.x - _a.x)/normeAB;
    vectAB.y = (_b.y - _a.y)/normeAB;
    p3d_point vectAC;
    vectAC.x = (_c.x - _a.x)/normeAC;
    vectAC.y = (_c.y - _a.y)/normeAC;


    for (int i = 1; i <= xlenght; i++)
    {
        for (int j = 1; j <= ylenght; j++)
        {
            p3d_point p;
            p.x = i*sampleRate*vectAB.x + j*sampleRate*vectAC.x + _a.x;
            p.y = i*sampleRate*vectAB.y + j*sampleRate*vectAC.y + _a.y;
            p.z = _a.z;
            t.push_back(p);
        }
    }

    return t;
}

GeometricForm* RectangleForm::getTransformedForm(p3d_matrix4 M)
{
    p3d_vector3 b;
    transformVector(_norm,M,b);
    return new RectangleForm(transformPoint(_a,M),
                             transformPoint(_b,M),
                             transformPoint(_c,M),
                             transformPoint(_d,M),
                             b);
}

void RectangleForm::draw()
{
    double colorvector[4];
    colorvector[0] = 0.0;       //red
    colorvector[1] = 0.0;       //green
    colorvector[2] = 1.0;       //blue
    colorvector[3] = 0.5;       //transparency

    glEnable(GL_BLEND);
    g3d_set_color(Any,colorvector);
    p3d_vector3 p0,p1,p2,p3;
    p0[0] = _a.x; p0[1] = _a.y; p0[2] = _a.z;
    p1[0] = _b.x; p1[1] = _b.y; p1[2] = _b.z;
    p2[0] = _c.x; p2[1] = _c.y; p2[2] = _c.z;
    p3[0] = _d.x; p3[1] = _d.y; p3[2] = _d.z;


    g3d_drawOneTriangle(p0,p1,p2,Blue,colorvector);
    g3d_drawOneTriangle(p3,p1,p2,Blue,colorvector);
}
