#include "gtpPlan.hpp"

#include "GTP/Tasks/task.hpp"
#include "taskSolution.hpp"
#include "jsoncpp/json/value.h"

using namespace std;

string fileName = "planDraw";

PlanTree::PlanTree()
{
    string s = string(getenv("HOME_MOVE3D"));
    s += "/";
    s+= fileName;
    myfilepath = s;

    ofstream myfile;
    myfile.open(myfilepath.c_str());
    myfile << "";
    myfile.close();

}

PlanTree::~PlanTree()
{

}

bool PlanTree::checkIfPlanInTree(OnePlan* OP)
{
    for (unsigned int i = 0; i < OP->getPlan().size(); i++)
    {
        PlanNode* PN = OP->getPlan().at(i);

        if (PN->getPrevious())
        {
            if (PN->getPrevious()->isInNexts(PN))
                continue;
            else
                return false;
        }
    }
    return true;
}

OnePlan* PlanTree::getPlanFromNode(PlanNode* PN)
{
    vector<PlanNode*> tmp;
    tmp.push_back(PN);
    PlanNode* curr = PN;
    while(curr->getPrevious())
    {
        tmp.push_back(curr->getPrevious());
        curr = curr->getPrevious();
    }
    OnePlan* OP = new OnePlan(PN->getPlanTree());
    for(unsigned int i = tmp.size() - 1 ; i >=0; i++)
    {
        OP->addToPlan(tmp.at(i));
    }
    return OP;
}

PlanNode* PlanTree::getNodeFromId(int idFirst, int idSecond)
{
    for(unsigned int i = 0; i < _tree.size(); i++)
    {
        if (idFirst == _tree.at(i)->getId().first && idSecond == _tree.at(i)->getId().second )
        {
            return _tree.at(i);
        }
    }
    return NULL;
}

PlanNode* PlanTree::getNodeFormTS(TaskSolution* TS)
{
    for(unsigned int i = 0; i < _tree.size(); i++)
    {
        if (TS->getTask()->getId() == _tree.at(i)->getId().first && TS->getId() == _tree.at(i)->getId().second )
        {
            return _tree.at(i);
        }
    }
    return NULL;
}


bool PlanTree::addPlanNode(PlanNode* PN, TaskSolution* prevTS) /*{_tree.push_back(PN);*/
{
//    PlanNode* pr = NULL;

    if (PN->hasSucceeded())
    {
        for(unsigned int i = 0; i < _tree.size(); i++)
        {
            PlanNode* tmp = _tree.at(i);
            if (PN->getId().first == tmp->getId().first && PN->getId().second == tmp->getId().second )
            {
                cout << "Node already exists" <<endl;
                return false;
            }
        }
    }

    TaskSolution* prev = prevTS;

//    prev = PN->getTaskSolution()->getStartWS()->getTSThatEndWithThisWS();

    PlanNode* prPN = NULL;
    if(prev)
    {
        prPN = getNodeFromId(prev->getTask()->getId(), prev->getId());
    }

    _tree.push_back(PN);
    PN->setPrevious(prPN);
    if (!prPN)
    {
        cout << "no previous has been found for this task node, considered as first in a tree" << endl;
    }
    //adding the node to the file
    printInFile(PN);
    return true;
}

std::vector<double> PlanTree::getAllImpoInfo()
{
    vector<double> times;
    vector<double> timesMp;
    vector<double> nbTests;
    vector<double> nbGik;
    vector<double> nbMP;

    vector<double> timesSuccess;
    vector<double> timesMpSuccess;
    vector<double> nbTestsSuccess;
    vector<double> nbGikSuccess;
    vector<double> nbMPSuccess;

    vector<double> timesFail;
    vector<double> timesMpFail;
    vector<double> nbTestsFail;
    vector<double> nbGikFail;
    vector<double> nbMPFail;

    for (unsigned int i = 0 ; i < _tree.size(); i++)
    {
        PlanNode* pn = _tree.at(i);
        times.push_back(pn->SD->getSDMapEntry("times")->getDoubleMapEntry("timeComp"));
        timesMp.push_back(pn->SD->getSDMapEntry("times")->getSDMapEntry("times")->getDoubleMapEntry("timeMp"));
        nbTests.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbTests"));
        nbGik.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbGikCalled"));
        nbMP.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbMPCalled"));

        if(pn->hasSucceeded())
        {
            timesSuccess.push_back(pn->SD->getSDMapEntry("times")->getDoubleMapEntry("timeComp"));
            timesMpSuccess.push_back(pn->SD->getSDMapEntry("times")->getSDMapEntry("times")->getDoubleMapEntry("timeMp"));
            nbTestsSuccess.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbTests"));
            nbGikSuccess.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbGikCalled"));
            nbMPSuccess.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbMPCalled"));
        }
        else
        {
            timesFail.push_back(pn->SD->getSDMapEntry("times")->getDoubleMapEntry("timeComp"));
            timesMpFail.push_back(pn->SD->getSDMapEntry("times")->getSDMapEntry("times")->getDoubleMapEntry("timeMp"));
            nbTestsFail.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbTests"));
            nbGikFail.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbGikCalled"));
            nbMPFail.push_back(pn->SD->getSDMapEntry("statistics")->getIntMapEntry("nbMPCalled"));
        }
    }

    double allTime = 0;
    double allTimeMp = 0;
    double meanTime = 0;
    double meanTimeMp = 0;

    double meanTests = 0;
    double meanGik = 0;
    double meanMP = 0;
    for (unsigned int i = 0; i < times.size(); i++)
    {
        meanTime += times.at(i);
        allTime += times.at(i);
        meanTimeMp += timesMp.at(i);
        allTimeMp += timesMp.at(i);

        meanTests += nbTests.at(i);
        meanGik += nbGik.at(i);
        meanMP += nbMP.at(i);

    }
    meanTime = meanTime/times.size();
    meanTimeMp = meanTimeMp/times.size();
    meanTests = meanTests/times.size();
    meanGik = meanGik/times.size();
    meanMP = meanMP/times.size();


    double timeVariance = 0;
    double timeMpVariance = 0;
    double testVariance = 0;
    double gikVariance = 0;
    double mpVariance = 0;
    for (unsigned int i = 0; i < times.size(); i++)
    {
        timeVariance += pow(times.at(i) - meanTime, 2);
        timeMpVariance += pow(timesMp.at(i) - meanTimeMp, 2);
        testVariance += pow(nbTests.at(i) - meanTests, 2);
        gikVariance += pow(nbGik.at(i) - meanGik, 2);
        mpVariance += pow(nbMP.at(i) - meanMP, 2);

    }

    timeVariance = timeVariance/times.size();
    timeMpVariance = timeMpVariance/times.size();
    testVariance = testVariance/times.size();
    gikVariance = gikVariance/times.size();
    mpVariance = mpVariance/times.size();

    double timeScareDist = sqrt(timeVariance);
    double timeMpScareDist = sqrt(timeMpVariance);
    double testScareDist = sqrt(testVariance);
    double gikScareDist = sqrt(gikVariance);
    double mpScareDist = sqrt(mpVariance);

    cout << "total time to compute tree = " << allTime << " s with " << times.size() << " actions " << " success rate " << timesSuccess.size()/(double)_tree.size() << endl;
    cout << "time  : mean = " << meanTime << " variance = " << timeVariance << " standard Deviation = " << timeScareDist << endl;
    cout << "tests : mean = " << meanTests << " variance = " << testVariance << " standard Deviation = " << testScareDist << endl;
    cout << "gik   : mean = " << meanGik << " variance = " << gikVariance << " standard Deviation = " << gikScareDist << endl;
    cout << "mp    : mean = " << meanMP << " variance = " << mpVariance << " standard Deviation = " << mpScareDist << endl;
    cout << "timeMp: mean = " << meanTimeMp << " variance = " << timeMpVariance << " standard Deviation = " << timeMpScareDist << endl;


    double allTimeFail = 0;
    double meanTimeFail = 0;
    double allTimeMpFail = 0;
    double meanTimeMpFail = 0;

    double meanTestsFail = 0;
    double meanGikFail = 0;
    double meanMPFail = 0;
    for (unsigned int i = 0; i < timesFail.size(); i++)
    {
        meanTimeFail += timesFail.at(i);
        allTimeFail += timesFail.at(i);

        meanTimeMpFail += timesMpFail.at(i);
        allTimeMpFail += timesMpFail.at(i);

        meanTestsFail += nbTestsFail.at(i);
        meanGikFail += nbGikFail.at(i);
        meanMPFail += nbMPFail.at(i);

    }
    meanTimeFail = meanTimeFail/timesFail.size();
    meanTimeMpFail = meanTimeMpFail/timesFail.size();
    meanTestsFail = meanTestsFail/timesFail.size();
    meanGikFail = meanGikFail/timesFail.size();
    meanMPFail = meanMPFail/timesFail.size();


    double timeVarianceFail = 0;
    double timeMpVarianceFail = 0;
    double testVarianceFail = 0;
    double gikVarianceFail = 0;
    double mpVarianceFail = 0;
    for (unsigned int i = 0; i < timesFail.size(); i++)
    {
        timeVarianceFail += pow(timesFail.at(i) - meanTimeFail, 2);
        timeMpVarianceFail += pow(timesMpFail.at(i) - meanTimeMpFail, 2);
        testVarianceFail += pow(nbTestsFail.at(i) - meanTestsFail, 2);
        gikVarianceFail += pow(nbGikFail.at(i) - meanGikFail, 2);
        mpVarianceFail += pow(nbMPFail.at(i) - meanMPFail, 2);

    }

    timeVarianceFail = timeVarianceFail/timesFail.size();
    timeMpVarianceFail = timeMpVarianceFail/timesFail.size();
    testVarianceFail = testVarianceFail/timesFail.size();
    gikVarianceFail = gikVarianceFail/timesFail.size();
    mpVarianceFail = mpVarianceFail/timesFail.size();

    double timeScareDistFail = sqrt(timeVarianceFail);
    double timeMpScareDistFail = sqrt(timeMpVarianceFail);
    double testScareDistFail = sqrt(testVarianceFail);
    double gikScareDistFail = sqrt(gikVarianceFail);
    double mpScareDistFail = sqrt(mpVarianceFail);

    cout << "total time to compute failed actions in tree = " << allTimeFail << " s with " << timesFail.size() << " actions " << endl;
    cout << "time  : mean = " << meanTimeFail << " variance = " << timeVarianceFail << " standard Deviation = " << timeScareDistFail << endl;
    cout << "tests : mean = " << meanTestsFail << " variance = " << testVarianceFail << " standard Deviation = " << testScareDistFail << endl;
    cout << "gik   : mean = " << meanGikFail << " variance = " << gikVarianceFail << " standard Deviation = " << gikScareDistFail << endl;
    cout << "mp    : mean = " << meanMPFail << " variance = " << mpVarianceFail << " standard Deviation = " << mpScareDistFail << endl;
    cout << "timeMp: mean = " << meanTimeMpFail << " variance = " << timeMpVarianceFail << " standard Deviation = " << timeMpScareDistFail << endl;





    double allTimeSuccess = 0;
    double meanTimeSuccess = 0;

    double allTimeMpSuccess = 0;
    double meanTimeMpSuccess = 0;

    double meanTestsSuccess = 0;
    double meanGikSuccess = 0;
    double meanMPSuccess = 0;
    for (unsigned int i = 0; i < timesSuccess.size(); i++)
    {
        meanTimeSuccess += timesSuccess.at(i);
        allTimeSuccess += timesSuccess.at(i);

        meanTimeMpSuccess += timesMpSuccess.at(i);
        allTimeMpSuccess += timesMpSuccess.at(i);

        meanTestsSuccess += nbTestsSuccess.at(i);
        meanGikSuccess += nbGikSuccess.at(i);
        meanMPSuccess += nbMPSuccess.at(i);

    }
    meanTimeSuccess = meanTimeSuccess/timesSuccess.size();
    meanTimeMpSuccess = meanTimeMpSuccess/timesMpSuccess.size();
    meanTestsSuccess = meanTestsSuccess/timesSuccess.size();
    meanGikSuccess = meanGikSuccess/timesSuccess.size();
    meanMPSuccess = meanMPSuccess/timesSuccess.size();


    double timeVarianceSuccess = 0;
    double timeMpVarianceSuccess = 0;
    double testVarianceSuccess = 0;
    double gikVarianceSuccess = 0;
    double mpVarianceSuccess = 0;
    for (unsigned int i = 0; i < timesSuccess.size(); i++)
    {
        timeVarianceSuccess += pow(timesSuccess.at(i) - meanTimeSuccess, 2);
        timeMpVarianceSuccess += pow(timesMpSuccess.at(i) - meanTimeMpSuccess, 2);
        testVarianceSuccess += pow(nbTestsSuccess.at(i) - meanTestsSuccess, 2);
        gikVarianceSuccess += pow(nbGikSuccess.at(i) - meanGikSuccess, 2);
        mpVarianceSuccess += pow(nbMPSuccess.at(i) - meanMPSuccess, 2);

    }

    timeVarianceSuccess = timeVarianceSuccess/timesSuccess.size();
    timeMpVarianceSuccess = timeMpVarianceSuccess/timesMpSuccess.size();
    testVarianceSuccess = testVarianceSuccess/timesSuccess.size();
    gikVarianceSuccess = gikVarianceSuccess/timesSuccess.size();
    mpVarianceSuccess = mpVarianceSuccess/timesSuccess.size();

    double timeScareDistSuccess = sqrt(timeVarianceSuccess);
    double timeMpScareDistSuccess = sqrt(timeMpVarianceSuccess);
    double testScareDistSuccess = sqrt(testVarianceSuccess);
    double gikScareDistSuccess = sqrt(gikVarianceSuccess);
    double mpScareDistSuccess = sqrt(mpVarianceSuccess);

    cout << "total time to compute successfull actions in tree = " << allTimeSuccess << " s with " << timesSuccess.size() << " actions " << endl;
    cout << "time  : mean = " << meanTimeSuccess << " variance = " << timeVarianceSuccess << " standard Deviation = " << timeScareDistSuccess << endl;
    cout << "tests : mean = " << meanTestsSuccess << " variance = " << testVarianceSuccess << " standard Deviation = " << testScareDistSuccess << endl;
    cout << "gik   : mean = " << meanGikSuccess << " variance = " << gikVarianceSuccess << " standard Deviation = " << gikScareDistSuccess << endl;
    cout << "mp    : mean = " << meanMPSuccess << " variance = " << mpVarianceSuccess << " standard Deviation = " << mpScareDistSuccess << endl;
    cout << "timeMp: mean = " << meanTimeMpSuccess << " variance = " << timeMpVarianceSuccess << " standard Deviation = " << timeMpScareDistSuccess << endl;






    vector<double> vres;
    vres.push_back(allTime);

    vres.push_back(meanTime);
    vres.push_back(timeVariance);
    vres.push_back(timeScareDist);

    vres.push_back(meanTests);
    vres.push_back(testVariance);
    vres.push_back(testScareDist);

    vres.push_back(meanGik);
    vres.push_back(gikVariance);
    vres.push_back(gikScareDist);

    vres.push_back(meanMP);
    vres.push_back(mpVariance);
    vres.push_back(mpScareDist);

    return vres;

}

void PlanTree::printAll()
{
    cout << "printing plan tree" <<endl;
    cout << "number of nodes = " << _tree.size() << endl;

    for (unsigned int i = 0 ; i < _tree.size(); i++)
    {
        cout << "node Id = (" << _tree.at(i)->getId().first << " , " <<  _tree.at(i)->getId().second << ")" << endl;
        if (_tree.at(i)->getPrevious())
            cout << "    with previous node with Id (" << _tree.at(i)->getPrevious()->getId().first << " , " <<  _tree.at(i)->getPrevious()->getId().second << ")" << endl;
        else
            cout << "    first in the tree" << endl;
    }

}

void PlanTree::printInFile(PlanNode* pn)
{
//    if (!pn->hasSucceeded())
//        return;
    if (pn && pn->getTask())
    {
        fstream myfile;

        myfile.open(myfilepath.c_str(),std::fstream::in | std::fstream::out | std::fstream::app);


        string st = pn->getTask()->getDescr();
        while(st.find(" ") != string::npos)
        {
            st.replace(st.find(" "), 1, "");
        }
        while(st.find("\t") != string::npos)
        {
            st.replace(st.find("\t"), 1, "");
        }
        while(st.find("\n") != string::npos)
        {
            st.replace(st.find("\n"), 1, "|");
        }

        myfile << pn->getId().first << " " <<
                  pn->getId().second << " " <<
                  st << " ";

        if (pn->getPrevious())
        {
            myfile << pn->getPrevious()->getId().first << " " << pn->getPrevious()->getId().second;
        }

        myfile << "\n";
        myfile.close();
    }


}

double PlanTree::computeDist(PlanNode* p1, PlanNode* p2)
{

    cout << "needs to be tested" << endl;
    if (!p1 || !p2)
        return P3D_HUGE;
    else if (p1 == p2)
        return 0;
    double res = 0;
    PlanNode* tmp = p1;
    vector<pair<double,PlanNode*> > firstPNBranch;
    while (tmp)
    {

        tmp = tmp->getPrevious();
        res += 1;
        pair<double, PlanNode*> PNB;

        if (tmp == p2)
            return res;

        PNB.first = res;
        PNB.second = tmp;
        firstPNBranch.push_back(PNB);

    }

    tmp = p2;

    res = 0;
    while (tmp)
    {

        tmp = tmp->getPrevious();
        res += 1;
        if (tmp == p1)
            return res;
        for (unsigned int i = 0; i< firstPNBranch.size();i++)
        {
            if (tmp == firstPNBranch.at(i).second)
                return res + firstPNBranch.at(i).first;
        }
    }


    return P3D_HUGE;
}

PlanNode* PlanTree::getClosestNode(WorldState *ws, type_data td)
{
    // tout doit etre sur 1
    // construire la map avec des 1 partout
    // construire la map avec les valeurs reels
    // remplacer dans la premiere map avec les resultats rammener sur 1 de la deuxieme
    // pour les agent: type, name, base, conf
    // pour les objet, la meme sans la conf
    // pour les support c soit type, name ou height + base
    // type means the same macro, but not the same name.
    // 

    vector<Robot*> rs;
    map<PlanNode* , map<string, double> > nodeMap;
    for (unsigned int i = 0; i<_tree.size(); i++)
    {
        PlanNode* PN = _tree.at(i);
        map<string,double> tmp;

        //.find("mainAgent")->second
        if (td.agents.count("mainAgent") == 1 && PN->getTask()->getCurrentDataType().agents.count("mainAgent") == 1)
        {
            Robot* a1 = global_Project->getActiveScene()->getRobotByName(td.agents.find("mainAgent")->second);
            Robot* a2 = global_Project->getActiveScene()->getRobotByName(PN->getTask()->getCurrentDataType().agents.find("mainAgent")->second);

            rs.push_back(a1);
            rs.push_back(a2);

            bool testComp = false;
            if (a1->getHriAgent()->type == a2->getHriAgent()->type)
            {
                tmp["mainAgentType"] = 0;
                testComp = true;
            }
            if (a1->getName() == a2->getName())
            {
                tmp["mainAgentName"] = 0;
                testComp = true;
            }
            if (testComp)
            {
                if (a1->getRobotStruct()->nb_dof == a2->getRobotStruct()->nb_dof)
                {
                    confPtr_t q1 = ws->getRobotConf(a1);
                    confPtr_t q2 = PN->getTaskSolution()->getStartWS()->getRobotConf(a2);

                    tmp["mainAgentDist"] = sqrt(pow((*q1)[6]-(*q2)[6],2) +  pow((*q1)[7]-(*q2)[7],2) );

                    if (a2->getRobotStruct()->nb_dof > 11)
                    {
                        double dist = 0;
                        for (int j = 12; j < a1->getRobotStruct()->nb_dof; j++)
                        {
                            dist += pow((*q1)[j]-(*q2)[j],2);
                        }
                        tmp["mainAgentConf"] = sqrt(dist);
                    }
                }
            }
            else
            {
                continue;
            }
        }




        if (td.agents.count("targetAgent") == 1 && PN->getTask()->getCurrentDataType().agents.count("targetAgent") == 1)
        {
            Robot* a1 = global_Project->getActiveScene()->getRobotByName(td.agents.find("targetAgent")->second);
            Robot* a2 = global_Project->getActiveScene()->getRobotByName(PN->getTask()->getCurrentDataType().agents.find("targetAgent")->second);

            rs.push_back(a1);
            rs.push_back(a2);

            bool testComp = false;
            if (a1->getHriAgent()->type == a2->getHriAgent()->type)
            {
                tmp["targetAgentType"] = 0;
                testComp = true;
            }
            if (a1->getName() == a2->getName())
            {
                tmp["targetAgentName"] = 0;
                testComp = true;
            }
            if (testComp)
            {
                if (a1->getRobotStruct()->nb_dof == a2->getRobotStruct()->nb_dof)
                {
                    confPtr_t q1 = ws->getRobotConf(a1);
                    confPtr_t q2 = PN->getTaskSolution()->getStartWS()->getRobotConf(a2);

                    tmp["targetAgentDist"] = sqrt(pow((*q1)[6]-(*q2)[6],2) +  pow((*q1)[7]-(*q2)[7],2) );

                    if (a2->getRobotStruct()->nb_dof > 11)
                    {
                        double dist = 0;
                        for (int j = 12; j < a1->getRobotStruct()->nb_dof; j++)
                        {
                            dist += pow((*q1)[j]-(*q2)[j],2);
                        }
                        tmp["targetAgentConf"] = sqrt(dist);
                    }
                }
            }
        }



        if (td.objects.count("supportObject") == 1 && PN->getTask()->getCurrentDataType().objects.count("supportObject") == 1)
        {
            Robot* a1 = global_Project->getActiveScene()->getRobotByName(td.objects.find("supportObject")->second);
            Robot* a2 = global_Project->getActiveScene()->getRobotByName(PN->getTask()->getCurrentDataType().objects.find("supportObject")->second);

            rs.push_back(a1);
            rs.push_back(a2);

            bool testComp = false;

            p3d_oriented_BB *OBB1 = p3d_get_oriented_BB_obj(a1->getJoint(0)->getJointStruct()->o);
            p3d_oriented_BB *OBB2 = p3d_get_oriented_BB_obj(a2->getJoint(0)->getJointStruct()->o);

            ws->load();
            double h1 = OBB1->UpBackLeft.z;

            PN->getTaskSolution()->getStartWS()->load();
            double h2 = OBB2->UpBackLeft.z;

            double heightDiff = fabs(h2 - h1);

            p3d_destroy_oriented_BB(OBB1);
            p3d_destroy_oriented_BB(OBB2);


            if (heightDiff < EPS2)
            {
                tmp["supportObjectHeight"] = 0;
                testComp = true;
            }
            if (a1->getName() == a2->getName())
            {
                tmp["supportObjectName"] = 0;
                testComp = true;
            }

            if (testComp)
            {
                if (a1->getRobotStruct()->nb_dof == a2->getRobotStruct()->nb_dof)
                {
                    confPtr_t q1 = ws->getRobotConf(a1);
                    confPtr_t q2 = PN->getTaskSolution()->getStartWS()->getRobotConf(a2);

                    double dist = 0;
                    for (int j = 0; j < a1->getRobotStruct()->nb_dof; j++)
                    {
                        dist += pow((*q1)[j]-(*q2)[j],2);
                    }
                    tmp["supportObjectConf"] = sqrt(dist);
                }
            }
        }


        //currentSupport ( probably relying on fact )


        if (td.objects.count("mainObject") == 1 && PN->getTask()->getCurrentDataType().objects.count("mainObject") == 1)
        {
            Robot* a1 = global_Project->getActiveScene()->getRobotByName(td.objects.find("mainObject")->second);
            Robot* a2 = global_Project->getActiveScene()->getRobotByName(PN->getTask()->getCurrentDataType().objects.find("mainObject")->second);

            rs.push_back(a1);
            rs.push_back(a2);

            double sizeMain;

            GeometricComputations GC = GeometricComputations();
            p3d_oriented_BB *main_OBB = p3d_get_oriented_BB_obj(a1->getJoint(0)->getJointStruct()->o);

            sizeMain = GC.distBetweenPoints(main_OBB->BottomBackLeft, main_OBB->UpBackLeft)*
                    GC.distBetweenPoints(main_OBB->BottomBackLeft, main_OBB->BottomBackRight)*
                    GC.distBetweenPoints(main_OBB->BottomBackLeft, main_OBB->BottomFrontLeft);
            p3d_destroy_oriented_BB(main_OBB);


            //compute size of BB of target
            double sizeTarget;

            p3d_oriented_BB *target_OBB = p3d_get_oriented_BB_obj(a2->getJoint(0)->getJointStruct()->o);

            sizeTarget = GC.distBetweenPoints(target_OBB->BottomBackLeft, target_OBB->UpBackLeft)*
                    GC.distBetweenPoints(target_OBB->BottomBackLeft, target_OBB->BottomBackRight)*
                    GC.distBetweenPoints(target_OBB->BottomBackLeft, target_OBB->BottomFrontLeft);

            p3d_destroy_oriented_BB(target_OBB);

            bool testComp = false;


            if (fabs(sizeTarget - sizeMain) < EPS2)
            {
                tmp["objectType"] = 0;
                testComp = true;
            }
            if (a1->getName() == a2->getName())
            {
                tmp["objectName"] = 0;
                testComp = true;
            }

            if (testComp)
            {
                if (a1->getRobotStruct()->nb_dof == a2->getRobotStruct()->nb_dof)
                {
                    confPtr_t q1 = ws->getRobotConf(a1);
                    confPtr_t q2 = PN->getTaskSolution()->getStartWS()->getRobotConf(a2);

                    double dist = 0;
                    for (int j = 0; j < a1->getRobotStruct()->nb_dof; j++)
                    {
                        dist += pow((*q1)[j]-(*q2)[j],2);
                    }
                    tmp["objectConf"] = sqrt(dist);
                }
            }
        }


        if (td.id == PN->getTask()->getTypeId())
            tmp["taskDist"] = 0;


        WorldState* ws1 = PN->getTaskSolution()->getStartWS();

        if (ws->getRobotConfs().size() != ws1->getRobotConfs().size())
            continue;
        double ds = 0;
        for (unsigned int j = 0; j< ws->getRobotConfs().size(); j++)
        {
            bool testIn = false;
            for (unsigned int k = 0; k < rs.size(); k++)
            {
                if (rs.at(k)->getName() == ws->getRobotConfs().at(j)->getRobot()->getName())
                {
                    testIn = true;
                    break;
                }
            }
            if (!testIn)
            {
                confPtr_t q;
                ds += ws->getRobotConfs().at(j)->distance(ws1->getRobotConf(ws->getRobotConfs().at(j)->getRobot()).get());
            }

        }
        tmp["wsDist"] = ds;
        //worldState


        nodeMap[PN] = tmp;
    }


    map<string, double> maxDist;
    for (unsigned int i = 0; i<_tree.size(); i++)
    {
        map<string, double> m = nodeMap[_tree.at(i)];

        for(map<string, double>::iterator it=m.begin() ; it!=m.end() ; ++it)
        {
            if (maxDist.count(it->first) != 1)
                maxDist[it->first] = it->second;
            else if (maxDist[it->first] < it->second)
                maxDist[it->first] = it->second;
        }
    }

    for (unsigned int i = 0; i<_tree.size(); i++)
    {
        map<string, double> m = nodeMap[_tree.at(i)];

        map<string,double> distances;
        distances["mainAgentName"] = 1;
        distances["mainAgentType"] = 1;
        distances["mainAgentDist"] = 1;
        distances["mainAgentConf"] = 1;

        distances["targetAgentName"] = 1;
        distances["targetAgentType"] = 1;
        distances["targetAgentDist"] = 1;
        distances["targetAgentConf"] = 1;

        distances["supportObjectName"] = 1;
        distances["supportObjectHeight"] = 1;
        distances["supportObjectConf"] = 1;

        distances["currentSupportName"] = 1;
        distances["currentSupportHeight"] = 1;
        distances["currentSupportDist"] = 1;

        distances["objectName"] = 1;
        distances["objectType"] = 1;
        distances["objectDist"] = 1;

        distances["wsDist"] = 1;

        distances["taskDist"] = 1;

        for(map<string, double>::iterator it=m.begin() ; it!=m.end() ; ++it)
        {
            distances[it->first] = it->second/(maxDist[it->first]);
        }
        nodeMap[_tree.at(i)] = distances;
    }

    PlanNode* closePN = _tree.at(0);

    double smallest = P3D_HUGE;
    for (unsigned int i = 0; i<_tree.size(); i++)
    {
        map<string, double> m = nodeMap[_tree.at(i)];

        double testA = 0;
        for(map<string, double>::iterator it=m.begin() ; it!=m.end() ; ++it)
        {
            testA += it->second;
        }
        if (testA < smallest)
        {
            closePN = _tree.at(i);
        }
    }


    return closePN;
}

void PlanTree::saveInFile(std::string path)
{

    std::ofstream file_id;
    string filename = path + "/gtpsol";
    file_id.open(filename.c_str());
    printf("Saving current plan to file %s ...\n",filename.c_str());
    Json::Value root(Json::arrayValue);
    for(unsigned int i=0;i<_tree.size();i++)
    {
        Json::Value onePN(Json::objectValue);
        PlanNode* PN = _tree.at(i);

        Json::Value curId(Json::objectValue);
        curId["taskId"] = PN->getId().first;
        curId["altId"] = PN->getId().second;
        onePN["id"] = curId;

        if (PN->getPrevious())
        {
            Json::Value prevId(Json::objectValue);

            prevId["taskId"] = PN->getPrevious()->getId().first;
            prevId["altId"] = PN->getPrevious()->getId().second;
            onePN["previous"] = prevId;
        }
        if (PN->getNexts().size()>0)
        {
            Json::Value nextsIds(Json::arrayValue);
            for (unsigned int i = 0; i < PN->getNexts().size(); i++)
            {
                Json::Value nextId(Json::objectValue);
                nextId["taskId"] = PN->getNexts().at(i)->getId().first;
                nextId["altId"] = PN->getNexts().at(i)->getId().second;
                nextsIds.append(nextId);
            }
            onePN["next"] = nextsIds;
        }

        onePN["suceeded"] = PN->hasSucceeded();

        onePN["SD"] = PN->SD->getJsonValue();

        Json::Value taskJson(Json::objectValue);
        taskJson["type"] = PN->getTask()->getType();
        taskJson["typeId"] = PN->getTask()->getTypeId();
        onePN["task"] = taskJson;


//        std::map<Robot*,std::vector<TaskSubSolution*> > solutionMap; //--------------


        if (PN->getTaskSolution())
        {
            Json::Value taskSolJson(Json::objectValue);
            Json::Value solutionMap(Json::arrayValue);
            vector<TaskSubSolution*> vTSS = PN->getTaskSolution()->getAllTSS();
            for (unsigned int j = 0; j < vTSS.size(); j++)
            {
                TaskSubSolution* TSS = vTSS.at(j);
                Json::Value oneTSS(Json::objectValue);
                oneTSS["subSolutionName"] = TSS->getSubSolutionName();
                oneTSS["id"] = TSS->getId();
                oneTSS["robot"] = TSS->getRobot()->getName();

                Json::Value prevTSS(Json::arrayValue);
                for (unsigned int k = 0; k < TSS->getPrevious().size(); k++)
                {
                    prevTSS.append(TSS->getPrevious().at(k)->getId());
                }
                oneTSS["previousTSS"] = prevTSS;

                Json::Value nextTSS(Json::arrayValue);
                for (unsigned int k = 0; k < TSS->getNext().size(); k++)
                {
                    nextTSS.append(TSS->getNext().at(k)->getId());
                }
                oneTSS["nextTSS"] = nextTSS;

                oneTSS["initRangeParam"] = TSS->getInitRangeParam();
                oneTSS["navigation"] = TSS->isNavigation();

                if (TSS->isNavigation())
                {
                    Json::Value navJson(Json::arrayValue);
                    for (unsigned int k = 0; k < TSS->getNavVector().size(); k++)
                    {
                        Json::Value onePoint(Json::arrayValue);
                        onePoint.append(TSS->getNavVector().at(k).x);
                        onePoint.append(TSS->getNavVector().at(k).y);
                        onePoint.append(TSS->getNavVector().at(k).z);
                        navJson.append(onePoint);
                    }
                    oneTSS["navigationTraj"] = navJson;
                }
                else if (TSS->getTraj())
                {
                    string trajFile = path +"/"+ TSS->getSubSolutionName() + ".traj";
                    oneTSS["trajFile"] = trajFile;
                    if (TSS->getRobot()->getHriAgent()->type == HRI_PR2)
                    {
                        p3d_multiLocalPath_enable_all_groupToPlan( TSS->getRobot()->getRobotStruct() );
                        unFixAllJoints(TSS->getRobot()->getRobotStruct());
                    }
                    p3d_save_traj(trajFile.c_str(),TSS->getTraj());
                }

                oneTSS["isLast"] = TSS->isLast();
                oneTSS["startWS"] = TSS->getStartState()->toJson();
//                cout << "startWS of " << TSS->getId() << " id = " <<  TSS->getStartState()->getId() << endl;
                oneTSS["endWS"] = TSS->getEndState()->toJson();
//                cout << "endWS of " << TSS->getId() << " id = " <<  TSS->getEndState()->getId() << endl;
                oneTSS["isSmoothed"] = TSS->isSmoothed();
                oneTSS["armId"] = TSS->getArmId();

                solutionMap.append(oneTSS);


                //            Sem::PlanningPart _planningPart; -------------------------------------------
            }
            taskSolJson["start_WS_id"] = PN->getTaskSolution()->getStartWS()->getId();
//            cout << "start_WS id = " <<  PN->getTaskSolution()->getStartWS()->getId() << endl;
            taskSolJson["end_WS_id"] = PN->getTaskSolution()->getStopWS()->getId();
//            cout << "end_WS id = " <<  PN->getTaskSolution()->getStopWS()->getId() << endl;

            taskSolJson["solutionMap"] = solutionMap;
            taskSolJson["trajComputed"] = PN->getTaskSolution()->isTrajComputed();
            taskSolJson["cost"] = PN->getTaskSolution()->getCost();

            onePN["taskSolution"] = taskSolJson;

        }


        root.append(onePN);

    }
    Json::StyledWriter styledWriter;
    file_id << styledWriter.write(root);

    file_id.close();
    printf("Done\n");
}











