#include "solutiondata.h"
#include <sstream>

INIT_MOVE3D_STATIC_LOGGER(SolutionData, "planners.GTP.GTPTools.SolutionData");

SolutionData::SolutionData()
{
    _size = 30;
    _space = "";
}

SolutionData::SolutionData(SolutionData* SD)
{
    copySolutionData(SD);
}

SolutionData::~SolutionData()
{
    for (std::map<std::string,SolutionData*>::iterator it = _SDMap.begin(); it != _SDMap.end(); ++it) {
        delete it->second;
    }
}

void SolutionData::copySolutionData(SolutionData* SD)
{
    if (SD)
    {
        _intMap = SD->getIntMap();
        _doubleMap = SD->getDoubleMap();
        _stringMap = SD->getStringMap();
        std::map<std::string,SolutionData*> s = SD->getSDMap();
        for (std::map<std::string,SolutionData*>::iterator it = s.begin(); it != s.end(); ++it) {
            _SDMap[it->first] = new SolutionData(it->second);
        }
        _space = SD->getSpace();
        _size = SD->getSize();
    }
}

const std::map<std::string, int> &SolutionData::getIntMap()
{
    return _intMap;
}

const std::map<std::string, double> &SolutionData::getDoubleMap()
{
    return _doubleMap;
}

const std::map<std::string, std::string> &SolutionData::getStringMap()
{
    return _stringMap;
}

const std::map<std::string, SolutionData*> &SolutionData::getSDMap()
{
    return _SDMap;
}

std::multimap<std::string, std::string> SolutionData::getStringAsMultimapMap()
{
    std::multimap<std::string, std::string> myMap;
    for (std::map<std::string,std::string>::iterator it=_stringMap.begin(); it!=_stringMap.end(); ++it)
        myMap.insert(std::pair<std::string,std::string>(it->first,it->second));
    return myMap;
}

void SolutionData::setMapEntry(std::string key, int val)
{
    _intMap[key] = val;
}

void SolutionData::setMapEntry(std::string key, double val)
{
    _doubleMap[key] = val;
}

void SolutionData::setMapEntry(std::string key, std::string val)
{
    _stringMap[key] = val;
}

void SolutionData::setMapEntry(std::string key, SolutionData* val)
{
    _SDMap[key] = val;
}

void SolutionData::incrementMapEntry(std::string key)
{
    if (_intMap.count(key) > 0)
        _intMap[key] += 1;
    else if (_doubleMap.count(key) > 0)
        _doubleMap[key] += 1;
    else
        _intMap[key] = 1;
}

void SolutionData::incrementMapEntry(std::string key,int val)
{
    if (_intMap.count(key) > 0)
        _intMap[key] += val;
    else if (_doubleMap.count(key) > 0)
        _doubleMap[key] += val;
    else
        _intMap[key] = val;
}

int SolutionData::getIntMapEntry(std::string key)
{
    return _intMap[key];
}

double SolutionData::getDoubleMapEntry(std::string key)
{
    return _doubleMap[key];
}

std::string SolutionData::getStringMapEntry(std::string key)
{
    return _stringMap[key];
}

SolutionData* SolutionData::getSDMapEntry(std::string key)
{
    if (_SDMap.count(key) != 1)
        _SDMap[key] = new SolutionData();

    return _SDMap[key];
}


std::string SolutionData::getHumanReadableData()
{
    int sl = _space.length();
    std::string s;
    s = "";
    for (std::map<std::string,int>::iterator it = _intMap.begin(); it != _intMap.end(); ++it) {
        s += _space + it->first;
        for (unsigned int i = 0; i < _size - it->first.length() - sl; i++)
        {
            s+=" ";
        }
        std::stringstream ss;
        ss << it->second;
        s +="= " +  ss.str() + "\n";
    }

    for (std::map<std::string,double>::iterator it = _doubleMap.begin(); it != _doubleMap.end(); ++it) {
        s += _space + it->first;
        for (unsigned int i = 0; i < _size - it->first.length() - sl; i++)
        {
            s+=" ";
        }
        std::stringstream ss;
        ss << it->second;
        s +="= " +  ss.str() + "\n";
    }

    for (std::map<std::string,std::string>::iterator it = _stringMap.begin(); it != _stringMap.end(); ++it) {
        s += _space + it->first;
        for (unsigned int i = 0; i < _size - it->first.length() - sl; i++)
        {
            s+=" ";
        }
        s +="= " + it->second + "\n";
    }

    for (std::map<std::string,SolutionData*>::iterator it = _SDMap.begin(); it != _SDMap.end(); ++it) {
        it->second->setSpace(_space+" ");
        s += _space+ it->first + ":\n";
        s += it->second->getHumanReadableData();
    }

    return s;


}

void SolutionData::print()
{
    M3D_INFO("\n"+getHumanReadableData());
}

Json::Value SolutionData::getJsonValue()
{
    Json::Value val(Json::objectValue);

    for (std::map<std::string,int>::iterator it = _intMap.begin(); it != _intMap.end(); ++it) {
        val[it->first] = it->second;
    }

    for (std::map<std::string,double>::iterator it = _doubleMap.begin(); it != _doubleMap.end(); ++it) {
        val[it->first] = it->second;
    }

    for (std::map<std::string,std::string>::iterator it = _stringMap.begin(); it != _stringMap.end(); ++it) {
        val[it->first] = it->second;
    }

    for (std::map<std::string,SolutionData*>::iterator it = _SDMap.begin(); it != _SDMap.end(); ++it) {
        val[it->first] = it->second->getJsonValue();
    }

    return val;
}

void SolutionData::loadFromJsonValue(Json::Value root)
{
    for( Json::ValueIterator itr = root.begin() ; itr != root.end() ; itr++ ) {
        Json::Value r = *itr;
        if (r.isInt())
            setMapEntry(itr.key().asString(),r.asInt());
        else if (r.isDouble())
            setMapEntry(itr.key().asString(),r.asDouble());
        else if (r.isString())
            setMapEntry(itr.key().asString(),r.asString());
        else
            getSDMapEntry(itr.key().asString())->loadFromJsonValue(r);
    }
}

std::string SolutionData::getSpace()
{
    return _space;
}

void SolutionData::setSpace(std::string s)
{
    _space = s;
}

int SolutionData::getSize()
{
    return _size;
}

void SolutionData::setSize(int s)
{
    _size = s;
}

void SolutionData::clear()
{
    _intMap.clear();
    _doubleMap.clear();
    _stringMap.clear();
    _SDMap.clear();
}






