#include "taskSubSolution.hpp"
#include "taskSolution.hpp"

#include "planner/TrajectoryOptim/Classic/smoothing.hpp"
#include "HRI_costspace/HRICS_costspace.hpp"

extern std::vector<p3d_point> currNavTraj;

TaskSubSolution::TaskSubSolution()
{

}

TaskSubSolution::TaskSubSolution(TaskSolution* sol, string name, int id, p3d_traj* traj, Robot *r, double initRP)
{
    _subSolutionName = name;
    _id = id;
    _trajectory = traj;
    _sol = sol;
    _isLast = true;
    _isSmoothed = false;
    _robot = r;
    _initRangeParam = initRP;
    _armId = -1;

    _sol->addToSolutions(this);

    _navigation = false;

}

TaskSubSolution::~TaskSubSolution()
{
    if (_trajectory)
        p3d_del_traj(_trajectory);
}

void TaskSubSolution::addToPrevious(TaskSubSolution* prev)
{
    for(unsigned int i = 0; i < _previous.size(); i++)
    {
        if(_previous.at(i)->getId() == prev->getId())
        {
            return;
        }
    }
    _previous.push_back(prev);
}

bool TaskSubSolution::setTrajInRobotStruct()
{
    if (!_robot)
        return false;

    if (!_sol->isTrajComputed())
    {
        cout << "the trajectory is not computed for the task alternetive " << endl;
    }
    else if (_trajectory)
    {
        _robot->getRobotStruct()->tcur = _trajectory;
        currNavTraj.clear();
        return true;
    }
    else if (isNavigation())
    {
        currNavTraj = _nav_path;
        _robot->getRobotStruct()->tcur = NULL;

        //compute traj from series of points
//        path_to_draw.clear();
//        for (unsigned int i = 0; i<_nav_path.size();i++)
//        {
//            Eigen::Vector2d v;
//            v[0] = _nav_path.at(i).x; v[1] = _nav_path.at(i).y;
//            path_to_draw.push_back(v);
//        }
        API::Trajectory p3d_trajectory(_robot);
        _start_WS->load();
        confPtr_t q_f(_robot->getCurrentPos());
        if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_PR2)
        {
            for (int e = 37; e<=48; e++)
                (*q_f)[e] = 0;
        }
        else if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_KUKAROBOT)
        {
            for (int e = 21; e<=26; e++)
                (*q_f)[e] = 0;
        }
        p3d_trajectory.push_back(q_f);

        for (unsigned int i = 0; i < _nav_path.size(); i++)
        {
            confPtr_t q_tmp(_robot->getCurrentPos());
//            cout << " cell = " << _nav_path.at(i).x << " , " << _nav_path.at(i).y << endl;
            (*q_tmp)[6] = _nav_path.at(i).x;
            (*q_tmp)[7] = _nav_path.at(i).y;
            (*q_tmp)[11] = _nav_path.at(i).z;
            if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_PR2)
            {
                for (int e = 37; e<=48; e++)
                    (*q_tmp)[e] = 0;
            }
            else if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_KUKAROBOT)
            {
                for (int e = 21; e<=26; e++)
                    (*q_tmp)[e] = 0;
            }
            p3d_trajectory.push_back(q_tmp);

        }
        p3d_traj* trajPt = NULL;
        _robot->getRobotStruct()->tcur = NULL;

        p3d_trajectory.replaceP3dTraj(trajPt);
        if (!_robot->getRobotStruct()->tcur)
        {
            cout << "Fail to create traj from points" << endl;
            return false;
        }
        return true;
    }
    return false;

}

double TaskSubSolution::getDuration()
{
    return 0;
}

void TaskSubSolution::setNavigation(bool nav)
{
    _navigation = nav;
}

bool TaskSubSolution::isNavigation()
{
    return _navigation;
}

void TaskSubSolution::setNavTraj(vector<p3d_point> p)
{
    _nav_path = p;
}

vector<p3d_point> TaskSubSolution::getNavVector()
{
    return _nav_path;
}


void TaskSubSolution::setNext(vector<TaskSubSolution*> next)
{
    _next = next;
    for(unsigned int i = 0; i < next.size(); i++)
    {
        next.at(i)->addToPrevious(this);
    }
    _isLast = false;

}

void TaskSubSolution::addNext(TaskSubSolution* next)
{
    _next.push_back(next);
    next->addToPrevious(this);
    _isLast = false;
}


bool TaskSubSolution::refineTraj()
{
    std::cout << "task to be smoothed" << std::endl;
    std::cout << getTask()->getDescr() << std::endl;
    if (!_robot)
        return false;
    std::cout << "call the smoothing for " << _subSolutionName << std::endl;
    if(!_isSmoothed)
    {
        API::Smoothing optimTrj(_robot,_trajectory);
        if(ENV.getBool(Env::enableHri))
            HRICS_CostFunctionCommon->prepareForSmoothing(optimTrj);
        double old_cost=optimTrj.cost();
        cout<<"Previous cost: "<<old_cost<<endl;
        optimTrj.runShortCut(PlanEnv->getInt(PlanParam::smoothMaxIterations));
        optimTrj.replaceP3dTraj();
        double cost=optimTrj.cost();
        cout<<"New cost: "<<cost<<endl;
        cout<<"Smoothing gain: "<<100*(old_cost-cost)/old_cost<<" %"<<endl;

        if(ENV.getBool(Env::enableHri))
            HRICS_CostFunctionCommon->endSmoothing(optimTrj);

    }
//    _isSmoothed = true;
    return true;
}

bool TaskSubSolution::reComputeTraj()
{
    return true;
}

p3d_traj* TaskSubSolution::getTraj()
{
//    refineTraj();
    return _trajectory;
}
Task* TaskSubSolution::getTask()
{
    return _sol->getTask();
}

string TaskSubSolution::getDescr()
{
    return "SubTask: \nName \t\t= " + _subSolutionName + "\nId \t\t= " + to_string(_id) + "\n" ;
}

void TaskSubSolution::saveToDisk(FILE *file){
    if(!file){
        printf("error: TaskSubSolution::saveToFile: file is NULL\n");
        return;
    }
    fprintf(file,"\nbeg_tss\n");
    fprintf(file,"name %s\n",getSubSolutionName().c_str());
    fprintf(file,"tssId %d\n",_id);
    fprintf(file,"armId %d\n",_armId);
    fprintf(file,"robot %s\n",_robot->getName().c_str());
    if(_start_WS)
        _start_WS->saveToDisk(file,"start");
    if(_trajectory && _trajectory->rob){
        p3d_rob *robotPt=_trajectory->rob;
        fprintf(file,"beg_traj %s\n",robotPt->name);
        p3d_localpath *lp=_trajectory->courbePt;
        while(lp){
            configPt q0=p3d_alloc_config(robotPt);
            configPt qi=lp->config_at_param(robotPt,lp,0);
            configPt qf=lp->config_at_param(robotPt,lp,lp->range_param);
            fprintf(file,"beg_localpath %s\n",p3d_get_localpath_type(lp->type_lp));
            fprintf(file,"q_init ");
            if(p3d_dist_config(robotPt,qi,q0)<EPS6)
                fprintf(file,"zero");
            else{
                for(int i=6;i<robotPt->nb_dof;i++){
                    if(fabs(qi[i])<EPS6)
                        fprintf(file,"0 ");
                    else
                        fprintf(file,"%f ",qi[i]);
                }
            }
            fprintf(file,"\nq_end ");
            if(p3d_dist_config(robotPt,qf,q0)<EPS6)
                fprintf(file,"zero");
            else{
                for(int i=6;i<robotPt->nb_dof;i++){
                    if(fabs(qf[i])<EPS6)
                        fprintf(file,"0 ");
                    else
                        fprintf(file,"%f ",qf[i]);
                }
            }
            fprintf(file,"\nend_localpath\n");
            lp=lp->next_lp;
            p3d_destroy_config(robotPt,q0);
            p3d_destroy_config(robotPt,qi);
            p3d_destroy_config(robotPt,qf);
        }
        fprintf(file,"end_traj\n");
    }
    if(_end_WS)
        _end_WS->saveToDisk(file, "end");

    fprintf(file,"end_tss\n");

}
