#ifndef TASKCONSTRAINTS_HPP
#define TASKCONSTRAINTS_HPP

#include "API/facts.hpp"
#include "GTP/GTPTools/geometricTools/geometricComputations.hpp"
#include "GTP/GTPTools/geometricTools/geometricForms-pkg.hpp"
#include "worldState.hpp"


typedef struct struct_region
{
    std::string nameId;
    GeometricForm* GeoForm;
}regionConst;

typedef struct struct_ressource
{
    std::string nameId;
    Robot* entity;
} ressourceConst;

typedef struct struct_position
{
    std::string nameId;
    Robot* entity;
    std::string objInEntity; //ex hand of PR2_ROBOT, first joint if empty
    p3d_point translations;
    p3d_point rotations;

} positionConst;

class WorldState;

class TaskConstraints
{
public:
    TaskConstraints();
    TaskConstraints(TaskConstraints* TC);

    void addConstraint(Fact f);
    void addConstraint(ressourceConst r);
    void addConstraint(regionConst r);
    void addConstraint(positionConst r);
    void addConstraint(std::string n, Robot* e);
    void addConstraint(std::string n, Robot *obj, std::string area, bool isSupport);
    void addConstraint(std::string n, GeometricForm* gf);
    void addConstraint(std::string n, Robot* e, std::string objInEntity, p3d_point t, p3d_point r );
    void addConstraint(TaskConstraints *TC);


    std::vector<Fact> getFactConstraints() {return factConstraints;}
    std::vector<ressourceConst> getRessourceConstraints() {return ressourceConstraints;}
    std::vector<regionConst> getRegionConstraints() {return regionConstraints;}
    std::vector<positionConst> getPositionConstraints() {return positionConstraints;}

    void cleanAll();

    bool areFactConstraintRespected(WorldState* WS);
    std::string getDescr();


private:
    std::vector<Fact> factConstraints;
    std::vector<ressourceConst> ressourceConstraints;
    std::vector<regionConst> regionConstraints;
    std::vector<positionConst> positionConstraints;

};

#endif // TASKCONSTRAINTS_HPP
