#include "worldState.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif



extern MightabilitiesGrid* currentMightability;

using namespace std;

int WorldState::_idCount = 0;

//########################## worldState ############################

WorldState::WorldState()
{
    _id = _idCount++;
    areFactsComputed = false;
    WSF = NULL;
    endingStateOf = NULL;
    _MG = NULL;
    _sce = NULL;
}

WorldState::WorldState(Scene* sce)
{
    _sce = sce;
    _id= _idCount++;
    areFactsComputed = false;
    WSF = NULL;
    endingStateOf = NULL;
    _TCForNextTask = NULL;
    _MG = NULL;
}

WorldState::WorldState(WorldState *WC)
{
    _id = _idCount++;
    _sce = WC->getScene();
    vector<confPtr_t > v = WC->getRobotConfs();
    for (unsigned int i = 0; i < v.size(); i++)
    {
        saveConf(v.at(i));
    }
    vector<Attachement> va = WC->getAttachements();
    for (unsigned int j = 0; j < va.size(); j++)
    {
        addAttachement(va.at(j));
    }
    areFactsComputed = false;
    WSF = NULL;
    endingStateOf = WC->getTSThatEndWithThisWS();
    endingStateOfIndex = WC->getTSIndexThatEndWithThisWS();

    if (WC->getTCForNextTask())
        _TCForNextTask= new TaskConstraints(WC->getTCForNextTask());
    else
        _TCForNextTask = NULL;



    _MG = WC->getMightabilities();
}

WorldState::WorldState(Json::Value root, Scene* sce)
{
    _sce = sce;
    fromJson(root);
}

void WorldState::load()
{

    resetCurrentAttachements();

    for(unsigned int i = 0; i < _robotsConfs.size(); i++)
    {
        loadConf(_robotsConfs.at(i));
    }

    loadAttachements();
    loadSurfaceAttachements();
}

void WorldState::loadConf(confPtr_t q)
{
    q->getRobot()->setAndUpdate(*q);
}


void WorldState::saveAll()
{
    if (!_sce)
        return;
    for (unsigned int i = 0; i < _sce->getNumberOfRobots(); i++)
    {
        saveRobot(_sce->getRobot(i));
    }
    saveCurrentAttachement();
}

void WorldState::saveRobot(Robot* rob)
{
    saveConf(rob->getCurrentPos());
}

void WorldState::saveConf(confPtr_t q)
{
    confPtr_t qq =(confPtr_t) new Configuration(*q);
    _robotsConfs.push_back(qq);
}

//return false if no occurance found
bool WorldState::replaceConf(confPtr_t q)
{
    areFactsComputed = false;
    if (q)
    {
        confPtr_t qq =(confPtr_t) new Configuration(*q);

        for (unsigned int i = 0; i < _robotsConfs.size(); i++)
        {
            if (qq->getRobot()->getName() == _robotsConfs.at(i)->getRobot()->getName())
            {
                _robotsConfs[i] = qq;
                return true;
            }
        }
    }
    return false;
}



void WorldState::clear()
{
//    for(unsigned int i = 0; i < _robotsConfs.size(); i++)
//    {
//        delete _robotsConfs.at(i);
//    }
    _robotsConfs.clear();
}

bool WorldState::isEqualTo(WorldState* WS)
{
    if (_robotsConfs.size() != WS->getRobotConfs().size())
        return false;
    for (unsigned int i = 0; i < _robotsConfs.size(); i++)
    {
        confPtr_t c1 = _robotsConfs.at(i);
        for(unsigned int j = 0; j < WS->getRobotConfs().size(); j++)
        {
            confPtr_t c2 = WS->getRobotConfs().at(j);
            if (c1->getRobot()->getName() == c2->getRobot()->getName())
            {
                if (!c1->equal(*c2, false))
                    return false;
                else
                    break;
            }
        }
    }
    return true;
}

Attachement WorldState::getAttachement(Robot* rob, Robot* obj)
{
    Attachement att;
    for(unsigned int i = 0; i < attachements_list.size(); i++ )
    {
        att = attachements_list.at(i);
        if (att.agent->getName() == rob->getName() && att.object->getName() == obj->getName())
        {
            return att;
        }
    }
    att.agent = NULL;
    att.object = NULL;
    return att;
}

void WorldState::addAttachement(Robot* rob, Robot* obj)
{
    Attachement a;
    a.agent = rob;
    a.object = obj;
    attachements_list.push_back(a);
}

void WorldState::addAttachement(Robot* rob, Robot* obj, gpGrasp gr, int armId)
{
    Attachement a;
    a.agent = rob;
    a.object = obj;
    a.grasp = gr;
    a.armId = armId;
    attachements_list.push_back(a);
}

void WorldState::addAttachement(Attachement a)
{
    attachements_list.push_back(a);
}

void WorldState::clearAttachement()
{
    attachements_list.clear();
}

void WorldState::removeAttachements(Robot* rob, Robot* obj){
    for(int i=(int)attachements_list.size()-1;i>=0;i--){
        Attachement att=attachements_list.at(i);
        if(att.agent==rob && att.object==obj){
            attachements_list.erase(attachements_list.begin()+i);
        }
    }
}

bool WorldState::hasAttachement(Robot* rob){
    for(unsigned int i=0;i<attachements_list.size();i++){
        Attachement att=attachements_list.at(i);
        if(att.agent==rob)
            return true;
    }
    return false;
}

bool WorldState::hasAttachement(Robot* rob,Robot* obj){
    for(unsigned int i=0;i<attachements_list.size();i++){
        Attachement att=attachements_list.at(i);
        if(att.agent==rob && att.object==obj)
            return true;
    }
    return false;
}

void WorldState::saveCurrentAttachement()
{
//    if (!_sce)
//        return;

//    for (unsigned int i = 0; i < _sce->getNumberOfRobots(); i++)
//    {
//        if (_sce->getRobot(i)->isAgent())
//        {
//            p3d_rob* r = _sce->getRobot(i)->getRobotStruct();
//            for (unsigned int i = 0; i < (*r->armManipulationData).size(); i++)
//            {
//                p3d_rob* o =  (*r->armManipulationData).at(i).getCarriedObject();
//                if (o)
//                {
//                    gpGrasp gr();
//                    addAttachement(r,o, gr,i);
//                }
//            }
//        }
//    }
}

void WorldState::resetCurrentAttachements()
{
    if (!_sce)
        return;

    for (unsigned int i = 0; i < _sce->getNumberOfRobots(); i++)
    {
        if (_sce->getRobot(i)->isAgent())
        {
            p3d_rob* r = _sce->getRobot(i)->getRobotStruct();
            while (r->grasped_objects)
            {
                p3d_place(r, r->grasped_objects->objectPt);
            }
        }
    }
    p3d_destroy_all_grasps();
}

void WorldState::loadAttachements()
{
    for(unsigned int i = 0; i < attachements_list.size(); i++)
    {
        Attachement a = attachements_list.at(i);
        int jointId;
        if(strstr(a.agent->getName().c_str(),"bonebreaker")){
            jointId=8;
        }
        else{
            jointId = a.agent->getHriAgent()->manip->external_hand[a.armId]->joint_id_in_rob;
        }

        if (!a.agent->isObjectAttached(a.object,jointId))
            a.agent->attachObj(a.object,jointId);
    }
}

Attachement WorldState::getCurrentAttachment() {
    if (attachements_list.size() > 0)
        return attachements_list.at(attachements_list.size() - 1);
    else
        return Attachement();
}

void WorldState::loadSurfaceAttachements()
{
    for(unsigned int i = 0; i < surfaceAttachements_list.size(); i++)
    {
        SurfaceAttachement a = surfaceAttachements_list.at(i);
        if (!a.support->isObjectAttached(a.attache,a.jointId+1))
        {
            //TODO: place by the matrix
            a.support->attachObj(a.attache,a.jointId+1);
        }
    }
}

void WorldState::clearSurfaceAttachements()
{
    surfaceAttachements_list.clear();
}

confPtr_t WorldState::getRobotConf(Robot* rob)
{
    confPtr_t q;
    for (unsigned int i = 0; i < _robotsConfs.size(); i++)
    {
        q = _robotsConfs.at(i);
        if (q->getRobot()->getName() == rob->getName())
        {
            break;
        }
    }
    return q;
}

vector<Fact> WorldState::getFacts()
{
    if (!areFactsComputed)
    {
        WSF = new WorldStateFacts(_sce);
        WSF->setConfigurations(_robotsConfs);
        areFactsComputed = true;
        _sce->getFactsManager()->computeFacts(WSF,true);
    }
    else
    {
        if(!WSF)
        {
            areFactsComputed = false;
            return getFacts();
        }
    }
    vector<Fact> vf = WSF->getMeaningfulFacts();
//    WSF->printFacts(vf);
    return vf;
}

vector<Fact> WorldState::getAllFacts()
{
    load();
    if (!areFactsComputed)
    {
        WSF = new WorldStateFacts(_sce);
        WSF->setConfigurations(_robotsConfs);
        areFactsComputed = true;
        _sce->getFactsManager()->computeFacts(WSF,true);
    }
    else
    {
        if(!WSF)
        {
            areFactsComputed = false;
            return getAllFacts();
        }
    }
    vector<Fact> vf = WSF->getFacts();
//    WSF->printFacts(vf);
    return vf;
}

Fact WorldState::getThisFact(Fact FI)
{
    load();
    return _sce->getFactsManager()->computeAndGetFact(FI.getType().factType,FI.getType().factSubType, FI.getMainEntity(), FI.getTargetEntity());
}

bool WorldState::setTSThatEndWithThisWS(TaskSolution* TS, bool forceSetting)
{
    bool ok = true;
    if (forceSetting)
    {
//        if (TS)
//            endingStateOfIndex.second = TS->getId();
//        if (TS && TS->getTask())
//            endingStateOfIndex.first = TS->getTask()->getId();

        endingStateOf = TS;
    }
    else
    {
        if (TS &&
                TS->getTask() &&
                endingStateOfIndex.first == TS->getTask()->getId() &&
                endingStateOfIndex.second == TS->getId())
            endingStateOf = TS;
        else
            ok = false;

    }
    return ok;

}

void WorldState::clearTCForNextTask()
{
    _TCForNextTask = new TaskConstraints();
}

void WorldState::initMightabilities()
{
     _MG = new MightabilitiesGrid();
//    Eigen::Vector3i size;
//    size[0] = 20;
//    size[1] = 20;
//    size[2] = 20;
//    std::vector<double> envSize;
//    envSize.push_back(0);
//    envSize.push_back(1);
//    envSize.push_back(0);
//    envSize.push_back(1);
//    envSize.push_back(0);
//    envSize.push_back(1);

//    _MG = new MightabilitiesGrid(size, envSize);
////    _MG->createAllCells();
//    _MG->setWorldState(this);

}

MightabilitiesGrid* WorldState::getMightabilities()
{
    if (!_MG)
        _MG = new MightabilitiesGrid();
    currentMightability = _MG;
    return _MG;
}

double WorldState::distanceTo(WorldState* ws)
{

    if (ws->getRobotConfs().size() != _robotsConfs.size())
        return P3D_HUGE;
    double res = 0;
    for(unsigned int i = 0; i < _robotsConfs.size(); i++)
    {
        if (ws->getRobotConfs().at(i)->getRobot()->getName() != _robotsConfs.at(i)->getRobot()->getName())
        {
            cout << "ERROR: problem with robot order!!" << endl;
            return P3D_HUGE;
        }
//        cout << "robot " <<  _robotsConfs.at(i)->getRobot()->getName() << endl;
        for(int j = 0; j < _robotsConfs.at(i)->getRobot()->getRobotStruct()->nb_dof; j++)
        {
            if (ws->getRobotConfs().at(i)->getRobot()->getHriAgent()
                    && ws->getRobotConfs().at(i)->getRobot()->getHriAgent()->type == HRI_PR2
                    && (j>36))
            {
                continue;
            }
//            if (fabs((*_robotsConfs.at(i))[j] - (*ws->getRobotConfs().at(i))[j]))
//                cout << "at " << j << " got " << fabs((*_robotsConfs.at(i))[j] - (*ws->getRobotConfs().at(i))[j]) << endl;
            res += fabs((*_robotsConfs.at(i))[j] - (*ws->getRobotConfs().at(i))[j]);
        }
    }
    return res;
}

void WorldState::updateRobotVirtualObj(Robot* r)
{
    confPtr_t q = getRobotConf(r);
    double Tx, Ty, Tz;
    double Rx, Ry, Rz;
    if (r->getHriAgent()->type == HRI_PR2)
    {


        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(r->getRobotStruct(), "rIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        (*q)[37] = Tx;
        (*q)[38] = Ty;
        (*q)[39] = Tz;
        (*q)[40] = Rx;
        (*q)[41] = Ry;
        (*q)[42] = Rz;

        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(r->getRobotStruct(), "lIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        (*q)[43] = Tx;
        (*q)[44] = Ty;
        (*q)[45] = Tz;
        (*q)[46] = Rx;
        (*q)[47] = Ry;
        (*q)[48] = Rz;
    }
    else if (r->getHriAgent()->type == HRI_KUKAROBOT)
    {
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(r->getRobotStruct(), "IkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        (*q)[21] = Tx;
        (*q)[22] = Ty;
        (*q)[23] = Tz;
        (*q)[24] = Rx;
        (*q)[25] = Ry;
        (*q)[26] = Rz;
    }
    replaceConf(q);
}

void WorldState::saveToDisk(FILE *file, string text)
{
    if(!file){
        printf("error: WorldState::saveToDisk: file is NULL\n");
        return;
    }
    fprintf(file,"beg_world_state\n");
    fprintf(file,"type %s\n",text.c_str());
    for(unsigned int i=0;i<_robotsConfs.size();i++){
        confPtr_t config=_robotsConfs.at(i);
        if(!config){
            printf("error: WorldState::saveToDisk: _robotsConfs.at(%d) is NULL\n",i);
            continue;
        }
        configPt q=config->getConfigStruct();
        if(!q){
            printf("error: WorldState::saveToDisk: config->getConfigStruct() is NULL\n");
            continue;
        }
        if(!config->getRobot()){
            printf("error: WorldState::saveToDisk: config->getRobot() is NULL\n");
            continue;
        }
        p3d_rob *robotPt=config->getRobot()->getRobotStruct();
        if(!robotPt){
            printf("error: WorldState::saveToDisk: config->getRobot()->getRobotStruct() is NULL\n");
            continue;
        }
        fprintf(file,"config %s ",robotPt->name);
        configPt q0=p3d_alloc_config(robotPt);
        if(p3d_dist_config(robotPt,q,q0)<EPS6)
            fprintf(file,"zero");
        else{
            for(int k=6;k<robotPt->nb_dof;k++){
                if(fabs(q[k])<EPS6)
                    fprintf(file,"0 ");
                else
                    fprintf(file,"%f ",q[k]);
            }
        }
        fprintf(file,"\n");
        p3d_destroy_config(robotPt,q0);
    }
    for(unsigned int i=0;i<attachements_list.size();i++){
        Attachement att=attachements_list.at(i);
        fprintf(file,"attachement %s %s %d\n",att.agent->getRobotStruct()->name
                                          ,att.object->getRobotStruct()->name
                                          ,att.armId);
    }
    fprintf(file,"end_world_state\n");
}

gpGrasp WorldState::getGraspFromId(int graspId, Robot *object, int armId, Robot* rob)
{
    bool ok = true;
    std::list<gpGrasp> gl;
    ArmManipulationData m = rob->getRobotStruct()->armManipulationData->at(armId);

    gpHand_type handType = m.getHandProperties().type;

    int te = gpGet_grasp_list(object->getName(), handType, gl);

    if (te != GP_OK)
    {
        cout << "ERROR fail to load grasps" << endl;
        ok = false;
    }

    gpGrasp gr;
    if (ok)
    {
        for (std::list<gpGrasp>::iterator igrasp = gl.begin(); igrasp != gl.end(); ++igrasp)
        {
            if (igrasp->ID == graspId)
            {
                gr = *igrasp;
                break;
            }
        }
    }
    return gr;

}

Json::Value WorldState::toJson()
{
    Json::Value root(Json::objectValue);

    Json::Value confs(Json::arrayValue);
    for (unsigned int i = 0; i<_robotsConfs.size(); i++)
    {
        confPtr_t q = _robotsConfs.at(i);

        Json::Value dofs(Json::arrayValue);
        for (unsigned int j = 0; j < q->getRobot()->getRobotStruct()->nb_dof; j++)
        {
            dofs.append((*q)[j]);
        }
        Json::Value oneConf(Json::objectValue);
        oneConf["robot"] = q->getRobot()->getName();
        oneConf["dofs"] = dofs;
        confs.append(oneConf);

    }
    root["confs"] = confs;

    Json::Value attachements(Json::arrayValue);
    for (unsigned int i = 0; i < attachements_list.size(); i++)
    {
        Attachement a = attachements_list.at(i);
        Json::Value oneAttachement(Json::objectValue);
        oneAttachement["agent"] = a.agent->getName();
        oneAttachement["object"] = a.object->getName();
        oneAttachement["armId"] = a.armId;
        oneAttachement["graspId"] = a.grasp.ID;
        attachements.append(oneAttachement);
    }
    root["attachements"] = attachements;

    Json::Value surfaceAttachements(Json::arrayValue);
    for (unsigned int i = 0; i < surfaceAttachements_list.size(); i++)
    {
        SurfaceAttachement a = surfaceAttachements_list.at(i);
        Json::Value oneAttachement(Json::objectValue);
        oneAttachement["support"] = a.support->getName();
        oneAttachement["attache"] = a.attache->getName();
        oneAttachement["jointId"] = a.jointId;

        Json::Value matrix(Json::objectValue);
        Json::Value matrixLine(Json::arrayValue);
        for (int j = 0; j < 4;j++)
        {
            Json::Value matrixColumn(Json::arrayValue);
            for (int k = 0; k < 4; k++)
            {
                matrixColumn.append(a.transform[j][k]);
            }
            matrixLine.append(matrixColumn);
        }
        matrix["transform"] = matrix;
    }
    root["surfaceAttachements"] = surfaceAttachements;

    root["id"] = _id;

    if (endingStateOf)
    {
        Json::Value endingStateOfJson(Json::objectValue);
        endingStateOfJson["taskId"] = endingStateOf->getTask()->getId();
        endingStateOfJson["altId"] = endingStateOf->getId();
        root["endingStateOf"] = endingStateOfJson;
    }
    return root;
}

void WorldState::fromJson(Json::Value root)
{
    if (!_sce)
        return;
    _id = root["id"].asInt();
    _robotsConfs.clear();

    for( Json::ValueIterator itr = root["confs"].begin() ; itr != root["confs"].end() ; itr++ ) {
        Json::Value oneConf = (*itr);
        Robot* r = _sce->getRobotByName(oneConf["robot"].asString());
        confPtr_t q = (confPtr_t)new Configuration(r);
        int i = 0;
        for( Json::ValueIterator itr = oneConf["dofs"].begin() ; itr != oneConf["dofs"].end() ; itr++ ) {
            (*q)[i++] = (*itr).asDouble();
        }
        _robotsConfs.push_back(q);

    }

    for( Json::ValueIterator itr = root["attachements"].begin() ; itr != root["attachements"].end() ; itr++ ) {
        Json::Value oneAttachement = (*itr);
        Robot* r = _sce->getRobotByName(oneAttachement["agent"].asString());
        Robot* o = _sce->getRobotByName(oneAttachement["object"].asString());

        int armId = oneAttachement["armId"].asInt();
        int graspId = oneAttachement["graspId"].asInt();
        gpGrasp gr = getGraspFromId(graspId,o,armId,r);

        addAttachement(r,o,gr,armId);
    }


    for( Json::ValueIterator itr = root["surfaceAttachements"].begin() ; itr != root["surfaceAttachements"].end() ; itr++ ) {
        Json::Value oneAttachement = (*itr);

        SurfaceAttachement sa;
        sa.attache = _sce->getRobotByName(oneAttachement["attache"].asString());
        sa.support = _sce->getRobotByName(oneAttachement["support"].asString());
        sa.jointId = oneAttachement["jointId"].asInt();

        Json::Value matrix = oneAttachement["transform"];

        for (int j = 0; j < 4;j++)
            for (int k = 0; k < 4; k++)
                sa.transform[j][k] = matrix[j][k].asDouble();

        surfaceAttachements_list.push_back(sa);
    }

    if (root.isMember("endingStateOf"))
    {
        endingStateOfIndex.first = root["endingStateOf"]["taskId"].asInt();
        endingStateOfIndex.second = root["endingStateOf"]["altId"].asInt();
    }
}
