#include "moveTo.hpp"
#include "API/project.hpp"

#include "planner/plannerFunctions.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include <jsoncpp/json/json.h>

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif


#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>
#include "boost/lexical_cast.hpp"
using namespace std;


MoveTo::MoveTo()
{
    type = "moveto";
    typeId = moveTo;
    _configurationName = "";
    nbMaxLoops = 3;


    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", nbMaxLoops);

    _possibleMainAgentType.push_back(HRI_PR2);
    _possibleMainAgentType.push_back(HRI_HERAKLES);
    _possibleMainAgentType.push_back(HRI_KUKAROBOT);
}

bool MoveTo::initAll()
{
    _traj = NULL;
    result = new TaskAllSolutions(this);
    _armId =-1;
    _returnReportcount = 0;
    _useConf = NOTDEFINED;
    _configurationName = "";
    mainAgent = NULL;
    armUsed = -1;
    return true;
}

bool MoveTo::initialize()
{
    _traj = NULL;

    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->getSDMapEntry("times")->setMapEntry("timeMp", 0.0);
    SD->setMapEntry("returnReport", "ERROR");

    _returnReportcount = 0;
    _returnReportcountMax = 0;

    _fileInfo.clear();

    //remove col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_deactivate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }

    return  initMotionPlaner(mainAgent);
}

bool MoveTo::finalize()
{
    //put back col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_activate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }

    ENV.setBool(Env::drawTraj,false);
    return true;
}


//TODO:
// use a 3D/6D point
//
bool MoveTo::run(multimap<string,string> agents_name,
                 multimap<string,string> objects_name,
                 multimap<string,p3d_point> point3d,
                 multimap<string,string> additionalData,
                 WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!isPossibleMainAgent(mainAgent))
    {
        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;
    if (_useConf == CARTESIANNDPOINT)
        nbMaxLoops = 100;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnReportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;

        res = true;
    }

    if (res)
        SD->setMapEntry("returnReport","OK");

    computeCost(0);
    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}


TaskSolution* MoveTo::findAlternative(bool computeMP)
{
    int count = 0;
    if (!getResult()->getAlternativeWithId(0))
    {
        cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
        return NULL;
    }
    getResult()->getAlternativeWithId(0)->getStartWS()->load(); //

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    int id = -1;
    bool res = false;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnReportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
           continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        id = result->getNextId();

        if(!setToSolution(id,computeMP))
           continue;
        if(!testConstraints(id))
            continue;
        res = true;
    }
    if (res)
        SD->setMapEntry("returnReport","OK");

    computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (res && id != -1)
        return getResult()->getAlternativeWithId(id);
    return NULL;
}



bool MoveTo::findCandidateSolutions()
{
    if (_useConf == SAVEDCONF)
//    if (_useConf != 0 && _configurationName != "")
    {
        //Read file
//        if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
//            return true;
        _fileInfo = getConfiguration(_configurationName);
        if (_fileInfo.size() == 0)
        {
            cout << "Notthing in the file for this configuration : " << _configurationName << endl;
            SD->setMapEntry("returnReport","no_conf_in_file");
            return false;
        }
    }
    return true;
}

bool MoveTo::getRandomSol()
{
    return true;
}

bool MoveTo::findConfigurations()
{
    if (!mainAgent->getHriAgent())
        return false;


    if (_useConf == CARTESIAN6DPOINT || _useConf == CARTESIANNDPOINT)
        _targetConf = computeConfFromPoint(mainAgent,_givenSol,armUsed);
    else if (_useConf == GIVENCONF || _useConf == SAVEDCONF)
    {
        map<int,vector<double> > m;
        int rArm = -1;
        int lArm = -1;
        int head = -1;
        int nbArmDof = -1;
        if (mainAgent->getHriAgent()->type == HRI_HERAKLES)
        {
            rArm = 18;
            lArm = 27;
            head = 15;
            nbArmDof = 7;
        }
        else if (mainAgent->getHriAgent()->type == HRI_PR2 )
        {
            rArm = 16;
            lArm = 25;
            head = 13;
            nbArmDof = 7;
        }
        else if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
        {
            rArm = 12;
            nbArmDof = 7;
        }

        if (_useConf == GIVENCONF)
        {
            int c;
            _returnReportcount++;
            if (_givenSol.size() != nbArmDof)
            {
                if (_returnReportcountMax < _returnReportcount)
                {
                    _returnReportcountMax = _returnReportcount;
                    SD->setMapEntry("returnReport","nb_of_dof_unmatched");
                    return false;
                }
            }
            if (armUsed == 0)
                c = rArm;
            else if (armUsed == 1)
                c = lArm;

            m[c] = _givenSol;
        }
        else if (_useConf == SAVEDCONF)
        {
            for (std::map<std::string,std::vector<double> >::iterator it=_fileInfo.begin(); it!=_fileInfo.end(); ++it)
            {
                if (it->first == "RARM" || it->first == "LARM")
                {
                    _returnReportcount++;
                    if (it->second.size() != nbArmDof)
                    {
                        if (_returnReportcountMax < _returnReportcount)
                        {
                            _returnReportcountMax = _returnReportcount;
                            SD->setMapEntry("returnReport","nb_of_dof_unmatched");
                            return false;
                        }
                    }
                }
                if (it->first == "RARM")
                {
                    m[rArm] = it->second;
                    if (armUsed == 1)
                        armUsed = 2;
                    else
                        armUsed = 0;
                }
                else if (it->first == "LARM")
                {
                    m[lArm] = it->second;
                    if (armUsed == 0)
                        armUsed = 2;
                    else
                        armUsed = 1;
                }
                else if (it->first == "HEAD")
                {
                    m[head] = it->second;
                    armUsed = 3;
                }
            }
        }
        _targetConf = getConf(mainAgent, m);
    }

    if (_targetConf)
        return true;
    else
        return false;
}


bool MoveTo::findTrajectories()
{

    _WS->load();
    p3d_destroy_all_grasps();
    vector<SurfaceAttachement> sss;
    for(unsigned int i = 0; i < _WS->getAttachements().size(); i ++)
    {
        Attachement aa = _WS->getAttachements().at(i);
        PrivateTask::attachObject(aa.agent,aa.object,true);
        vector<SurfaceAttachement> tmp = createAndLoadSurfaceAttachements(aa.object);
        sss.insert(sss.end(), tmp.begin(), tmp.end());
    }
    int status = -1;


    mainAgent->setAndUpdate(*_targetConf);
    g3d_draw_allwin_active();
//    sleep(10);

    bool useGraph = false;
    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
    {
        pqp_deactivate_robot_collision(mainAgent->getRobotStruct());
        _traj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent), _targetConf, NULL, status);
        pqp_activate_robot_collision(mainAgent->getRobotStruct());
    }
    else if (armUsed == 2)
        _traj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent), _targetConf, NULL, status);
    else
    {
        _traj = compute_traj(_WS->getRobotConf(mainAgent), _targetConf, NULL, status);
        useGraph =true;
    }


    _returnReportcount++;
    if (!_traj)
    {
        cout << "no traj" << endl;
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","not_traj_found");
        }
        return false;
    }


//    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
//        return true;

    if (useGraph)
    {
        setGraph(new Graph(*API_activeGraph));
        SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",(int)getGraph()->getNodes().size());
        SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",_traj->nlp+1);
        SD->getSDMapEntry("graph")->setMapEntry("trajSize",_traj->range_param);
    }
    removeSurfaceAttachements(sss);
    return true;
}

bool MoveTo::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }
    return computeSolutionTrajectories(TS);

}

bool MoveTo::computeSolutionTrajectories(TaskSolution *TS)
{
    bool res=true;
    if(!TS)
    {
        res = false;
    }

    if(res){
        _WS->load();
        confPtr_t q = TS->getStopWS()->getRobotConf(mainAgent);
        _targetConf = q;
        res = findTrajectories();
    }

    if(res){
        TS->getSubSolutionWithId(0)->setTraj(_traj);
    }

    return res;
}

bool MoveTo::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = createSolution(computeMP);
    if(TS){
        TS->setId(solutionId);
        TS->setTaskAllSolutions(result);
        getResult()->addAlternativeSol(TS);
    }
    return (TS != NULL);
}

TaskSolution *MoveTo::createSolution(bool computeMP)
{

    _WS->load();
    TaskSolution* TS = new TaskSolution();

    TS->setTrajComputed(computeMP);
    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "move", 0, _traj, mainAgent, 0);

    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);

    vector<SurfaceAttachement> sss;
    for(unsigned int i = 0; i < _WS->getAttachements().size(); i ++)
    {
        Attachement aa = _WS->getAttachements().at(i);
        WS1->addAttachement(aa.agent,aa.object,aa.grasp,aa.armId);
        vector<SurfaceAttachement> tmp = createAndLoadSurfaceAttachements(aa.object);
        sss.insert(sss.end(), tmp.begin(), tmp.end());
    }

    _WS->setSurfaceAttachements(sss);

    bool t = WS1->replaceConf(_targetConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    vector<Attachement> atts;
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        if (att.agent == mainAgent)
        {
            PrivateTask::attachObject(mainAgent, att.object, true ,att.armId);
            atts.push_back(att);
        }
    }
    mainAgent->setAndUpdate(*_targetConf);
    for (unsigned int i = 0; i < atts.size(); i++)
        WS1->replaceConf(atts.at(i).object->getCurrentPos());

    for (unsigned int i = 0; i<sss.size(); i++)
        WS1->replaceConf(sss.at(i).attache->getCurrentPos());

    TSS1->setEndState(WS1);

    WS1->setTCForNextTask(_currConstraints);

    TS->setStopWS(WS1);
    WS1->setTSThatEndWithThisWS(TS);

    return TS;
}

bool MoveTo::setAgents(multimap<string,string> agents_name)
{
    setMainAgent(agents_name);
    if (!mainAgent || !mainAgent->getHriAgent())
    {
        SD->setMapEntry("returnReport","no_mainAgent");
       return false;
    }
    if (mainAgent->getHriAgent()->type != HRI_PR2
            && mainAgent->getHriAgent()->type != HRI_HERAKLES
            && mainAgent->getHriAgent()->type != HRI_KUKAROBOT)
    {
       cout << "this function is made to work with PR2, herakles and kuka only" << endl;
       SD->setMapEntry("returnReport","not_pr2_nor_herakles_nor_kuka_used");
       return false;
    }
    return true;

}

bool MoveTo::setObjects(multimap<string,string> objects_name)
{
    return true;
}

bool MoveTo::setPoints(multimap<string,p3d_point> points)
{
    _givenSol.clear();
    bool ok = true;
//    if (_useConf)
//    {
        if (points.count("target1") == 1 ||
                points.count("target2") == 1 ||
                points.count("target3") == 1 )
        {
            if (points.count("target1") == 1 &&
                    points.count("target2") == 1 &&
                    points.count("target3") == 1)
            {
                _givenSol.push_back(points.find("target1")->second.x*M_PI/180);
                _givenSol.push_back(points.find("target1")->second.y*M_PI/180);
                _givenSol.push_back(points.find("target1")->second.z*M_PI/180);
                _givenSol.push_back(points.find("target2")->second.x*M_PI/180);
                _givenSol.push_back(points.find("target2")->second.y*M_PI/180);
                _givenSol.push_back(points.find("target2")->second.z*M_PI/180);
                _givenSol.push_back(points.find("target3")->second.x*M_PI/180);
                _useConf = GIVENCONF;
            }
            else
                ok = false;
        }
        else if (points.count("cartesianTarget") == 1 || points.count("cartesianTarget1") == 1)
        {
            _useConf = CARTESIANNDPOINT;
            string str;
            if (points.count("cartesianTarget") == 1)
                str = "cartesianTarget";
            else
                str = "cartesianTarget1";
            _givenSol.push_back(points.find(str)->second.x);
            _givenSol.push_back(points.find(str)->second.y);
            _givenSol.push_back(points.find(str)->second.z);
            if (points.count("cartesianTarget2") == 1)
            {
                p3d_point a = points.find("cartesianTarget2")->second;
                _givenSol.push_back(a.x);
                _givenSol.push_back(a.y);
                _givenSol.push_back(a.z);
                if (a.x != 1000 && a.y != 1000 && a.z != 1000)
                    _useConf = CARTESIAN6DPOINT;
            }
        }
        else
        {
            // in this case, neither a point nor a conf has beed defined, we can suppose we will have a configuration later.
            _useConf = SAVEDCONF;
        }
//    }
    return ok;
}

bool MoveTo::setdata(multimap<string,string> additionalData)
{

    if (additionalData.count("confName") == 1 && _useConf == SAVEDCONF)
    {
        _configurationName = additionalData.find("confName")->second;
        if (_configurationName == "")
        {
            SD->setMapEntry("returnReport","configurationName_empty");
            return false;
        }
        return true;
    }
    else if (additionalData.count("hand") == 1 &&
             (_useConf == GIVENCONF ||
              _useConf == CARTESIAN6DPOINT ||
              _useConf == CARTESIANNDPOINT) )
    {
        if (additionalData.find("hand")->second == "right")
            armUsed = 0;
        else if (additionalData.find("hand")->second == "left")
            armUsed = 1;
        else
        {
            SD->setMapEntry("returnReport","wrong_hand");
            return false;
        }
        cout << "armUsed " << armUsed << endl;

        return true;
    }
    else
    {
        cout << "ERROR: A wrong inputs have been given" << endl;
        SD->setMapEntry("returnReport","wrong_inputs");
        return false;
    }
}

void MoveTo::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "move")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
            _WS->load();
            tss->refineTraj();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}


string MoveTo::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    return s;
}

string MoveTo::getTextFromValues()
{
    return "Task: " + type + " : move the right hand and/or the left hand to confName, defined in the file or by setting points as a conf or as a target"
            + "\n    agents: "
            + "\n        mainAgent: [AGENT ID]"
            + "\n    datas: "
            + "\n        confName: [CONFIGURATION NAME] #conflicting with setting points"
            + "\n        hand: [right or left] #only when setting points"
            + "\n    points: "
            + "\n        target1: x = q1, y = q2, z = q3"
            + "\n        target2: x = q4, y = q5, z = q6"
            + "\n        target3: x = q7 # this is to set up q conf, another possibility is to set up a cartesian goal"
            + "\n        cartesianTarget: x y z # of the point to reach by the hand defined in data"
            + "\n        cartesianTarget2: rx ry rz # to define the rotations. OPTIONAL."
            + "\n                                     in case of partially defined, use 1000 to set the undefined rotations";
}

map<string, vector<double> > MoveTo::getConfiguration(string configurationName)
{
    string path = string(getenv("HOME_MOVE3D")) + "/Configurations/moveTo/" + mainAgent->getName();

    std::ifstream ifs;
    Json::Value root;
    bool ok=true;

    ifs.open(path.c_str());

    if(!ifs){
        ok=false;
        cout << "could not open "<< path << endl;
        SD->setMapEntry("returnReport","file_does_not_exist");
    }
    if(ok){
        try{
            ifs>>root;
        }catch (std::runtime_error e){
            cout << "error while parsing file "<<path<<": "<<e.what() << endl;
            ok=false;
            SD->setMapEntry("returnReport","parsing_error1");
        }catch(std::exception e){
            cout << "error while parsing file "<<path<<": "<<e.what() << endl;
            ok=false;
            SD->setMapEntry("returnReport","parsing_error2");
        }
    }
    if(ok){
        if(!root){
            cout << "could not parse json file "<<path << endl;
            ok=false;
            SD->setMapEntry("returnReport","bad_json_format");
        }
    }

    map<string, vector<double> > confParts;
    if (ok && root.isMember(configurationName))
    {
        Json::Value v = root[configurationName];
//        cout << v.toStyledString() << endl;
        vector<string> membersName = v.getMemberNames();
        for (unsigned int i = 0; i < membersName.size(); i++)
        {
            vector<double> vectD;
            for(Json::Value::iterator it = v[membersName.at(i)].begin(); it !=v[membersName.at(i)].end(); ++it)
            {
                vectD.push_back((*it).asDouble());
            }
            confParts[membersName.at(i)] = vectD;
        }
    }
    return confParts;

}

confPtr_t MoveTo::getConf(Robot* r, std::map<int, std::vector<double> >  toChange)
{
    confPtr_t q = (confPtr_t) new Configuration(*r->getCurrentPos());

    for (std::map<int, std::vector<double> >::iterator it=toChange.begin(); it!=toChange.end(); ++it)
    {
        for (unsigned int i = it->first; i < it->first + it->second.size(); i++)
        {
            (*q)[i] = it->second.at(i - it->first);
        }
    }
    return q;
}

confPtr_t MoveTo::computeConfFromPoint(Robot *r, vector<double> v, int armId)
{
    confPtr_t q;
    bool ok = true;
    if (v.size() < 3)
        ok = false;
    if (ok)
    {
        double x,y,z,tx,ty,tz;
        x = v.at(0);
        y = v.at(1);
        z = v.at(2);
        tx = p3d_random(-M_PI, M_PI);
        ty = p3d_random(-M_PI, M_PI);
        tz = p3d_random(-M_PI, M_PI);
        if (v.size() == 6)
        {
            if (v.at(3) != 1000)
                tx = v.at(3)*M_PI/180;
            if (v.at(4) != 1000)
                ty = v.at(4)*M_PI/180;
            if (v.at(5) != 1000)
                tz = v.at(5)*M_PI/180;
        }
        p3d_matrix4 m;
        p3d_mat4Pos(m,x,y,z,tx,ty,tz);
        ok = computeInverseKinematic(r,m,q,armId,true);
    }
    return q;
}



