#include "point.hpp"
#include "API/project.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

using namespace std;

//####################### Point ########################

Point::Point() : PrivateTask()
{
    type = "point";
    typeId = point;
    _armId = -1;

    nbMaxLoops = 50;

    //_possibleMainAgent.push_back("PR2_ROBOT");
    //_possibleMainAgent.push_back("MAX_ROBOT");
    //_possibleMainAgent.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);

    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 1);
    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);

    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", nbMaxLoops);
}

bool Point::initAll()
{
    result = new TaskAllSolutions(this);
    _targetPoint.x=-1000;
    _targetPoint.y=-1000;
    _targetPoint.z=-1000;

    cX = 0;
    cY = 0;
    cZ = 0;
    cRx = 0;
    cRy = 0;
    cRz = 0;

    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);

    _nbTests = 0;
    _possibleMainAgentType.push_back(HRI_PR2);

    return true;
}

bool Point::run(multimap<string,string> agents_name,
                           multimap<string,string> objects_name,
                           multimap<string,p3d_point> point3d,
                           multimap<string,string> additionalData,
                           WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if(!setObjects(objects_name))
    {
        if (!setPoints(point3d))
        {
            cout << "ERROR: no target specified for task " << type << " with id " << task_id << endl;
            return false;
        }
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: no hand specified " << type << " with id " << task_id << endl;
        return false;
    }

//    if (!findCandidateSolutions())
//    {
//        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
    bool res = false;
    int count = 0;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        if(count++ > 50)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        res = true;
    }
    if (res)
        SD->setMapEntry("returnReport","OK");

    return res;
}

bool Point::findCandidateSolutions()
{
    return true;
}

bool Point::getRandomSol()
{
    if (_nbTests == 0)
    {
        //Find a point for the hand depending on the distance to the target
        int shoulderId = mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()];
        p3d_jnt* j = mainAgent->getJoint(shoulderId)->getJointStruct();

        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(j->abs_pos, &Tx, &Ty, &Tz,
                                       &Rx, &Ry, &Rz);
        _mainPoint.x = Tx;
        _mainPoint.y = Ty;
        _mainPoint.z = Tz;

        //distance shoulder-target
        _dist = sqrt(pow(Tx - _targetPoint.x ,2) + pow(Ty - _targetPoint.y ,2) + pow(Tz - _targetPoint.z ,2));
        _armReach = mainAgent->getHriAgent()->manip->reach_tolerance -0.15;

        if (_dist > _armReach)
            isNear = false;
        else
        {
            isNear = true;
            isStraight = true;
        }
        dist_to_move = _armReach * 0.05;
    }


    if (!isNear)
    {
        double current_dist = _armReach - dist_to_move*_nbTests;
        // arm extended
        if (current_dist < _armReach*0.3)
        {
            cout << "dist is too near, fail " << endl;
            //pour le moment, pas besoin d'aller plus loin
            return false;
        }
        cout << "dist = " << _dist << " current_dist = " << current_dist  << "_main point = " << _mainPoint.x << ", " << _mainPoint.y << ", " << _mainPoint.z << endl;
        cX = ((_dist-current_dist)*_mainPoint.x + current_dist*_targetPoint.x)/_dist;
        cY = ((_dist-current_dist)*_mainPoint.y + current_dist*_targetPoint.y)/_dist;
        cZ = ((_dist-current_dist)*_mainPoint.z + current_dist*_targetPoint.z)/_dist;
        cRx = 0;
        cRy = -M_PI/2;
        cRz = 0;


        p3d_vector3 ux;
        ux[0] = (_targetPoint.x - _mainPoint.x) /_dist;
        ux[1] = (_targetPoint.y - _mainPoint.y) /_dist;
        ux[2] = (_targetPoint.z - _mainPoint.z) /_dist;

        p3d_vector3 uy;
        uy[0] = - ux[1];
        uy[1] = ux[0];
        uy[2] = 0;

        p3d_vector3 uz;
        uz[0] = -ux[2]*uy[1];
        uz[1] = -ux[2]*uy[0];
        uz[2] = ux[0]*uy[1] - ux[1]*uy[0];

        p3d_matrix4 m = {{ux[0],uy[0],uz[0],cX},{ux[1],uy[1],uz[1],cY},{ux[2],uy[2],uz[2],cZ},{0,0,0,1}};




        p3d_matrix4 rotPI2 = {{0,0,1,0},{0,1,0,0},{-1,0,0,0},{0,0,0,1}};

        p3d_matrix4 a;
        p3d_mat4Mult(m,rotPI2,a);


        p3d_mat4Copy(a,_m);
    }
    else
    {
        confPtr_t q = _targetObject->getCurrentPos();
        p3d_matrix4 a = {{-1,0,0,(*q)[6]},{0,1,0,(*q)[7]},{0,0,-1,(*q)[8]+0.1},{0,0,0,1}};
        p3d_mat4Print(a, "nanana");
        p3d_mat4Copy(a,_m);
        // if first time test at 10cm from object
        // if fail tries on top of the object
        // the random position at 10 cm from the object
    }






//            double reach_tolerance;
//            int* shoulder_joint_ids;
//            int shoulder_joint_ids_no;

    _nbTests++;
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
    return true;
}

bool Point::findConfigurations()
{

    computeInverseKinematic(mainAgent,_m,_sol,0,true);
    if(_sol)
    {
//        mainAgent->setAndUpdate(*_sol);
        return true;
    }
    return false;
}


bool Point::findTrajectories()
{
    int status = -1;
    _traj = compute_traj(_WS->getRobotConf(mainAgent),_sol,NULL,status);

    _returnReportcount++;
    if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)
    {
        cout << "no traj" << endl;
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            if (status == PLAN_STATUS_START_INVALID)
                SD->setMapEntry("returnReport","approach_traj_invalid_start");
            else if (status == PLAN_STATUS_NOT_FOUND)
                SD->setMapEntry("returnReport","approach_traj_not_found");
            else if (status == PLAN_STATUS_GOAL_INVALID)
                SD->setMapEntry("returnReport","approach_traj_invalid_stop");
        }
        return false;
    }

    setGraph(new Graph(*API_activeGraph));
    SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",(int)getGraph()->getNodes().size());
    SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",_traj->nlp+1);
    SD->getSDMapEntry("graph")->setMapEntry("trajSize",_traj->range_param);

    return true;
}

bool Point::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    cout << getDescr() << endl;
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }

    TS->getStartWS()->load();
    TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);

    confPtr_t q1 = TSS0->getEndState()->getRobotConf(mainAgent);

    _sol = q1;

    findTrajectories();

    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(0)->setTraj(_traj);


    TS->setTrajComputed(true);

    return true;
}

bool Point::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, _traj, mainAgent, 0);

    TSS1->setArmId(_armId);


    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(_sol);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }

    TSS1->setEndState(WS1);

    TS->setStopWS(WS1);


    TS->addConf(_WS->getRobotConf(mainAgent));
    TS->addConf(_sol);
    return true;
}



bool Point::setAgents(multimap<string,string> agents_name)
{
    if (agents_name.count("mainAgent") != 1)
    {
        cout << "ERROR: no main agent for " << type << " task." << endl;
        return false;
    }
    mainAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("mainAgent")->second);
    if (!mainAgent)
        return false;
    return true;
}

bool Point::setObjects(multimap<string,string> objects_name)
{
    if (objects_name.count("target") != 1)
    {
        cout << "one and only one target at a time" << endl;
        return false;
    }
    _targetObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("target")->second);
    if (!_targetObject)
        return false;

    GeometricComputations GC = GeometricComputations();
    _targetPoint = GC.getCenterOfObj(_targetObject);
    cout << "target = " << _targetPoint.x << " , " <<  _targetPoint.y << " , " <<  _targetPoint.z << endl;
    return true;
}

bool Point::setPoints(multimap<string,p3d_point> points)
{
    if (points.count("target") != 1)
    {
        cout << "one and only one target at a time" << endl;
        return false;
    }
    _targetPoint = points.find("target")->second;
    return true;
}

bool Point::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
        cout << "no arm choosen, go for right" << endl;
        setArmToUse(Right);
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        cout << "setting default arm to right" << endl;
        setArmToUse(Right);
    }

    return true;
}

string Point::getDescr()
{
    return "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type+ "\nMain agent \t\t= " + mainAgent->getName() + "\ntarget \t= is to be specified";
}


bool Point::setArmToUse(armLabel arm_t)
{
    m_arm = arm_t;
    //define, from the arm string and the type of main robot the hand type:
    // need to be changed!!!!!!!!!!!!!! should be stocked in agent.h
    if (mainAgent->getName().find("PR2") !=std::string::npos ||
            mainAgent->getName().find("MAX") !=std::string::npos ||
            mainAgent->getName().find("BOB") !=std::string::npos)
    {
        if (arm_t == Left)
            handType = GP_PR2_GRIPPER_LEFT;
        else
            handType = GP_PR2_GRIPPER; //by default right hand
        return true;
    }
    else if (mainAgent->getName().find("HUMAN") !=std::string::npos)
    {
        if (arm_t == Left)
            handType = GP_SAHAND_LEFT;
        else
            handType = GP_SAHAND_RIGHT; //by default right hand
        return true;
    }
    return false;

}



int Point::getArmId()
{
    if (m_arm == Left)
        return 1;
    else if (m_arm == Right)
        return 0;
    else
    {
        cout << "arm not defined as it should be, by default choosing id 0, string defined = " << m_arm << endl;
        return 0;
    }
}

string Point::getTextFromValues()
{
    return "Task: " + type + " this task is not implemented yet!";
}
