#include "give.hpp"
#include "API/project.hpp"
#include "GTPTools/worldState.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

//int maxTests = 100;


using namespace std;

//####################### Give ########################
Give::Give() : CooperativeTask()
{
    type = "give";
    typeId = give;
    _armId = -1;

    nbMaxLoops = 50;
    maxTests = 100;

    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 7);
    SD->setMapEntry("maxTests", maxTests);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", maxTests*nbMaxLoops);

    _hasTargedPoint = false;

    _possibleMainAgentType.push_back(HRI_PR2);
}

bool Give::initAll()
{

    result = new TaskAllSolutions(this);
    _armId = 0;
    mainObject = NULL;
    mainAgent = NULL;
    return true;
}

bool Give::initialize()
{
    approachTraj = NULL;

    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->setMapEntry("returnReport", "ERROR");
//    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", maxTests*nbMaxLoops);

    _currTargetHandId = -1;

    _returnReportcount = 0;
    _returnReportcountMax = 0;

    return initMotionPlaner(mainAgent);
}

bool Give::finalize()
{
    activateFingertipCol(true, mainAgent, mainObject, NULL);
    return true;
}

bool Give::findCandidateSolutions()
{
    if (!checkAndSetArmAndObj())
        return false;

    if (!checkAttachement())
    {
        SD->setMapEntry("returnReport", "no_corresponding_attachement_in_ws");
        return false;
    }
    if (!checkHRIAgent(mainAgent))
        return false;

    if (!checkHRIAgent(targetAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    GeometricComputations GC = GeometricComputations();

    _mainAgentShoulder.x = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[0][3];
    _mainAgentShoulder.y = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[1][3];
    _mainAgentShoulder.z = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[2][3];

    _mainAgentReach = mainAgent->getHriAgent()->manip->reach_tolerance;
    _targetAgentReach = targetAgent->getHriAgent()->manip->reach_tolerance;


    for (int i = 0; i < targetAgent->getHriAgent()->manip->shoulder_joint_ids_no; i++)
    {
        p3d_point targetShoulder;

        targetShoulder.x = targetAgent->getJoint(targetAgent->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[0][3];
        targetShoulder.y = targetAgent->getJoint(targetAgent->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[1][3];
        targetShoulder.z = targetAgent->getJoint(targetAgent->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[2][3];

        double d = GC.distBetweenPoints(_mainAgentShoulder, targetShoulder);

        cout << "dist = " << d << " Main reach = " << _mainAgentReach << " target reach = " << _targetAgentReach << endl;

        if (d < _mainAgentReach + _targetAgentReach)
        {
            pair<int, p3d_point> p;
            p.first = i;
            p.second = targetShoulder;
            _targetShoulder.push_back(p);
        }
    }

    if (_targetShoulder.size() == 0)
    {
        SD->setMapEntry("returnReport","agents_too_far");
        return false;
    }

    _WSF = new WorldStateFacts(_WS->getRobotConfs());

    return true;
}

bool Give::getRandomSol()
{

//    GeometricComputations GC = GeometricComputations();
    _WS->load();
    confPtr_t objectCurConf = mainObject->getCurrentPos();

    detachObject();

    bool res = false;
    int nbTestsPlace = -1;

    int reachedReturnReportCount = _returnReportcount;
    while (!res && nbTestsPlace < maxTests)
    {
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
        nbTestsPlace++;
        _returnReportcount = reachedReturnReportCount;
        mainObject->setAndUpdate(*objectCurConf);
//        cout<< "test nb = " << nbTests << endl;

        if (!findCurrentTestPoint())
            continue;


        configPt q = findObjectConf();
        if (!q)
            continue;
        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(), q);

        _objectFinalConf = confPtr_t(new Configuration(mainObject,q));

        if(isObjectInCollision(NULL))
            continue;

        if(isHandInCollision())
            continue;
        res = true;
    }

    _WS->load();
    return res;
}

bool Give::findConfigurations()
{
    _WS->load();
    removeVirtualObjects();
    detachObject();

    mainObject->setAndUpdate(*_objectFinalConf);
    activateFingertipCol(false, mainAgent, mainObject, NULL);



    //HACK
    if(mainAgent->getHriAgent()->type == HRI_PR2)
    {
        _curGrasp.hand_type = GP_PR2_GRIPPER;
    }

    p3d_matrix4 matPos;
    p3d_mat4Copy(mainObject->getJoint(0)->getJointStruct()->abs_pos, matPos);
    bool test = computeConf(_agentFinalConf, matPos, _curGrasp);
    activateFingertipCol(true, mainAgent, mainObject, NULL);

    _WS->load();
    if (test)
    {
        mainObject->setAndUpdate(*_objectFinalConf);
        mainAgent->setAndUpdate(*_agentFinalConf);
    }

    if (test && !checkTargetReach())
        return false;

    _WS->load();

    return test;
}

bool Give::findTrajectories()
{
    activateFingertipCol(false, mainAgent, mainObject, NULL);
    _WS->load();
    attachObject();

    bool ok = true;

    int status = -1;
    approachTraj = compute_traj(_WS->getRobotConf(mainAgent), _agentFinalConf, mainObject, status);

    if (!approachTraj)
        ok = false;

    _returnReportcount++;
    if (!ok)
    {
        cout << "no approach traj" << endl;
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_approach_traj");
        }
    }
    else
    {
        setGraph(new Graph(*API_activeGraph));
        SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",(int)getGraph()->getNodes().size());
        SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",approachTraj->nlp+1);
        SD->getSDMapEntry("graph")->setMapEntry("trajSize",approachTraj->range_param);
    }

    activateFingertipCol(true, mainAgent, mainObject, NULL);
    moveBackVirtualObjects();

    return ok;
}

bool Give::computeSolutionTrajectories(TaskSolution *TS)
{
    bool res=true;
    if(!TS)
    {
        res = false;
    }

    if(res){
        TS->getStartWS()->load();
        TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);

        _agentFinalConf = TSS0->getEndState()->getRobotConf(mainAgent);
        _objectFinalConf = TSS0->getEndState()->getRobotConf(mainObject);

        res = findTrajectories();
    }


    if(res){
        TS->getSubSolutionWithId(0)->setTraj(approachTraj);
        TS->setTrajComputed(true);
    }

    return res;
}

TaskSolution *Give::createSolution(bool computeMP)
{
    TaskSolution* TS = new TaskSolution();
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);

    _WS->load();

    TS->setStartWS(_WS);

    TSS1->setStartState(_WS);
    WorldState* WS1 = new WorldState(_WS);
    WS1->replaceConf(_agentFinalConf);
    WS1->replaceConf(_objectFinalConf);
    TSS1->setEndState(WS1);
    TS->setStopWS(WS1);

    WS1->setTSThatEndWithThisWS(TS);
    WS1->setTCForNextTask(_currConstraints);


    return TS;
}


bool Give::setAgents(multimap<string,string> agents_name)
{
    return (setMainAgent(agents_name) && setTargetAgent(agents_name));
}

bool Give::setObjects(multimap<string,string> objects_name)
{
    return (setMainObject(objects_name));
}

bool Give::setPoints(multimap<string,p3d_point> points)
{
    return setTargetPoint(points);
}


bool Give::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
        cout << "no arm choosen, go for right" << endl;
        setArmToUse(Right);
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        cout << "setting default arm to right" << endl;
        setArmToUse(Right);
    }

    return true;
}


bool Give::setTargetPoint(multimap<string,p3d_point> points)
{

    if (points.count("target") == 1)
    {
        cout << "got a target point" << endl;
        _hasTargedPoint = true;
        _targetPoint = points.find("target")->second;
        return true;
    }

//    noSupport = true;
    return true;
}

void Give::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "approach")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
            attachObject();
            tss->refineTraj();
            detachObject();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}

//This function is not efficient at all, TODO having a better sampling system
int nbMaxOfTestInSphere = 100;

bool Give::findCurrentTestPoint()
{
    //finding spheres
    //finding hexadron
    GeometricComputations GC = GeometricComputations();
    int nbTestaa=0;
    bool reachablePointFound = false;
    while (nbTestaa < nbMaxOfTestInSphere && !reachablePointFound)
    {
        nbTestaa++;
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");

        p3d_point tmp;
        tmp.x = p3d_random(_mainAgentShoulder.x - _mainAgentReach, _mainAgentShoulder.x + _mainAgentReach);
        tmp.y = p3d_random(_mainAgentShoulder.y - _mainAgentReach, _mainAgentShoulder.y + _mainAgentReach);
        tmp.z = p3d_random(_mainAgentShoulder.z - _mainAgentReach, _mainAgentShoulder.z + _mainAgentReach);

        if (GC.distBetweenPoints(tmp,_mainAgentShoulder) > _mainAgentReach)
            continue;

        for (unsigned int i = 0; i < _targetShoulder.size(); i++)
        {
            if (GC.distBetweenPoints(tmp, _targetShoulder.at(i).second) < _targetAgentReach)
                reachablePointFound = true;
        }

        if (reachablePointFound)
            _currentTest = tmp;
    }

    _returnReportcount++;
    if (!reachablePointFound)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_point_found_in_intersection");
        }
        return false;
    }

    return true;
}

configPt Give::findObjectConf()
{
    configPt q = p3d_copy_config(mainObject->getRobotStruct(), mainObject->getCurrentPos()->getConfigStruct());
    q[6] = _currentTest.x;
    q[7] = _currentTest.y;
    q[8] = _currentTest.z;


    q[9] = p3d_random(-M_PI,M_PI);
    q[10] = p3d_random(-M_PI,M_PI);
    q[11] = p3d_random(-M_PI,M_PI);


    return q;
}



bool Give::checkTargetReach()
{
    _WSF->setConfiguration(_objectFinalConf);
    global_Project->getActiveScene()->getFactsManager()->computeReachability(_WSF,targetAgent,mainObject,true);

    fact_id FI3;
    FI3.factType = "IsReachable";
    FI3.factSubType = "GIK";
    FI3.isMainAnAgent = true;

    Fact F = _WSF->getFact(FI3, targetAgent, mainObject);
//    F.printFact();
    return F.getBoolValue();
}

string Give::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if (targetAgent)
        s+= "\nMain agent \t\t= " + targetAgent->getName();
    if (mainObject)
        s+= "\nMain object \t= " + mainObject->getName() ;
    return s;
}

string Give::getTextFromValues()
{
    return "Task: " + type + " : Give the object to the target agent"
            + "\n    agents: "
            + "\n        mainAgent:  [AGENT ID]"
            + "\n        targetAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject: [OBJECT ID]"
            + "\n    datas:"
            + "\n        hand: [right OR left]"
            + "\n    points:"
            + "\n        target: [X Y THETA]";
}


