#include "placeHuman.hpp"
#include "API/project.hpp"
#include <libmove3d/hri/hri.h>

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

using namespace std;


double distToObject = 0.05;

//####################### PlaceHuman ########################

PlaceHuman::PlaceHuman() : Place()
{
    type = "placeHuman";
    typeId = placeHuman;
    _armId = -1;
    noSupport = false;


    nbMaxLoops = 50;
    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 5);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", nbMaxLoops);


    _hasTargedPoint = false;
    //_possibleMainAgentType.push_back("HERAKLES");
    _possibleMainAgentType.push_back(HRI_HERAKLES);
}


bool PlaceHuman::initAll()
{

    result = new TaskAllSolutions(this);
    if (_WS->getAttachements().size() == 0)
    {
        SD->setMapEntry("returnReport","no_attachement_in_ws");
        return false;
    }
    return true;
}


bool PlaceHuman::initialize()
{
    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;

    _currentSurface = NULL;



    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->getSDMapEntry("times")->setMapEntry("timeMp", 0.0);
    SD->setMapEntry("returnReport", "ERROR");


    _returnReportcount = 0;
    _returnReportcountMax = 0;

    //moving all virtual object to 0 and storing confs

    return true;
}


bool PlaceHuman::findCandidateSolutions()
{

    if (!checkAndSetArmAndObj())
        return false;
    //check if seted hand is the same as the one with the object

    if (!checkAttachement())
    {
        SD->setMapEntry("returnReport","no_corresponding_attachement_in_ws");
        return false;
    }

    if (!checkHRIAgent(mainAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    if (!findSupports())
        return false;

    if (!findRotations())
        return false;



    return true;
}

bool PlaceHuman::getRandomSol()
{

    confPtr_t robotCurConf = confPtr_t(new Configuration(*mainAgent->getCurrentPos()));
    confPtr_t objectCurConf = confPtr_t(new Configuration(*mainObject->getCurrentPos()));

    GeometricComputations GC = GeometricComputations();

    detachObject();

    bool res = false;
    int nbTestsPlace = -1;

    int reachedReturnReportCount = _returnReportcount;

    // in case of specified point
    if(_hasTargedPoint && _supportObject)
    {
        cout << "placing the object " << mainObject->getName() << " on object " << _supportObject->getName() << " at position ( "
             << _targetPoint.x << " , " << _targetPoint.y << ") with rotation : " << _targetPoint.z << endl;


    }

    int maxTests = 500;

    while (!res && nbTestsPlace < maxTests)
    {
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
        nbTestsPlace++;
        _returnReportcount = reachedReturnReportCount;
        mainObject->setAndUpdate(*objectCurConf);

        if (!findCurrentTestPoint())
            continue;

        configPt q = findObjectConf();
        if (!q)
            continue;

        if (!isObjectInSurface())
            continue;

        if(isObjectInCollision(_supportObject))
            continue;

        vector<p3d_point> vptt = GC.getUpperPointsInObject(mainObject,0.03);

        if (vptt.size() >= 4)
        {
            targetPoint.x = (vptt.at(0).x +vptt.at(3).x)/2;
            targetPoint.y = (vptt.at(0).y +vptt.at(3).y)/2;
            targetPoint.z = vptt.at(0).z + distToObject;
        }
        else if (vptt.size()>=2)
        {
            targetPoint.x = (vptt.at(0).x +vptt.at(1).x)/2;
            targetPoint.y = (vptt.at(0).y +vptt.at(1).y)/2;
            targetPoint.z = vptt.at(0).z + distToObject;
        }
        else if (vptt.size() >=1)
        {
            targetPoint = vptt.at(0);
            targetPoint.z = vptt.at(0).z + distToObject;
        }
        else
        {
            return false;
        }


        cout << "targetpoint = ( " << targetPoint.x << " , " << targetPoint.y << " , " << targetPoint.z << " )" << endl;

        ag_reach = mainAgent->getHriAgent()->manip->reach_tolerance;
        agShoulder.x = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[0][3];
        agShoulder.y = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[1][3];
        agShoulder.z = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[2][3];
        _returnReportcount++;
//        cout << "agreach = " << ag_reach << " dist = " << GC.distBetweenPoints2D(targetPoint,agShoulder) << endl;
        if(GC.distBetweenPoints2D(targetPoint,agShoulder) > ag_reach)
        {
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","obj_too_far");
            }
            return false;
        }


        res = true;



        if (!testFactConstraints())
            continue;

        _confSol = confPtr_t(new Configuration(*mainObject->getCurrentPos()));

    }


    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
//    attachObject();
    return res;
}

bool PlaceHuman::findConfigurations()
{
    detachObject();

    HRI_GIK_TASK_TYPE task;
    if (m_arm == Left)
        task = GIK_LATREACH;
    else if (m_arm == Right)
        task = GIK_RATREACH;

    p3d_vector3 Tcoord;

    Tcoord[0] = targetPoint.x;
    Tcoord[1] = targetPoint.y;
    Tcoord[2] = targetPoint.z;
    double distance_tolerance = 0.05;

    configPt q = p3d_get_robot_config(mainAgent->getRobotStruct());


    int IKSucceded = hri_agent_single_task_manip_move(mainAgent->getHriAgent(),
                                                     task,
                                                      &Tcoord,
                                                      distance_tolerance,
                                                      &q);

    IKSucceded = hri_agent_single_task_manip_move(mainAgent->getHriAgent(),
                                                         task,
                                                          &Tcoord,
                                                          distance_tolerance,
                                                          &q);


    SD->getSDMapEntry("statistics")->incrementMapEntry("nbGikCalled");

    if (IKSucceded != 1)
    {
        p3d_vector3 v;
        p3d_mat4ExtractTrans(mainAgent->getRobotStruct()->joints[mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()]+7]->abs_pos,v);
        double dist = sqrt(pow(v[0] - targetPoint.x,2) + pow(v[1] - targetPoint.y,2) + pow(v[2] - targetPoint.z,2));

        _returnReportcount++;
        if(dist > distance_tolerance)
        {
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","no_ik");
            }
            return false;
        }
    }


    p3d_set_and_update_this_robot_conf(mainAgent->getRobotStruct(),q);

    _returnReportcount++;
    if(pqp_robot_all_no_self_collision_test(mainAgent->getRobotStruct()) == 1)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","ik_in_colision");
        }
        return false;
    }

    approachConf = confPtr_t(new Configuration(mainAgent,q));

    return true;

}

bool PlaceHuman::findTrajectories()
{
    p3d_col_deactivate_robot(mainAgent->getRobotStruct());
    int status = -1;
    approachTraj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent),approachConf,mainObject, status);
    SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",1);
    SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",approachTraj->nlp+1);
    SD->getSDMapEntry("graph")->setMapEntry("trajSize",approachTraj->range_param);
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");
    p3d_col_activate_robot(mainAgent->getRobotStruct());

    _returnReportcount++;
    if(!approachTraj)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_traj_found");
        }
        return false;
    }
    return true;

}

bool PlaceHuman::computeSolutionTrajectories(TaskSolution *TS)
{
    bool res = true;
    cout << getDescr() << endl;
    if(!TS)
    {
        res = false;
    }
    if(res){


        TS->getStartWS()->load();

        confPtr_t q1 = TS->getStopWS()->getRobotConf(mainAgent);

        approachConf = q1;

        res = findTrajectories();
    }
    if(res){

        TS->getSubSolutionWithId(0)->setTraj(approachTraj);

        TS->setTrajComputed(true);
    }

    return res;
}

TaskSolution *PlaceHuman::createSolution(bool computeMP)
{
    TaskSolution* TS = new TaskSolution();
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    detachObject();


    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(approachConf);
    if (!t)
    {
        cout << "ERROR: error with robots names 1" << endl;
        delete TS;
        return NULL;
    }

    t = WS1->replaceConf(_confSol);
    if (!t)
    {
        cout << "ERROR: error with robots names 2" << endl;
        delete TS;
        return NULL;
    }

    TSS1->setEndState(WS1);

//    WS1->clearAttachement();
    WS1->removeAttachements(mainAgent,mainObject);

    TS->setStopWS(WS1);
    WS1->setTSThatEndWithThisWS(TS);
    WS1->setTCForNextTask(_currConstraints);

    return TS;
}



void PlaceHuman::smoothSolution(int alternativeId)
{
    alternativeId = 0;
}
