#include "manipulationTask.hpp"

#include "API/HPP_Wrap/Romeo_Wrap.hpp"
#include "P3d-pkg.h"
#include "API/project.hpp"

#include "planner/plannerFunctions.hpp"
#include "plannerFunctions.hpp"
#include "planner/cost_space.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#include "database/semantics/Sem_SemanticsManager.hpp"

#ifdef HPP
RomeoWrap *_romeoWrap;
#endif

#include <iostream>
#include <map>

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

using namespace std;

ManipulationTask::ManipulationTask() : PrivateTask()
{
    currentTransfo = 0;
    mainObject = NULL;
    _noArm = false;
}

bool ManipulationTask::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = createSolution(computeMP);
    if(TS){
        TS->setId(solutionId);
        TS->setTaskAllSolutions(result);
        getResult()->addAlternativeSol(TS);
    }
    return (TS != NULL);
}

bool ManipulationTask::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }
    return computeSolutionTrajectories(TS);

}

bool ManipulationTask::setMainObject(multimap<string,string> objects_name)
{
    if (objects_name.count("mainObject") != 1)
    {
        cout << "ERROR: no main object for " << type << " task." << endl;
        SD->setMapEntry("returnReport","no_main_object");
        return false;
    }
    mainObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("mainObject")->second);
    if (!mainObject)
    {
        SD->setMapEntry("returnReport","no_main_object_name");
        return false;
    }
    return true;
}

bool asd = true;

bool ManipulationTask::run(multimap<string,string> agents_name,
                           multimap<string,string> objects_name,
                           multimap<string,p3d_point> point3d,
                           multimap<string,string> additionalData,
                           WorldState *WS, bool computeMP)
{
    WS->load();
    _WS = WS;
    bool computeMP_aux=false;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!isPossibleMainAgent(mainAgent))
    {
        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setObjects(objects_name))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }


    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    bool res = false;

    //find a certain number of alternatives
    //TODO change settable to stop condition
    if (ENV.getBool(Env::enableHri))
    {
        if(SD->getSDMapEntry("statistics")->getIntMapEntry("nbTotPossibleTests")<=0 || nbMaxLoops <=0)
            res=false;
        else{
            unsigned int nbLoops = static_cast<unsigned int>(std::min(SD->getSDMapEntry("statistics")->getIntMapEntry("nbTotPossibleTests"),nbMaxLoops))*2;
            for (unsigned int i=0;i<nbLoops;i++){
                TaskSolution *ts=createAlternative(false);
                if(ts!=NULL){
                    std::vector<TaskSubSolution*> all_tss = ts->getAllTSS();
                    double c(0);
                    ts->getStartWS()->load();
                    c += global_costSpace->cost(*this->getMainAgent()->getCurrentPos());
                    for(unsigned int t=0;t<all_tss.size();t++){
                        TaskSubSolution *tss = all_tss[t];
                        tss->getEndState()->load();
                        c += global_costSpace->cost(*this->getMainAgent()->getCurrentPos());
                    }
                    exploredSolutions.push_back(make_pair(c,ts));
                }
            }
        }


        computeMP_aux=computeMP;
        res=false;
        std::sort(exploredSolutions.rbegin(),exploredSolutions.rend());
        if(exploredSolutions.size()){
            cout << "ManipulationTask found "<<exploredSolutions.size() << " alternatives, with cost ranging from "<<exploredSolutions.front().first
                 << " to " << exploredSolutions.back().first<<endl;
        }else{
            cout << "no exploredSolutions !!! "<<endl;
        }
        while(!res && exploredSolutions.size()){
            TaskSolution *ts=exploredSolutions.back().second;
            exploredSolutions.pop_back();
            if(computeMP_aux)
                if(!computeSolutionTrajectories(ts)){
                    continue;
                }
            ts->setId(0);
            ts->setTaskAllSolutions(result);
            result->addAlternativeSol(ts);
            res=true;
        }

        if (res)
            SD->setMapEntry("returnReport","OK");
    } else {


//        int count = 0;
//        while(!res && count < nbMaxLoops)
//        {
            TaskSolution *ts=createAlternative(computeMP);
            if (ts)
            {
                ts->setId(0);
                ts->setTaskAllSolutions(result);
                result->addAlternativeSol(ts);
                res=true;
            }
//            count++;
//        }

        if (res)
            SD->setMapEntry("returnReport","OK");
    }
    return res;
}

TaskSolution* ManipulationTask::findAlternative(bool computeMP)
{
//    int count = 0;
    if (!getResult()->getAlternativeWithId(0))
    {
        cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
        return NULL;
    }
    getResult()->getAlternativeWithId(0)->getStartWS()->load(); //

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    int id = -1;
    bool res = false;

    TaskSolution *TS;
    // use explored solution first
    while ( !res && exploredSolutions.size()){
        TS = exploredSolutions.back().second;
        exploredSolutions.pop_back();
        if (TS->isTrajComputed() == false && computeMP ){
            if(computeSolutionTrajectories(TS)){
                id = getResult()->getNextId();
                TS->setId(id);
                TS->setTaskAllSolutions(result);
                result->addAlternativeSol(TS);
                cout << "findAlternative: using already explored alternative"<<endl;
                res=true;
            }
        }
    }

    if (!res){
        TS = createAlternative(computeMP);
        cout << "findAlternative: using newly created alternative"<<endl;
        if (TS){
            id = getResult()->getNextId();
            TS->setId(id);
            TS->setTaskAllSolutions(result);
            result->addAlternativeSol(TS);
            res=true;
        }
    }

    if (res)
        SD->setMapEntry("returnReport","OK");

    computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (res && id != -1)
        return getResult()->getAlternativeWithId(id);
    return NULL;
}

TaskSolution *ManipulationTask::createAlternative(bool computeMP)
{
    int count = 0;
    // if (!getResult()->getAlternativeWithId(0))
    // {
    //     cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
    //     return NULL;
    // }
    // getResult()->getAlternativeWithId(0)->getStartWS()->load(); //

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }


    //int id = -1;
    bool res = false;
    TaskSolution *TS=NULL;
    while (!res)
    {
//        cout << "count loop = " << count << endl;
        _returnReportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
           continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        // id = result->getNextId();

        TS = createSolution (computeMP);
        if (!TS)
            continue;
        // if(!setToSolution(id,computeMP))
        //    continue;
        if(!testConstraints(TS))
            continue;
        res = true;
    }
    if (res)
        SD->setMapEntry("returnReport","OK");

    //computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return TS;

}


void ManipulationTask::showMainObjectRandomTransform()
{
    double Tx, Ty, Tz, Rx, Ry, Rz;
    p3d_matrix4 res;

    confPtr_t qO = mainObject->getCurrentPos();
    p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos, mainObject->getRobotStruct()->objectTransform[currentTransfo].value, res);
    p3d_mat4ExtractPosReverseOrder(res, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);
    (*qO)[6] = Tx; (*qO)[7] = Ty; (*qO)[8] = Tz; (*qO)[9] = Rx; (*qO)[10] = Ry; (*qO)[11] = Rz;
    mainObject->setAndUpdate(*qO);

    if (currentTransfo < (mainObject->getRobotStruct()->objectTransform.size() - 1))
        currentTransfo ++;
    else
        currentTransfo = 0;
}

bool ManipulationTask::attachObject()
{
    return PrivateTask::attachObject(mainAgent, mainObject, true);
}

bool ManipulationTask::detachObject()
{
    return PrivateTask::attachObject(mainAgent, mainObject, false);
}



bool ManipulationTask::checkHRIAgent(Robot* r)
{
    if (!r->getHriAgent())
    {
        cout << "hri agent doesn't exist for "<< r->getName()<<" , unable to compute "<< type << endl;
        SD->setMapEntry("returnReport","no_hri_agent");
        return false;
    }
    if (!mainAgent->getHriAgent()->manip)
    {
        cout << "hri agent doesn't have manip data, unable to compute task" << endl;
        SD->setMapEntry("returnReport","no_manip_data_in_agent");
        return false;
    }
    return true;
}

bool ManipulationTask::setMainAgentHand()
{
    if (mainAgent->getHriAgent()->manip->external_hand_no <= 0)
    {
        cout << "hri agent doesn't have external hands" << endl;
        SD->setMapEntry("returnReport","no_external_hand_in_struct");
        return false;
    }

    for (int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
    {
        if (!mainAgent->getHriAgent()->manip->external_hand)
            return false;
        if (!mainAgent->getHriAgent()->manip->external_hand[i])
            return false;
        if (mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
        {
            _armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
            _hands[_armId] = global_Project->getActiveScene()->getRobotByName(mainAgent->getHriAgent()->manip->external_hand[i]->rob_hand->name);
            break;
        }
    }
    if (!_hands[_armId])
    {
        cout << "ERROR, the hand of the robot was not found" << endl;
        SD->setMapEntry("returnReport","no_external_hand_in_world");
        return false;
    }
    return true;
}


bool ManipulationTask::isObjectInCollision(Robot* supportObject)
{
    if (supportObject)
        p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), supportObject->getRobotStruct());

    _returnReportcount++;
    bool isColliding = mainObject->getCurrentPos()->isInCollision();
//    if(p3d_col_test_robot(mainObject->getRobotStruct(),0) == 1)//there is collision
    if (isColliding)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","object_in_collision");
        }
//        p3d_print_col_pair();
        vector<p3d_rob*> colRob = pqp_robot_robot_list_collision_test(mainObject->getRobotStruct());
        for (unsigned int i = 0; i < colRob.size(); i++)
        {
            SD->getSDMapEntry("colision")->incrementMapEntry(colRob.at(i)->name);
        }
    }

    if (supportObject)
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), supportObject->getRobotStruct());
    return isColliding;
}


bool ManipulationTask::isHandInCollision()
{

    p3d_matrix4 rot;
    p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos,_curGrasp.frame, rot);

    double Tx, Ty, Tz;
    double Rx, Ry, Rz;
    p3d_mat4ExtractPosReverseOrder(rot, &Tx, &Ty, &Tz,
                                   &Rx, &Ry, &Rz);


    confPtr_t q_hand = _hands[_armId]->getCurrentPos();


    double Itx = (*q_hand)[6], Ity = (*q_hand)[7],  Itz = (*q_hand)[8], Irx = (*q_hand)[9], Iry = (*q_hand)[10], Irz = (*q_hand)[11];

    (*q_hand)[6] = Tx; (*q_hand)[7] = Ty; (*q_hand)[8] = Tz;
    (*q_hand)[9] = Rx; (*q_hand)[10] = Ry; (*q_hand)[11] = Rz; (*q_hand)[12] = 0.15;
    _hands[_armId]->setAndUpdate(*q_hand);

    p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _hands[_armId]->getRobotStruct());

    bool colTest = true;// = p3d_col_test_robot(_hands[_armId]->getRobotStruct(),0) == 1;

    vector<p3d_rob*> colRob = pqp_robot_robot_list_collision_test(_hands[_armId]->getRobotStruct());

    if (colRob.size() == 0 || (colRob.size() == 1 && colRob.at(0) == mainObject->getRobotStruct()))
        colTest = false;

    if (colTest)
    {
        for (unsigned int i = 0; i < colRob.size(); i++)
        {
            SD->getSDMapEntry("colision")->incrementMapEntry(colRob.at(i)->name);
        }
    }


    p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _hands[_armId]->getRobotStruct());


    (*q_hand)[6] = Itx; (*q_hand)[7] = Ity; (*q_hand)[8] = Itz;
    (*q_hand)[9] = Irx; (*q_hand)[10] = Iry; (*q_hand)[11] = Irz;
    _hands[_armId]->setAndUpdate(*q_hand);

    _returnReportcount++;
    if(colTest) // there is collision
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","hand_in_collision");

        }
    }

    return colTest;
}



void ManipulationTask::removeVirtualObjects()
{
    virtualObjs.clear();
    for (unsigned int v = 0; v < global_Project->getActiveScene()->getNumberOfRobots(); v++)
    {
        Robot* re = global_Project->getActiveScene()->getRobot(v);
        if (re->getName().find("VIRTUAL") != string::npos)
        {
            pair<Robot*,configPt> rcp;
            rcp.first = re;
            rcp.second = re->getCurrentPos()->getConfigStruct();
            virtualObjs.push_back(rcp);

            configPt qtmp = p3d_copy_config(re->getRobotStruct(), re->getCurrentPos()->getConfigStruct());
            qtmp[6] = 0;
            qtmp[7] = 0;
            qtmp[8] = 0;
            p3d_set_and_update_this_robot_conf(re->getRobotStruct(),qtmp);
        }
    }
}

void ManipulationTask::moveBackVirtualObjects()
{
    for (unsigned int v = 0; v < virtualObjs.size(); v++)
    {
        p3d_set_and_update_this_robot_conf(virtualObjs.at(v).first->getRobotStruct(),virtualObjs.at(v).second);
    }
}


bool ManipulationTask::checkAttachement()
{
    if (_WS->getAttachements().size() == 0)
    {
        return false;
    }
    Attachement att = _WS->getAttachement(mainAgent,mainObject);
    if(att.object == NULL || att.agent == NULL)
    {
        return false;
    }
    _curGrasp = att.grasp;
    return true;
}

bool ManipulationTask::checkAndSetArmAndObj()
{
    if (!mainObject && _noArm)
    {
        SD->setMapEntry("returnReport","no_obj_and_no_arm");
        return false;
    }
    else if (!mainObject && !_noArm)//if no obj setobjfrom attachment
    {
        mainObject = getMainObjectFromAtt();
        if (!mainObject)
        {
            SD->setMapEntry("returnReport","no_attachement_for_arm");
            return false;
        }
    }
    else if (mainObject && _noArm)
    {
        int aI = getArmIdFromAtt();
        if (aI == -1)
        {
            SD->setMapEntry("returnReport","no_attachement_for_object");
            return false;
        }
        setArmToUse(getArmLabelFromId(aI));
    }
    else if (mainObject && !_noArm)
    {
        if (!checkArmAtt())
        {
            SD->setMapEntry("returnReport","wrong_arm_choice");
            return false;
        }
    }
    else
    {
        SD->setMapEntry("returnReport","unexpected_error");
        return false;
    }
    return true;
}

Robot* ManipulationTask::getMainObjectFromAtt()
{
    if (_WS->getAttachements().size() == 0)
    {
        SD->setMapEntry("returnReport","no_attachement_in_ws");
        return false;
    }
    vector<Attachement> vatt = _WS->getAttachements();
    for (unsigned int i = 0; i < vatt.size(); i++)
    {
        Attachement att = vatt.at(i);
        if (att.agent && att.object)
        {
            if (att.agent->getName() == mainAgent->getName() && att.armId == getArmId())
            {
                return att.object;
            }
        }
    }
    return NULL;
}

int ManipulationTask::getArmIdFromAtt()
{
    if (_WS->getAttachements().size() == 0)
    {
        SD->setMapEntry("returnReport","no_attachement_in_ws");
        return false;
    }
    vector<Attachement> vatt = _WS->getAttachements();
    for (unsigned int i = 0; i < vatt.size(); i++)
    {
        Attachement att = vatt.at(i);
        if (att.agent && att.object)
        {
            if (att.agent->getName() == mainAgent->getName() && att.object->getName() == mainObject->getName())
            {
                return att.armId;
            }
        }
    }
    return -1;
}


bool ManipulationTask::checkArmAtt()
{
    if (_WS->getAttachements().size() == 0)
    {
        return false;
    }
    Attachement att = _WS->getAttachement(mainAgent,mainObject);
    if(att.object == NULL || att.agent == NULL)
    {
        return false;
    }
    else if (att.armId != getArmId())
    {
        return false;
    }
    return true;
}

