#include "pick.hpp"
#include "API/project.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include "planner/plannerFunctions.hpp"
#include "planner/cost_space.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#define APPROACHDIST 0.1
#define ESCAPEDIST 0.1


using namespace std;

//####################### Pick ########################

INIT_MOVE3D_STATIC_LOGGER(Pick, "planners.gtp.tasks.pick");

Pick::Pick() : ManipulationTask()
{
    type = "pick";
    typeId = pick;
    current_grasp_id = -1;
    _armId = 0;
    nbMaxLoops = 100;


    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);

    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 500);
    SD->setMapEntry("returnReport", "ERROR");

    _noArmPreference = true;
    _returnReportcount = 0;
    _returnReportcountMax = 0;

    graspLists.clear();

}

bool Pick::initAll()
{
    result = new TaskAllSolutions(this);

    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 2);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    _armId =0;
    return true;
}

bool Pick::initialize()
{
    approachTraj = NULL;
    graspTraj = NULL;
    escapeTraj = NULL;
    closeHandTraj = NULL;


    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("times")->setMapEntry("timeMp", 0.0);

    SD->setMapEntry("returnReport", "ERROR");

    _returnReportcount = 0;
    _returnReportcountMax = 0;

    if (checkAttachement())
    {
        SD->setMapEntry("returnReport", "attachement_already_exist_first");
        return false;
    }

    return true;
}

bool Pick::finalize()
{
    return true;
}

bool Pick::findCandidateSolutions()
{

    if (!checkHRIAgent(mainAgent))
        return false;

    Sem::ActionsManager * actMgr=Sem::SemanticsManager::mainSemanticsManager()->actionsManager();
    std::vector<Sem::PlanningPart> planParts = actMgr->getJointsForAction(Sem::ActionType(getType()),
                                                                          mainAgent->getHriAgent()->type,
                                                                          false);
    int aI = -1;
    if (!_noArmPreference)
        aI = _armId;
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", 0);
    for (unsigned int p = 0; p < planParts.size(); p++)
    {
        for (unsigned int a=0;a<planParts[p].armIds().size() ; a++){

            //if (_noArm)
            setArmToUse(planParts[p].armIds().at(a));

            if (!setMainAgentHand())
                return false;

            if (!getGrasps())
                return false;
            if (_noArmPreference)
                SD->getSDMapEntry("statistics")->incrementMapEntry("nbTotPossibleTests", graspLists[_armId].size());
        }

    }
    _armId = aI;

    if (!_noArmPreference)
        SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", (int)graspLists[_armId].size());


    if (checkAttachement())
    {
        SD->setMapEntry("returnReport", "attachement_already_exist");
        return false;
    }

    return true;
}

bool Pick::getRandomSol()
{
    // TODO: armmanagement
    //need to define a way to choose the arm to use
    //closest to the arm need more thought

    _WS->load();
    _returnReportcount++;

    if (_noArmPreference){
        std::vector<Sem::PlanningPart> planParts = Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->getJointsForAction(Sem::ActionType(getType()),mainAgent->getHriAgent()->type,false);
        _planPartToUse = planParts.at(p3d_random_integer(0,planParts.size()-1));
        setArmToUse(_planPartToUse.armIds().at(p3d_random_integer(0,_planPartToUse.armIds().size()-1)));
        M3D_TRACE("will plan with arm id=" << _armId);
    }else{
        std::vector<Sem::PlanningPart> planParts = Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->getJointsForAction(Sem::ActionType(getType()),mainAgent->getHriAgent()->type,false);
        std::vector<int> possiblePlanParts;
        for (unsigned int i = 0; i<planParts.size() ; i++){
            Sem::PlanningPart p = planParts.at(i);
            if (std::find(p.armIds().begin(), p.armIds().end(), _armId) != p.armIds().end()){
                //contain the arm we want
                possiblePlanParts.push_back(i);
            }
        }
        _planPartToUse = planParts.at(possiblePlanParts.at( p3d_random_integer(0,possiblePlanParts.size() -1 ) ));
    }

    if (graspLists[_armId].size() > 0)
    {
        current_grasp_id = p3d_random_integer(0, graspLists[_armId].size()-1);
        mainAgent->getHriAgent()->graspId = graspLists[_armId].at(current_grasp_id).ID;
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
//        cout << "nbTest " << nbTests << endl;
        return true;
    }

    cout<< "nb grasp = 0" << endl;
    if (_returnReportcountMax < _returnReportcount)
    {
        _returnReportcountMax = _returnReportcount;
        SD->setMapEntry("returnReport","no_grasp_in_list");
    }


    return false;
}

bool Pick::findConfigurations()
{

    confPtr_t qInit = _WS->getRobotConf(mainAgent);
    _curGrasp = graspLists[_armId].at(current_grasp_id);

    //graspLists[_armId].erase(graspLists[_armId].begin()+current_grasp_id);

    if(isHandInCollision())
        return false;

    confPtr_t qTmp = (confPtr_t) new Configuration(*qInit);
    openGripper(qTmp,true,_armId);

    mainAgent->setAndUpdate(*qTmp);
    p3d_update_this_robot_pos_without_cntrt(mainAgent->getRobotStruct());

//    p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
    activateFingertipCol(false, mainAgent, mainObject, NULL);

    p3d_vector3 v1, v2;
    _curGrasp.direction(v1);

    v2[0] = 0;
    v2[1] = 0;
    v2[2] = 1;


    bool test = computeManipulationConfs(approachConf, graspConfClosed, graspConfOpen, escapeConf, mainObject->getJoint(0)->getJointStruct()->abs_pos,
                                         _curGrasp, v1, APPROACHDIST, v2, ESCAPEDIST);

    correctConfigs();

    mainAgent->setAndUpdate(*qInit);
    return test;
}



bool Pick::findTrajectories()
{
    bool ok=true;
    int status = -1;
    _WS->load();
    //compute traj between actual conf and approachConf

    std::vector<p3d_rob*> col_deactivated;
    for (unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots(); ++i){
        p3d_rob *r1,*r2;
        r1=mainObject->getRobotStruct();
        r2=global_Project->getActiveScene()->getRobot(i)->getRobotStruct();
        bool col=p3d_col_test_robot_other(r1,r2,0);
        if(col){
            col_deactivated.push_back(r2);
            M3D_DEBUG("object "<<r2->name<<" is in collision with mainObject, will be deactivated for escape");
        }
    }


    if(ok){
        graspTraj = compute_traj_using_localPath(approachConf, graspConfOpen, mainObject, status);

        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)

        {
            cout << "no grasp traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","grasp_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","grasp_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","grasp_traj_invalid_stop");
            }
            ok = false;
        }
    }

    if(ok) {
        if (graspConfOpen->equal(*graspConfClosed)) {
            closeHandTraj = NULL;
        } else {
            _WS->load();
            mainAgent->setAndUpdate(*graspConfOpen);

            activateFingertipCol(false, mainAgent, mainObject, NULL);
            closeHandTraj = compute_traj_using_localPath(graspConfOpen, graspConfClosed, mainObject, status);
            activateFingertipCol(true, mainAgent, mainObject, NULL);

            _returnReportcount++;
            if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)

            {
                cout << "no closing hand traj" << endl;
                if (_returnReportcountMax < _returnReportcount)
                {
                    _returnReportcountMax = _returnReportcount;
                    if (status == PLAN_STATUS_START_INVALID)
                        SD->setMapEntry("returnReport","closing_hand_traj_invalid_start");
                    else if (status == PLAN_STATUS_NOT_FOUND)
                        SD->setMapEntry("returnReport","closing_hand_traj_not_found");
                    else if (status == PLAN_STATUS_GOAL_INVALID)
                        SD->setMapEntry("returnReport","closing_hand_traj_invalid_stop");
                }
                ok = false;
            }
        }
    }

    if(ok){
        _WS->load();
        mainAgent->setAndUpdate(*graspConfClosed);
        attachObject();


        // disable fingerType collisions
        //     gpDeactivate_object_fingertips_collisions(p3d_rob *robot, p3d_obj *object, gpHand_properties &hand, int handID)
        //    gpHand_properties armHandProp = ( *mainAgent->getRobotStruct()->armManipulationData ) [getArmId()].getHandProperties();
        //    gpDeactivate_object_fingertips_collisions(mainAgent->getRobotStruct(), mainObject->getRobotStruct()->joints[1]->o, armHandProp , getArmId());

        //TODO: get support and remove collision only between support and obj


        //p3d_col_deactivate_robot(mainObject->getRobotStruct());
        vector<SurfaceAttachement> sAtt = createAndLoadSurfaceAttachements(mainObject);
        activateFingertipCol(false, mainAgent, mainObject, NULL);
        for(unsigned int i = 0 ; i<col_deactivated.size();i++)
            p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), col_deactivated[i]);
        escapeTraj = compute_traj_using_localPath(graspConfClosed, escapeConf, mainObject, status);
        //p3d_col_activate_robot(mainObject->getRobotStruct());
        activateFingertipCol(true, mainAgent, mainObject, NULL);
        for(unsigned int i = 0 ; i<col_deactivated.size();i++)
            p3d_col_activate_robot_robot(mainObject->getRobotStruct(), col_deactivated[i]);

        removeSurfaceAttachements(sAtt);

        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)
        {
//            cout << "no escape traj found" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","escape_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","escape_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","escape_traj_invalid_stop");
            }
            ok = false;
        }
    }
    detachObject();
    if(ok){
        _WS->load();
        approachTraj = compute_traj(_WS->getRobotConf(mainAgent), approachConf, mainObject, status);

        //activateFingertipCol(true, mainAgent, mainObject, NULL);
        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)
        {
            cout << "no approach traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","approach_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","approach_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","approach_traj_invalid_stop");
            }
            ok = false;
        }else{
            setGraph(new Graph(*API_activeGraph));
            SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",(int)getGraph()->getNodes().size());
            SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",approachTraj->nlp+1);
            SD->getSDMapEntry("graph")->setMapEntry("trajSize",approachTraj->range_param);
        }
    }

    return ok;
}

bool Pick::computeSolutionTrajectories(TaskSolution *TS)
{
    bool res=true;
    //cout << getDescr() << endl;
    if(!TS)
    {
        res = false;
    }

    if(res){
        TS->getStartWS()->load();
        TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);
        TaskSubSolution* TSS1 = TS->getSubSolutionWithId(1);
        TaskSubSolution* TSS2 = TS->getSubSolutionWithId(2);
        TaskSubSolution* TSS3 = TS->getSubSolutionWithId(3);

        confPtr_t q0 = TSS0->getEndState()->getRobotConf(mainAgent);
        confPtr_t q1 = TSS1->getEndState()->getRobotConf(mainAgent);
        confPtr_t q2 = TSS2->getEndState()->getRobotConf(mainAgent);
        confPtr_t q3 = TSS3->getEndState()->getRobotConf(mainAgent);

        approachConf = q0;
        graspConfOpen = q1;
        graspConfClosed = q2;
        escapeConf = q3;

        _armId = TSS0->getArmId();
        _planPartToUse = TSS0->getPlanningPart();

        res = findTrajectories();
    }


    if(res){
        TS->getSubSolutionWithId(0)->setTraj(approachTraj);
        TS->getSubSolutionWithId(1)->setTraj(graspTraj);
        TS->getSubSolutionWithId(2)->setTraj(closeHandTraj);
        TS->getSubSolutionWithId(3)->setTraj(escapeTraj);

        TS->setTrajComputed(true);
    }

    return res;
}

TaskSolution *Pick::createSolution(bool computeMP)
{
    correctConfigs();
    TaskSolution* TS = new TaskSolution();
    TS->setTrajComputed(computeMP);
    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    TaskSubSolution* TSS2 = new TaskSubSolution(TS, "engage", 1, graspTraj, mainAgent, 0);
    TaskSubSolution* TSS3 = new TaskSubSolution(TS, "grasp", 2, closeHandTraj, mainAgent, 0);
    TaskSubSolution* TSS4 = new TaskSubSolution(TS, "escape", 3, escapeTraj, mainAgent, 0);

    TSS1->setArmId(_armId);
    TSS2->setArmId(_armId);
    TSS3->setArmId(_armId);
    TSS4->setArmId(_armId);

    TSS1->setPlanningPart(_planPartToUse);
    TSS2->setPlanningPart(_planPartToUse);
    TSS3->setPlanningPart(_planPartToUse);
    TSS4->setPlanningPart(_planPartToUse);

    detachObject();


    vector<TaskSubSolution*> v2;
    v2.push_back(TSS2);
    TSS1->setNext(v2);

    vector<TaskSubSolution*> v3;
    v3.push_back(TSS3);
    TSS2->setNext(v3);

    vector<TaskSubSolution*> v4;
    v4.push_back(TSS4);
    TSS3->setNext(v4);

    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(approachConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    TSS1->setEndState(WS1);
    TSS2->setStartState(WS1);

    WorldState* WS2 = new WorldState(WS1);
    t = WS2->replaceConf(graspConfOpen);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    TSS2->setEndState(WS2);
    TSS3->setStartState(WS2);


    WorldState* WS3 = new WorldState(WS2);
    t = WS3->replaceConf(graspConfClosed);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    gpGrasp gr = _curGrasp;
    WS3->addAttachement(mainAgent, mainObject, gr, getArmId());
    TSS3->setEndState(WS3);
    TSS4->setStartState(WS3);

    WorldState* WS4 = new WorldState(WS3);
    WS4->load();
    vector<SurfaceAttachement> sss = createAndLoadSurfaceAttachements(mainObject);
    WS3->setSurfaceAttachements(sss);
//    WS4->setSurfaceAttachements(sss);
    t = WS4->replaceConf(escapeConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }
    mainAgent->setAndUpdate(*escapeConf);

    for (unsigned int i = 0; i<sss.size(); i++)
        WS4->replaceConf(sss.at(i).attache->getCurrentPos());

    _WS->load();
    //    mainAgent->setAndUpdate(*graspConfClosed);
    ////    p3d_mat4Print(mainObject->getJoint(0)->getJointStruct()->abs_pos,"avant");
    //    double avZ = mainObject->getJoint(0)->getJointStruct()->abs_pos[2][3];
    //    attachObject();
    //    mainAgent->setAndUpdate(*escapeConf);
    //    p3d_mat4Print(mainObject->getJoint(0)->getJointStruct()->abs_pos,"apres");

    confPtr_t q = mainObject->getCurrentPos();
    //    if (avZ == mainObject->getJoint(0)->getJointStruct()->abs_pos[2][3])
    //TODO adding the direction to the escapeDist
    (*q)[8] =  (*q)[8] + ESCAPEDIST;


    WS4->replaceConf(q);


//    WS4->addAttachement(mainAgent, mainObject, gr, getArmId());
    TSS4->setEndState(WS4);


    WS4->setTSThatEndWithThisWS(TS);
    WS4->setTCForNextTask(_currConstraints);

    TS->setStopWS(WS4);


    TS->addConf(_WS->getRobotConf(mainAgent));
    TS->addConf(approachConf);
    TS->addConf(graspConfOpen);
    TS->addConf(graspConfClosed);
    TS->addConf(escapeConf);


    return TS;

}

bool Pick::setAgents(multimap<string,string> agents_name)
{
    return setMainAgent(agents_name);
}

bool Pick::setObjects(multimap<string,string> objects_name)
{
    return setMainObject(objects_name);
}

bool Pick::setPoints(multimap<string,p3d_point> points)
{
    return true;
}

bool Pick::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
//        cout << "no arm choosen, go for right" << endl;
//        setArmToUse(Right);
        _noArmPreference = true;
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
//        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        _noArmPreference = true;
        return false;
    }

    _noArmPreference = false;
    return true;
}

void Pick::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "approach")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
//            attachObject();
            activateFingertipCol(false, mainAgent, mainObject, NULL);
            tss->refineTraj();
            activateFingertipCol(true, mainAgent, mainObject, NULL);
//            detachObject();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}

string Pick::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if (mainObject)
        s+="\nMain object \t= " + mainObject->getName();
    s+= "\nwith armId \t= " + to_string(getArmId());
    return s;
}

string Pick::getTextFromValues()
{
    return "Task: " + type + " : picking the objects using precomputed grasps"
            + "\n    agents: "
            + "\n        mainAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject: [OBJECT ID]"
            + "\n    datas:"
            + "\n        hand: [right OR left]";
}

bool Pick::getGrasps()
{
    std::list<gpGrasp> gl;
    int te = gpGet_grasp_list(mainObject->getName(), handType, gl);

    if (te != GP_OK)
    {
        SD->setMapEntry("returnReport","no_grasp_list");
        return false;
    }


    cout << "some grasps might cause faillure, the last one is the cause:" << endl;
    for (std::list<gpGrasp>::iterator igrasp = gl.begin(); igrasp != gl.end(); ++igrasp)
    {
        cout << "grasp id = " << igrasp->ID << endl;

        //std::advance(igrasp, i);
        graspLists[_armId].push_back(*igrasp);
    }

    if(graspLists[_armId].size() > 0)
        return true;



    SD->setMapEntry("returnReport","empty_grasp_list");
    return false;

}

void Pick::correctConfigs()
{
    if (mainAgent->getHriAgent()->type == HRI_PR2)
    {
        if (approachConf)
        {
            (*approachConf)[20] = angle_limit_PI((*approachConf)[20]);
            (*approachConf)[22] = angle_limit_PI((*approachConf)[22]);
            (*approachConf)[29] = angle_limit_PI((*approachConf)[29]);
            (*approachConf)[31] = angle_limit_PI((*approachConf)[31]);
        }
        if (graspConfClosed)
        {
            (*graspConfClosed)[20] = angle_limit_PI((*graspConfClosed)[20]);
            (*graspConfClosed)[22] = angle_limit_PI((*graspConfClosed)[22]);
            (*graspConfClosed)[29] = angle_limit_PI((*graspConfClosed)[29]);
            (*graspConfClosed)[31] = angle_limit_PI((*graspConfClosed)[31]);
        }
        if (graspConfOpen)
        {
            (*graspConfOpen)[20] = angle_limit_PI((*graspConfOpen)[20]);
            (*graspConfOpen)[22] = angle_limit_PI((*graspConfOpen)[22]);
            (*graspConfOpen)[29] = angle_limit_PI((*graspConfOpen)[29]);
            (*graspConfOpen)[31] = angle_limit_PI((*graspConfOpen)[31]);
        }
        if (escapeConf)
        {
//            cout <<"lllllllllllllllll" << endl;
//            escapeConf->print(true);
            (*escapeConf)[20] = angle_limit_PI((*escapeConf)[20]);
            (*escapeConf)[22] = angle_limit_PI((*escapeConf)[22]);
            (*escapeConf)[29] = angle_limit_PI((*escapeConf)[29]);
            (*escapeConf)[31] = angle_limit_PI((*escapeConf)[31]);
//            cout <<"ffffffffffffffffff" << endl;
//            escapeConf->print(true);
        }
    }

}


double Pick::computeHeuristicCost()
{
    _WS->load();

    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        p3d_col_deactivate_robot(global_Project->getActiveScene()->getRobot(i)->getRobotStruct());
//        if (global_Project->getActiveScene()->getRobot(i)->getObjectRob())
//            global_Project->getActiveScene()->getRobot(i)->getObjectRob()->setToVirtual(true);
    }
    double di = P3D_HUGE;
    confPtr_t q_curr = confPtr_t(new Configuration(*mainAgent->getCurrentPos()));;
    if (mainAgent
            && mainAgent->getHriAgent()
            && mainAgent->getHriAgent()->type == HRI_PR2
            && mainAgent->getHriAgent()->manip
            && mainObject)
    {
        GeometricComputations GC = GeometricComputations();
        confPtr_t q = confPtr_t(new Configuration(*mainAgent->getCurrentPos()));

        p3d_point center;
        center.x = mainObject->getJoint(0)->getJointStruct()->abs_pos[0][3];
        center.y = mainObject->getJoint(0)->getJointStruct()->abs_pos[1][3];
        center.z = mainObject->getJoint(0)->getJointStruct()->abs_pos[2][3];

        p3d_point base;
        base.x = (*q)[6];
        base.y = (*q)[7];
        base.z = 0;

        p3d_vector3 v, v_norm;
        v[0] = center.x - base.x;
        v[1] = center.y - base.y;
        v[2] = 0;

        p3d_vectNormalize(v,v_norm);

        double theta = acos(v_norm[1]) +M_PI/2;

        double d = mainAgent->getHriAgent()->manip->reach_tolerance - 0.15;
        d =GC.distBetweenPoints2D(center, base) -d;



        p3d_point p;
        p.x = base.x + d* v_norm[0];
        p.y = base.y + d* v_norm[1];
        p.z = base.z + d* v_norm[2];

        (*q)[6] = p.x;
        (*q)[7] = p.y;
        (*q)[11] = theta;
        mainAgent->setAndUpdate(*q);
//        setArmsToNeutralPos(mainAgent);

//        g3d_draw_allwin_active();
//        sleep(3);


        configPt q2 = p3d_get_robot_config(mainAgent->getRobotStruct());

        for (int i = 0; i < mainAgent->getHriAgent()->manip->shoulder_joint_ids_no; i ++)
        {
            HRI_GIK_TASK_TYPE task;
            if (i == 0)
                task = GIK_RATREACH;
            else if (i == 1)
                task = GIK_LATREACH;

            p3d_point clPoint = GC.getCenterOfObj(mainObject);
            p3d_vector3 Tcoord;

            Tcoord[0] = clPoint.x;
            Tcoord[1] = clPoint.y;
            Tcoord[2] = clPoint.z;

            int IKSucceded = hri_agent_single_task_manip_move(mainAgent->getHriAgent(),
                                                              task,
                                                              &Tcoord,
                                                              0.1,
                                                              &q2);

            if (IKSucceded == 1)
            {
                p3d_set_and_update_this_robot_conf(mainAgent->getRobotStruct(),q2);
//                g3d_draw_allwin_active();
//                sleep(3);
                di = p3d_dist_q1_q2(mainAgent->getRobotStruct(),q_curr->getConfigStruct(),q2);
                break;
            }



        }

    }


    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
//        global_Project->getActiveScene()->getRobot(i)->getObjectRob()->setToVirtual(false);

        p3d_col_deactivate_robot(global_Project->getActiveScene()->getRobot(i)->getRobotStruct());
    }

    mainAgent->setAndUpdate(*q_curr);


    return di;
}

double Pick::getSSSizeLeft()
{
    vector<int> va;
    double t = 0;
    if (_noArmPreference)
    {
        t += graspLists[1-_armId].size();
    }


    t += graspLists[_armId].size();
    return t;
}

std::vector<int> Pick::getJointsIdsToUse()
{
    return _planPartToUse.getAllJointIds();
}

void Pick::setPlanPartToUse(Sem::PlanningPart &planPart)
{
    _planPartToUse=planPart;
}

std::pair<Sem::PlanningPart, bool> Pick::getPlanPartToUse()
{
    return std::make_pair(_planPartToUse,true);
}


