#include "pickMultiRrt.hpp"
#include "API/project.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

using namespace std;


//####################### Pick ########################

Pick_MultiRRT::Pick_MultiRRT()
{
    type = "pick_multirrt";
    typeId = pick_multirrt;
    current_grasp_id = -1;
    _armId = -1;
    //_possibleMainAgent.push_back("PR2_ROBOT");
    //_possibleMainAgent.push_back("MAX_ROBOT");
    //_possibleMainAgent.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);
}

bool Pick_MultiRRT::init()
{
    result = new TaskAllSolutions(this);

    approachTraj = NULL;
    graspTraj = NULL;
    escapeTraj = NULL;
    closeHandTraj = NULL;
    return initMotionPlaner(mainAgent);
}

bool Pick_MultiRRT::findCandidateSolutions()
{
    p3d_matrix4 res;
    double Tx, Ty, Tz;
    double Rx, Ry, Rz;
    std::list<gpGrasp> gl;
    bool tr;

    graspConfClosedVect.clear();
    graspConfOpenVect.clear();
    approachConfVect.clear();
    escapeConfVect.clear();

    initConf = confPtr_t(new Configuration(mainAgent, _WS->getRobotConf(mainAgent)->getConfigStruct()));
    p3d_set_new_robot_config(mainAgent->getRobotStruct(), (char *)"init", initConf->getConfigStruct(), NULL, NULL);

    if (mainAgent->getHriAgent()->manip->external_hand_no <= 0)
        return false;

    Robot* hand;
    int armId = -1;
    for(int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
    {
        if (!mainAgent->getHriAgent()->manip->external_hand)
            return false;
        if (!mainAgent->getHriAgent()->manip->external_hand[i])
            return false;
        if(mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
        {
            hand = global_Project->getActiveScene()->getRobotByName(mainAgent->getHriAgent()->manip->external_hand[i]->rob_hand->name);
            armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
            break;
        }
    }
    _armId = armId;
    if (!hand)
        return false;

    confPtr_t qA = hand->getCurrentPos();
    confPtr_t qO = mainObject->getCurrentPos();
    double ItxA = (*qA)[6], ItyA = (*qA)[7],  ItzA = (*qA)[8], IrxA = (*qA)[9], IryA = (*qA)[10], IrzA = (*qA)[11];
    double ItxO = (*qO)[6], ItyO = (*qO)[7],  ItzO = (*qO)[8], IrxO = (*qO)[9], IryO = (*qO)[10], IrzO = (*qO)[11];

    gpGet_grasp_list(mainObject->getName(), handType, gl);

    for (std::list<gpGrasp>::iterator igrasp = gl.begin(); igrasp != gl.end(); ++igrasp)
    {
        cout << "RV -- mainObject name : " << mainObject->getName() << endl;
        p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos, (*igrasp).frame, res);
        p3d_mat4ExtractPosReverseOrder(res, &Tx, &Ty, &Tz,
                                       &Rx, &Ry, &Rz);
        (*qA)[6] = Tx; (*qA)[7] = Ty; (*qA)[8] = Tz;
        (*qA)[9] = Rx; (*qA)[10] = Ry; (*qA)[11] = Rz;
        hand->setAndUpdate(*qA);

        //g3d_draw_allwin_active();
/*
        p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), hand->getRobotStruct());
        if(p3d_col_test_robot(hand->getRobotStruct(),0) == 1) // there is collision
        {
            p3d_print_col_pair();
            cout << "collision detected" << endl;
            continue;
        }
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), hand->getRobotStruct());
*/
        (*qA)[6] = ItxA; (*qA)[7] = ItyA; (*qA)[8] = ItzA; (*qA)[9] = IrxA; (*qA)[10] = IryA; (*qA)[11] = IrzA;
        hand->setAndUpdate(*qA);
        
        p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
        MANIPULATION_TASK_MESSAGE status;// = _MP->getMotionPlanner(mainAgent->getName())->computeManipulationData(armId, mainObject->getRobotStruct(), (*igrasp));
        
        if (status == MANIPULATION_TASK_OK)
        {
            cout << "pick found" <<endl;

//            graspConfClosed = confPtr_t(new Configuration(mainAgent,_MP->getMotionPlanner(mainAgent->getName())->getManipulationData().getGraspConfig()));
//            graspConfOpen = confPtr_t(new Configuration(mainAgent,_MP->getMotionPlanner(mainAgent->getName())->getManipulationData().getOpenConfig()));
//            approachConf = confPtr_t(new Configuration(mainAgent,_MP->getMotionPlanner(mainAgent->getName())->getManipulationData().getApproachFreeConfig()));
//            escapeConf = confPtr_t(new Configuration(mainAgent,_MP->getMotionPlanner(mainAgent->getName())->getManipulationData().getApproachGraspConfig()));

            p3d_set_new_robot_config(mainAgent->getRobotStruct(), (char *)&(igrasp->ID), approachConf->getConfigStruct(), NULL, NULL);

            graspConfClosedVect.push_back(graspConfClosed);
            graspConfOpenVect.push_back(graspConfOpen);
            approachConfVect.push_back(approachConf);
            escapeConfVect.push_back(escapeConf);

            graspList.push_back(*igrasp);
        }        
    }

    if (graspList.size()>0) {
        int status = -1;
        if (compute_traj(initConf, approachConfVect[0], mainObject, status))  // Construct the RRT, we don't care about the trajectory
        {
            cout << "Pick -- graspList size is : " << graspList.size() << endl;
            tr = true;
        }
    }
    else {
        cout << "No grasp found ..." << endl;
        tr = false;
    }

    p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());

    (*qO)[6] = ItxO; (*qO)[7] = ItyO; (*qO)[8] = ItzO; (*qO)[9] = IrxO; (*qO)[10] = IryO; (*qO)[11] = IrzO;
    (*qA)[6] = ItxA; (*qA)[7] = ItyA; (*qA)[8] = ItzA; (*qA)[9] = IrxA; (*qA)[10] = IryA; (*qA)[11] = IrzA;
    mainObject->setAndUpdate(*qO);
    hand->setAndUpdate(*qA);

    return tr;
}

bool Pick_MultiRRT::getRandomSol()
{
    if (graspList.size() > 0)
    {
        current_grasp_id = p3d_random_integer(0, graspList.size()-1);
        if (_alternativeIds.size() == 0)
        {
            _alternativeIds.insert(current_grasp_id);
        } else if (_alternativeIds.size() == graspList.size())
        {
            cout << "ERROR : all alternative grasp found" << endl;
            return false;
        } else
        {
            if (_alternativeIds.find(current_grasp_id) != _alternativeIds.end())
                getRandomSol();
            else
                _alternativeIds.insert(current_grasp_id);
        }
        cout << "Selected grasp id is : " << current_grasp_id << endl;

        mainAgent->getHriAgent()->graspId = graspList.at(current_grasp_id).ID;
        return true;
    }
    return false;
}

bool Pick_MultiRRT::findConfigurations()
{
    if (graspConfClosedVect.size() > 0)
    {
        graspConfClosed = graspConfClosedVect[current_grasp_id];
        graspConfOpen = graspConfOpenVect[current_grasp_id];
        approachConf = approachConfVect[current_grasp_id];
        escapeConf = escapeConfVect[current_grasp_id];
        
        return true;
    }

    return false;
}

bool Pick_MultiRRT::findTrajectories()
{
    approachTraj = p3d_extract_traj(global_Move3DPlanner->trajFound(), global_Move3DPlanner->getActivGraph()->getNumberOfNodes(), global_Move3DPlanner->getActivGraph(),
                                            initConf, approachConf);
    if (!approachTraj)
    {
        cout << "no approach traj" << endl;
        return false;
    }

    int status = -1;
    graspTraj = compute_traj_using_localPath(approachConf, graspConfOpen, mainObject, status);

    if (!approachTraj)
    {
        cout << "no grasp traj" << endl;
        return false;
    }
    attachObject();

    // disable fingerType collisions
//     gpDeactivate_object_fingertips_collisions(p3d_rob *robot, p3d_obj *object, gpHand_properties &hand, int handID)
//    gpHand_properties armHandProp = ( *mainAgent->getRobotStruct()->armManipulationData ) [_armId].getHandProperties();
//    gpDeactivate_object_fingertips_collisions(mainAgent->getRobotStruct(), mainObject->getRobotStruct()->joints[1]->o, armHandProp , _armId);

    //TODO: get support and remove collision only between support and obj
    p3d_col_deactivate_robot(mainObject->getRobotStruct());
    escapeTraj = compute_traj_using_localPath(graspConfOpen, escapeConf, mainObject, status);
    p3d_col_activate_robot(mainObject->getRobotStruct());


    if (!escapeTraj)
        cout << "no escape traj found, the pick is still a success!" << endl;

    closeHandTraj = NULL;

    return true;
}

bool Pick_MultiRRT::computeSolutionTrajectories(TaskSolution *TS)
{
    bool res=true;
    cout << getDescr() << endl;
    if(!TS)
    {
        res = false;
    }

    if(res){
        TS->getStartWS()->load();
        TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);
        TaskSubSolution* TSS1 = TS->getSubSolutionWithId(1);
        TaskSubSolution* TSS3 = TS->getSubSolutionWithId(3);

        confPtr_t q1 = TSS0->getEndState()->getRobotConf(mainAgent);
        confPtr_t q2 = TSS1->getEndState()->getRobotConf(mainAgent);
        confPtr_t q3 = TSS3->getEndState()->getRobotConf(mainAgent);

        approachConf = q1;
        graspConfOpen = q2;
        escapeConf = q3;

        res = findTrajectories();
    }

    if(res){
        TS->getSubSolutionWithId(0)->setTraj(approachTraj);
        TS->getSubSolutionWithId(1)->setTraj(graspTraj);
        TS->getSubSolutionWithId(3)->setTraj(escapeTraj);

        TS->setTrajComputed(true);
    }

    return true;
}

TaskSolution *Pick_MultiRRT::createSolution(bool computeMP)
{
    TaskSolution* TS = new TaskSolution();
    TS->setTrajComputed(computeMP);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    TaskSubSolution* TSS2 = new TaskSubSolution(TS, "engage", 1, graspTraj, mainAgent, 0);
    TaskSubSolution* TSS3 = new TaskSubSolution(TS, "grasp", 2, closeHandTraj, mainAgent, 0);
    TaskSubSolution* TSS4 = new TaskSubSolution(TS, "escape", 3, escapeTraj, mainAgent, 0);

    detachObject();


    vector<TaskSubSolution*> v2;
    v2.push_back(TSS2);
    TSS1->setNext(v2);

    vector<TaskSubSolution*> v3;
    v3.push_back(TSS3);
    TSS2->setNext(v3);

    vector<TaskSubSolution*> v4;
    v4.push_back(TSS4);
    TSS3->setNext(v4);

    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(approachConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    TSS1->setEndState(WS1);
    TSS2->setStartState(WS1);

    WorldState* WS2 = new WorldState(WS1);

    t = WS2->replaceConf(graspConfOpen);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    TSS2->setEndState(WS2);
    TSS3->setStartState(WS2);
    TSS3->setEndState(WS2);
    TSS4->setStartState(WS2);

    WorldState* WS3 = new WorldState(WS2);

    t = WS3->replaceConf(escapeConf);


    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    gpGrasp gr = graspList.at(current_grasp_id);



    //TODO: this is not clean!!
    mainAgent->setAndUpdate(*graspConfOpen);
    double TZi = (*mainAgent->getRobotStruct()->armManipulationData)[_armId].getManipulationJnt()->abs_pos[2][3];
    mainAgent->setAndUpdate(*escapeConf);
    double TZf = (*mainAgent->getRobotStruct()->armManipulationData)[_armId].getManipulationJnt()->abs_pos[2][3];

    confPtr_t q = mainObject->getCurrentPos();
//    q->print();
    (*q)[8] = (*q)[8]+ TZf-TZi;
    WS3->replaceConf(q);
//    q->print();

    WS3->addAttachement(mainAgent, mainObject, gr, _armId);
    TSS4->setEndState(WS3);

    TS->setStopWS(WS3);


    return TS;
}


bool Pick_MultiRRT::setAgents(multimap<string,string> agents_name)
{
    return setMainAgent(agents_name);
}

bool Pick_MultiRRT::setObjects(multimap<string,string> objects_name)
{
    return setMainObject(objects_name);
}

bool Pick_MultiRRT::setPoints(multimap<string,p3d_point> points)
{
    return true;
}

bool Pick_MultiRRT::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
        cout << "no arm choosen, go for right" << endl;
        setArmToUse(Right);
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        cout << "setting default arm to right" << endl;
        setArmToUse(Right);
    }

    return true;
}

void Pick_MultiRRT::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "approach")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
            attachObject();
            tss->refineTraj();
            detachObject();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}

string Pick_MultiRRT::getDescr()
{
    return "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type+ "\nMain agent \t\t= " + mainAgent->getName() +
            "\nMain object \t= " + mainObject->getName() + "\nNb alt \t= " + to_string(result->getNbAlternatives());
}

string Pick_MultiRRT::getTextFromValues()
{
    return "Task: " + type
            + "\n    agents: "
            + "\n        mainAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject: [OBJECT ID]"
            + "\n    datas:"
            + "\n        hand: [right OR left]";
}
