#include "placeMultiRrt.hpp"
#include "API/project.hpp"
#include "API/Roadmap/compco.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

using namespace std;

//####################### Place ########################
Place_MultiRRT::Place_MultiRRT() : Place()
{
    type = "place_multirrt";
    typeId = place_multirrt;
    _armId = -1;

    //_possibleMainAgent.push_back("PR2_ROBOT");
    //_possibleMainAgent.push_back("MAX_ROBOT");
    //_possibleMainAgent.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);
}

bool Place_MultiRRT::findCandidateSolutions()
{
    double Txh, Tyh, Tzh, Rxh, Ryh, Rzh;
    p3d_matrix4 rot;
    gpPlacement placement;
    bool tr;
    unsigned nbrTest = 0;

    GeometricComputations GC = GeometricComputations();

    graspConfClosedVect.clear();
    graspConfOpenVect.clear();
    approachConfVect.clear();
    escapeConfVect.clear();

    _curGrasp = _WS->getCurrentAttachment().grasp;

    Robot *hand;
    for (int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
    {
        if (!mainAgent->getHriAgent()->manip->external_hand)
            return false;
        if (!mainAgent->getHriAgent()->manip->external_hand[i])
            return false;
        if (mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
        {
            hand = global_Project->getActiveScene()->getRobotByName(mainAgent->getHriAgent()->manip->external_hand[i]->rob_hand->name);
            _armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
            break;
        }
    }
    if (!hand)
        return false;

    detachObject();

    confPtr_t qO = mainObject->getCurrentPos();
    confPtr_t qH = hand->getCurrentPos();
    confPtr_t qA = mainAgent->getCurrentPos();
    double ItxO = (*qO)[6], ItyO = (*qO)[7],  ItzO = (*qO)[8], IrxO = (*qO)[9], IryO = (*qO)[10], IrzO = (*qO)[11];
    double ItxH = (*qH)[6], ItyH = (*qH)[7],  ItzH = (*qH)[8], IrxH = (*qH)[9], IryH = (*qH)[10], IrzH = (*qH)[11];

    if (mainAgent->getRobotStruct()->nconf != 0)
        for (int i = 0; i < mainAgent->getRobotStruct()->nconf; i++)
            p3d_del_config(mainAgent->getRobotStruct(), mainAgent->getRobotStruct()->conf[i]);

    initConf = confPtr_t(new Configuration(mainAgent, mainAgent->getCurrentPos()->getConfigStruct()));
    p3d_set_new_robot_config(mainAgent->getRobotStruct(), (char *)"init", initConf->getConfigStruct(), NULL, NULL);

    _supports = _supportObject->getObjectRob()->getSupports();

    unsigned nbrPlacements = ENV.getInt(Env::nbOfSeeds);
    if (nbrPlacements == 0)
        nbrPlacements = 5;

    vector<configPt> rotations;
    if (mainObject->getObjectRob() && mainObject->getObjectRob()->isManipulable())
        rotations = mainObject->getObjectRob()->getStableRotations();
//     = getConfigPlacement(mainObject);
    if (rotations.size() == 0)
    {
        cout << "no rotations defined for the object : " << mainObject->getName() << endl;
        return false;
    }

    while (confSolVect.size() < nbrPlacements && nbrTest < MAX_NBR_TEST)
    {
        nbrTest++;

        int r1 = p3d_random_integer(0,_supports.size()-1);
        GeometricForm* s = _supports.at(r1);

        p3d_point  currentTest = s->getRandomPointInForm();

        configPt q = p3d_copy_config(mainObject->getRobotStruct(), rotations.at( p3d_random_integer(0, rotations.size() - 1) ));

        double randomRot = p3d_random(-M_PI,M_PI);

        q[6] = currentTest.x + q[6];
        q[7] = currentTest.y + q[7];
        q[8] = currentTest.z + q[8];
        if (q[9] == -10)
            q[9] = randomRot;
        else if (q[10] == -10)
            q[10] = randomRot;
        else if (q[11] == -10)
            q[11] = randomRot;

        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(), q);

        vector<p3d_point> vp = GC.getLowerPointsInObject(mainObject, 0.01);

        bool isOnSupport = true;
        for (unsigned int i = 0; i < vp.size(); i++)
        {
            if(!s->isPointInForm(vp.at(i)))
//            if(!GC.point_in_rect2(s.rectangle_corner.at(0),
//                               s.rectangle_corner.at(1),
//                               s.rectangle_corner.at(2),
//                               s.rectangle_corner.at(3),
//                               vp.at(i)))
            {
                isOnSupport = false;
                break;
            }
        }
        if (!isOnSupport)
            continue;

        p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _supportObject->getRobotStruct());
        if(p3d_col_test_robot(mainObject->getRobotStruct(),0) == 1) //there is collision
            continue;
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _supportObject->getRobotStruct());

        p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos, _curGrasp.frame, rot);
        p3d_mat4ExtractPosReverseOrder(rot, &Txh, &Tyh, &Tzh, &Rxh, &Ryh, &Rzh);
        (*qH)[6] = Txh; (*qH)[7] = Tyh; (*qH)[8] = Tzh; (*qH)[9] = Rxh; (*qH)[10] = Ryh; (*qH)[11] = Rzh; (*qH)[12] = 0.2;
        hand->setAndUpdate(*qH);

        //test if hand is in collision with something else
        p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), hand->getRobotStruct());
        bool colTest = p3d_col_test_robot(hand->getRobotStruct(),0) == 1;
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), hand->getRobotStruct());

        if(colTest) // there is collision
        {
            cout << "collision detected" << endl;
            p3d_col_activate_robot_robot(mainObject->getRobotStruct(), hand->getRobotStruct());
            continue;
        }

        confPtr_t confSol = confPtr_t(new Configuration(mainObject, q));

        (*qH)[6] = ItxH; (*qH)[7] = ItyH; (*qH)[8] = ItzH; (*qH)[9] = IrxH; (*qH)[10] = IryH; (*qH)[11] = IrzH;
        hand->setAndUpdate(*qH);
        mainObject->setAndUpdate(*confSol);

        p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
        p3d_col_deactivate_robot_robot(_supportObject->getRobotStruct(), mainObject->getRobotStruct());
        MANIPULATION_TASK_MESSAGE status = ManipPlMap[mainAgent->getName()]->computeManipulationData(_armId, mainObject->getRobotStruct(), _curGrasp);
        if (status == MANIPULATION_TASK_OK)
        {
            cout << "Placement found !" << endl;

            graspConfClosed = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getGraspConfig()));
            graspConfOpen = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getOpenConfig()));
            escapeConf = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getApproachFreeConfig()));
            approachConf = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getApproachGraspConfig()));

            p3d_set_new_robot_config(mainAgent->getRobotStruct(), (char *)&nbrTest, approachConf->getConfigStruct(), NULL, NULL);

            graspConfClosedVect.push_back(graspConfClosed);
            graspConfOpenVect.push_back(graspConfOpen);
            approachConfVect.push_back(approachConf);
            escapeConfVect.push_back(escapeConf);
            confSolVect.push_back(confSol);
        }
        p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
        p3d_col_activate_robot_robot(_supportObject->getRobotStruct(), mainObject->getRobotStruct());

        (*qO)[6] = ItxO; (*qO)[7] = ItyO; (*qO)[8] = ItzO; (*qO)[9] = IrxO; (*qO)[10] = IryO; (*qO)[11] = IrzO;
        mainObject->setAndUpdate(*qO);
        mainAgent->setAndUpdate(*qA);
    }

    attachObject();

    if (confSolVect.size() == 0 || nbrTest == MAX_NBR_TEST)
    {
        cout << "Place -- Unable to find " << nbrPlacements << " placements after " << nbrTest << " tries ..." << endl;
        tr = false;
    } else if (confSolVect.size() >= nbrPlacements)
    {
        cout << "Place -- confSolVect size is : " << confSolVect.size() << endl;
        p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
        int status = -1;
        if (compute_traj(initConf, approachConfVect[0], mainObject, status))  // Construct the RRT, we don't care about the trajectory
            tr = true;
        p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
    } else
    {
        cout << "No placement found ..." << endl;
        tr = false;
    }

    return tr;



/*


    // Find all acceptable placements for mainObject on supportObject
    std::list<gpPlacement> placement_list_full;
    std::list<gpPlacement> placement_list_on_support;
    std::list<p3d_rob*> robot_list;

    gpCompute_stable_placements(mainObject->getRobotStruct(), placement_list_full);
    
    cout << "Place -- placement_list_full size is : " << placement_list_full.size() << endl;

    robot_list.push_back(_supportObject->getRobotStruct());

    gpFind_placements_on_object(mainObject->getRobotStruct(), _supportObject->getRobotStruct(), robot_list, placement_list_full, step, _numOfRot, 0.01, placement_list_on_support);

    cout << "Place -- placement_list_on_support size is : " << placement_list_on_support.size() << endl;

    if (mainAgent->getRobotStruct()->nconf != 0)
        for (int i = 0; i < mainAgent->getRobotStruct()->nconf; i++)
            p3d_del_config(mainAgent->getRobotStruct(), mainAgent->getRobotStruct()->conf[i]);

    initConf = confPtr_t(new Configuration(mainAgent, mainAgent->getCurrentPos()->getConfigStruct()));
    p3d_set_new_robot_config(mainAgent->getRobotStruct(), (char *)"init", initConf->getConfigStruct(), NULL, NULL);

    // Eliminate all placements where robot EEF is in collision with support
    for (std::list<gpPlacement>::iterator iterP = placement_list_on_support.begin(); iterP != placement_list_on_support.end(); ++iterP)
    {
        detachObject();

        iterP->computePoseMatrix(To);
        p3d_mat4ExtractPosReverseOrder(To, &Txo, &Tyo, &Tzo, &Rxo, &Ryo, &Rzo);
        (*qO)[6] = Txo; (*qO)[7] = Tyo; (*qO)[8] = Tzo; (*qO)[9] = Rxo; (*qO)[10] = Ryo; (*qO)[11] = Rzo;
        mainObject->setAndUpdate(*qO);

        //p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos, _curGrasp.frame, Ta);
        p3d_mat4Mult(To, _curGrasp.frame, Ta);
        p3d_mat4ExtractPosReverseOrder(Ta, &Txa, &Tya, &Tza, &Rxa, &Rya, &Rza);
        (*qA)[6] = Txa; (*qA)[7] = Tya; (*qA)[8] = Tza; (*qA)[9] = Rxa; (*qA)[10] = Rya; (*qA)[11] = Rza;
        hand->setAndUpdate(*qA);

        if (p3d_col_test_robot_other(hand->getRobotStruct(), _supportObject->getRobotStruct(), 0) == 1) // there is collision
            continue;

        // Find the first configuration for all the posible configurations and add it as an init node of the multi-RRT

        (*qA)[6] = ItxA; (*qA)[7] = ItyA; (*qA)[8] = ItzA; (*qA)[9] = IrxA; (*qA)[10] = IryA; (*qA)[11] = IrzA;
        hand->setAndUpdate(*qA);

        p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
        p3d_col_deactivate_robot_robot(_supportObject->getRobotStruct(), mainObject->getRobotStruct());
        MANIPULATION_TASK_MESSAGE status = ManipPlMap[mainAgent->getName()]->computeManipulationData(_armId, mainObject->getRobotStruct(), _curGrasp);
        if (status == MANIPULATION_TASK_OK)
        {
            cout << "Placement found !" << endl;

            graspConfClosed = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getGraspConfig()));
            graspConfOpen = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getOpenConfig()));
            escapeConf = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getApproachFreeConfig()));
            approachConf = confPtr_t(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getApproachGraspConfig()));

            p3d_set_new_robot_config(mainAgent->getRobotStruct(), (char *)&(iterP->ID), approachConf->getConfigStruct(), NULL, NULL);

            graspConfClosedVect.push_back(graspConfClosed);
            graspConfOpenVect.push_back(graspConfOpen);
            approachConfVect.push_back(approachConf);
            escapeConfVect.push_back(escapeConf);

            _placementListOnSupport.push_back(*iterP);
        }
        p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
        p3d_col_activate_robot_robot(_supportObject->getRobotStruct(), mainObject->getRobotStruct());

        attachObject();
    }

    p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());

    if (_placementListOnSupport.size()>0) {
        cout << "Place -- _placementListOnSupport size is : " << _placementListOnSupport.size() << endl;
        if (compute_traj(initConf, approachConfVect[0], mainObject))  // Construct the RRT, we don't care about the trajectory
            tr = true;
    }
    else {
        cout << "No placement found ..." << endl;
        tr = false;
    }

    p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());

    (*qO)[6] = ItxO; (*qO)[7] = ItyO; (*qO)[8] = ItzO; (*qO)[9] = IrxO; (*qO)[10] = IryO; (*qO)[11] = IrzO;
    (*qA)[6] = ItxA; (*qA)[7] = ItyA; (*qA)[8] = ItzA; (*qA)[9] = IrxA; (*qA)[10] = IryA; (*qA)[11] = IrzA;
    mainObject->setAndUpdate(*qO);
    hand->setAndUpdate(*qA);

    return tr;
*/
}

bool Place_MultiRRT::getRandomSol()
{
    _placementIndex = p3d_random_integer(0, confSolVect.size() - 1);
    if (_alternativeIds.size() == 0)
    {
        _alternativeIds.insert(_placementIndex);
    } else if (_alternativeIds.size() == confSolVect.size())
    {
        cout << "ERROR : all alternative placement found" << endl;
        return false;
    } else
    {
        if (_alternativeIds.find(_placementIndex) != _alternativeIds.end())
            getRandomSol();
        else
            _alternativeIds.insert(_placementIndex);
    }

    cout << "Placement id is : " << _placementIndex << endl;
    return true;

}

bool Place_MultiRRT::findConfigurations()
{
    if (mainAgent->getHriAgent()->manip->external_hand_no <= 0)
        return false;

    graspConfClosed = graspConfClosedVect[_placementIndex];
    graspConfOpen = graspConfOpenVect[_placementIndex];
    escapeConf = escapeConfVect[_placementIndex];
    approachConf = approachConfVect[_placementIndex];

    _confSol = confSolVect[_placementIndex];

    return true;
}

bool Place_MultiRRT::findTrajectories()
{
    gpHand_properties armHandProp = ( *mainAgent->getRobotStruct()->armManipulationData ) [_armId].getHandProperties();
    gpDeactivate_object_fingertips_collisions(mainAgent->getRobotStruct(), mainObject->getRobotStruct()->joints[1]->o, armHandProp , _armId);

    attachObject();

    cout << "Place -- number of nodes in graph from global_Move3DPlanner : " << global_Move3DPlanner->getActivGraph()->getNumberOfNodes() << ", name of graph : " << global_Move3DPlanner->getActivGraph()->getName() << endl;

    approachTraj = p3d_extract_traj(global_Move3DPlanner->trajFound(), global_Move3DPlanner->getActivGraph()->getNumberOfNodes(), global_Move3DPlanner->getActivGraph(),
                                            initConf, approachConf);

    if (!approachTraj)
    {
        cout << "no approach traj" << endl;
        return false;
    }

    p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
    int status = -1;
    placeTraj = compute_traj_using_localPath(approachConf, graspConfClosed, mainObject, status);

    if (!placeTraj)
    {
        cout << "no place traj" << endl;
        return false;
    }
    p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
    detachObject();

    escapeTraj = compute_traj_using_localPath(graspConfClosed, escapeConf, mainObject, status);

    if (!escapeTraj)
    {
        cout << "no escape traj" << endl;
        return false;
    }

    releaseTraj = NULL;

    return true;
}
