
#include "API/HPP_Wrap/Romeo_Wrap.hpp"
#include "task.hpp"
#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"
#include "P3d-pkg.h"


#include "API/project.hpp"


#include "planner/plannerFunctions.hpp"
#include "plannerFunctions.hpp"
#include "planner/cost_space.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif


#include "database/semantics/Sem_SemanticsManager.hpp"

using namespace std;

#ifdef HPP
RomeoWrap *_romeoWrap;
#endif



#include <iostream>
#include <map>

Task::Task()
{
    _taskGraph = NULL;
    _planTree = NULL;
    result = NULL;
    SD = new SolutionData();
    SD->setMapEntry("returnReport", "ERROR");
    SD->getSDMapEntry("times")->setMapEntry("timeMp", 0.0);
}

int Task::getId()
{
    return task_id;
}

void Task::setId(int id)
{
    task_id = id;
}



bool Task::run(multimap<string,string> agents_name,
               multimap<string,string> objects_name,
               multimap<string,p3d_point> point3d,
               multimap<string,string> additionalData,
               WorldState* WS,bool computeMP)
{
    WS->load();
    _WS = WS;
    initAll();
    if (!setAgents(agents_name))
        return false;
    if (!setObjects(objects_name))
        return false;
    if (!setPoints(point3d))
        return false;
    if (!setdata(additionalData))
        return false;
    if (!findConfigurations())
        return false;
    if (computeMP)
        if(!findTrajectories())
            return false;

    return true;
}

double Task::heuristic(multimap<string,string> agents_name,
                       multimap<string,string> objects_name,
                       multimap<string,p3d_point> point3d,
                       multimap<string,string> additionalData,
                       WorldState* WS, bool computeMP)
{
    WS->load();
    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setObjects(objects_name))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }


    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    double res =  computeHeuristicCost();

    if (!finalize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    return res;
}

double Task::computeHeuristicCost()
{
    return P3D_HUGE;
}

p3d_traj* Task::compute_traj(confPtr_t q_init, confPtr_t q_goal, int &status)
{
    if (q_init->getRobot()->getName() != q_goal->getRobot()->getName())
        return false;

    //this should be implemented in the near future
    return false;
}

p3d_traj* Task::compute_traj_using_localPath(confPtr_t q_init, confPtr_t q_goal, int& status)
{
    return false;
}

bool Task::computeManipulationConfs(confPtr_t& , confPtr_t& , confPtr_t& , confPtr_t& , p3d_matrix4 , gpGrasp,
                                      p3d_vector3 , double , p3d_vector3 , double )
{
    cout << "wrong fct" << endl;
    return true;
}

bool Task::computeInverseKinematic(Robot* ag, p3d_matrix4 target_pos, confPtr_t &solution, int armId, bool fromCurPos)
{
    cout << "wrong fct" << endl;
    return true;
}

TaskSubSolution* Task::getSubSolAt(int alternativeId, int id)
{
    TaskSolution* ts = result->getAlternativeWithId(alternativeId);
    TaskSubSolution* tSubSol = NULL;
    if (!ts)
        return tSubSol;
    return ts->getSubSolutionWithId(id);
}

int Task::getSubSolNumberOfTrajs(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if (!TS)
        return 0;

    int nbRes = 0;
    vector<Robot*> vr = TS->getInvolvedRobots();
    for (unsigned int i = 0; i<vr.size(); i++)
    {
        nbRes += TS->getNbsubSolPerRobot(vr.at(i));
    }
    return nbRes;
}

double Task::getSSSizeLeft()
{
    return 0;
}

vector<SurfaceAttachement> Task::createRecursiveAttachements(Robot* r)
{
    vector<SurfaceAttachement> sAttList;
    WorldStateFacts* WSF = new WorldStateFacts(global_Project->getActiveScene());
    global_Project->getActiveScene()->getFactsManager()->computeAllIsOn(WSF,true);
    vector<Fact> ff = WSF->getFacts();
    vector<Robot*> openList;
    openList.push_back(r);

    while (openList.size() > 0)
    {
        Robot* tmp = openList.at(openList.size()-1);
        openList.pop_back();
        for(unsigned int i = 0; i < ff.size(); i++)
        {
            Fact f = ff.at(i);

            if (f.getTargetEntity()->getName() == tmp->getName() &&
                    f.getType().factType == "IsOn" &&
                    f.getType().factSubType == "" &&
                    f.getBoolValue())
            {
//                cout << "detected that " << f.getMainEntity()->getName() << " is On " << f.getTargetEntity()->getName() << endl;
                openList.push_back(f.getMainEntity());
                //create attachement
                SurfaceAttachement sa;
                sa.support = tmp;
                sa.attache = f.getMainEntity();
                sa.jointId = 0; //for now the attachement can be done on joint 0 only
                p3d_matrix4 m,sInv;
                p3d_matInvertXform(tmp->getJoint(0)->getJointStruct()->abs_pos,sInv);
                p3d_mat4Mult(f.getMainEntity()->getJoint(0)->getJointStruct()->abs_pos,sInv,m);
                p3d_mat4Copy(m,sa.transform);
                sAttList.push_back(sa);
            }
        }
    }

//    cout << " astt size = " << sAttList.size() << endl;
    return sAttList;
}

vector<SurfaceAttachement> Task::createAndLoadSurfaceAttachements(Robot* r)
{
    vector<SurfaceAttachement> sAttList = createRecursiveAttachements(r);
    removeSurfaceAttachements(sAttList);


    for(unsigned int i = 0; i < sAttList.size(); i++)
    {
        SurfaceAttachement a = sAttList.at(i);
//        if (!a.support->isObjectAttached(a.attache,a.jointId+1))
//        {
            //TODO: place by the matrix
//            cout << "attach " << a.attache->getName() << " to " << a.support->getName() << endl;
            a.support->attachObj(a.attache,a.jointId+1);
//        }
    }

    return sAttList;
}


void Task::removeSurfaceAttachements(vector<SurfaceAttachement> s)
{
    for (unsigned int i = 0; i < s.size(); i++)
    {
        SurfaceAttachement a = s.at(i);
        if (a.support->isObjectAttached(a.attache,a.jointId+1))
        {
            a.support->dettachObj(a.attache);
        }
    }
}
