#include "HRICS_BodiesDistance.h"
#include "HRICS_costfunction.h"

using namespace std;
namespace HRICS {
namespace CF {

BodiesDistance::BodiesDistance(CostFunctionCommon *common, double min, double max, double low, double up, double maxToMeanRatio, FunctionShape shape, bool useBoundingBoxes):
    VirtualCostFunction(common,min,max,low,up,maxToMeanRatio,shape),
    _useBoundingBoxes(useBoundingBoxes)
{

}

BodiesDistance::~BodiesDistance()
{

}

double BodiesDistance::getCost()
{
    double cost(0),weights(0);
    double mindist(lowerBound());
    Robot *human,*rob;
    rob=_common->getRobot();
    for(unsigned int h=0;h<_common->getHumans().size();h++){
        //std::map<HRI_AGENT_TYPE,std::vector<std::pair< int,double> > >::const_iterator humanweights,rweights;
        human=_common->getHumans()[h];
        const std::vector<std::pair<int,double> > &humanweights = _common->bodyWeightsCurrentCost(human->getHriAgent()->type);
        const std::vector<std::pair<int,double> > &rweights = _common->bodyWeightsCurrentCost(rob->getHriAgent()->type);
        for(vector<pair<int,double> >::const_iterator hb=humanweights.begin();
            hb!=humanweights.end();hb++){
            double weight(hb->second);
            for(vector<pair<int,double> >::const_iterator rb=rweights.begin();
                rb!=rweights.end() && weight ;rb++) // dont loop if weight is nul
            {
                if(rb->second !=0){
                    weight *= rb->second;
                    double dist(0);
                    p3d_obj *o1,*o2;
                    o1=human->getRobotStruct()->o[hb->first];
                    o2=rob->getRobotStruct()->o[rb->first];
                    double dummy(0);
                    dist=p3d_BB_obj_obj_extern_dist(o1,o2,&dummy);
                    if (dist <= lowerBound()){
                            //if AABB are closer than the outer limit, compute real distance btw obj
                        if(useBoundingBoxes()){
                            p3d_vector3 v1,v2;
                            dist = pqp_distance_OBB(o1,o2,v1,v2);
                        }else{
                            dist=p3d_distanceObjToObj(o1,o2);
                        }
                    }
                    mindist = std::min(dist,mindist);
                    cost += weight * getCost(dist);
                    weights += weight;
                }
            }
        }
    }
    assert(_maxToMeanRatio <=1);
    return _maxToMeanRatio * getCost(mindist) + (1-_maxToMeanRatio) * cost/weights;
}


std::pair<double,double> BodiesDistance::getCost(Robot *r1, unsigned int o1, Robot *r2, unsigned int o2)
{
    double cost(0);
    //std::map<HRI_AGENT_TYPE,std::map<std::string,double> >::const_iterator r1weights,r2weights;
    double w1(0),w2(0),dist(0);
    if(r1->getRobotStruct()->o[o1]->pqpModel != NULL && r2->getRobotStruct()->o[o2]->pqpModel != NULL){
        const std::map<std::string,double> &r1weights = _common->bodyWeightsByNameCurrentCost(r1->getHriAgent()->type);
        const std::map<std::string,double> &r2weights = _common->bodyWeightsByNameCurrentCost(r2->getHriAgent()->type);
        p3d_obj *obj1(r1->getRobotStruct()->o[o1]),*obj2(r2->getRobotStruct()->o[o2]);
        std::string name1(obj1->name),name2(obj2->name);
        std::map<std::string,double>::const_iterator itw1 = r1weights.find(name1.substr(name1.find_first_of('.')+1));
        std::map<std::string,double>::const_iterator itw2 = r2weights.find(name2.substr(name2.find_first_of('.')+1));
        if(itw1 != r1weights.end() && itw2 != r2weights.end()){
            w1 = itw1->second;
            w2 = itw2->second;
        }
        if(useBoundingBoxes()){
            double dummy(0);
            dist = p3d_BB_obj_obj_extern_dist(r1->getRobotStruct()->o[o1],r2->getRobotStruct()->o[o2],&dummy);
        }else{
            dist = p3d_distanceObjToObj(r1->getRobotStruct()->o[o1],r2->getRobotStruct()->o[o2]);
        }
        cost = getCost(dist);
    }
    return std::make_pair(cost,w1*w2);
}


bool BodiesDistance::useBoundingBoxes() const
{
    return _useBoundingBoxes;
}

void BodiesDistance::setUseBoundingBoxes(bool useBoundingBoxes)
{
    _useBoundingBoxes = useBoundingBoxes;
}






} // namespace CF
} // namespace HRICS

