#include "HRICS_CF_KeepItLow.h"
#include "API/planningAPI.hpp"
#include "HRICS_costfunction.h"

namespace HRICS {
namespace CF {

INIT_MOVE3D_STATIC_LOGGER(KeepItLow,"planners.hrics.cf.keepitlow");

KeepItLow::KeepItLow(CostFunctionCommon *commonCF,
                     double min, double max,
                     double low, double up,
                     double maxToMeanRatio, FunctionShape shape):
    VirtualCostFunction(commonCF, min,max,low,up,maxToMeanRatio,shape)
{

}

KeepItLow::~KeepItLow()
{
}

double KeepItLow::getCost()
{
    Robot* r=commonCF()->getRobot();
    HRI_AGENT_TYPE t=r->getHriAgent()->type;
    double max_cost(-std::numeric_limits<double>::infinity()),sum_cost(0),mean_cost(0);
    double weight_sum(0);
    try{
        std::map<int,double> jointWeights = commonCF()->jointWeightsMovingCurrentCost(t);
        for (std::map<int,double>::const_iterator it=jointWeights.begin();it!=jointWeights.end();++it){
            if(it->second!=0){
                Joint *j;
                j=r->getJoint(it->first);
                double z=j->getVectorPos()[2];
                double cost= it->second*getCost(z);
                max_cost=std::max(cost,max_cost);
                weight_sum+=it->second;
                sum_cost+=cost;
            }
        }
        if (weight_sum==0){
            M3D_WARN("no joints specified to compute cost");
        }
        mean_cost = sum_cost / weight_sum;
        return getCostMaxMean(max_cost,mean_cost);
    }catch(std::runtime_error e){
        M3D_ERROR(e.what());
    }
    return _maxCost;
}


} // namespace CF
} // namespace HRICS

