#include "HRICS_CF_KeepManipulatedVertical.h"
#include "HRICS_costfunction.h"

#include "API/planningAPI.hpp"
#include "utils/Geometry.h"

namespace HRICS {
namespace CF {

KeepManipulatedVertical::KeepManipulatedVertical(CostFunctionCommon *commonCF, double min, double max, double low, double up, double maxToMeanRatio, FunctionShape shape):
    VirtualCostFunction(commonCF,min, max,low, up, maxToMeanRatio, shape)
{

}

KeepManipulatedVertical::~KeepManipulatedVertical()
{

}

double KeepManipulatedVertical::getCost()
{
    std::vector<Robot *> objects = getObjects();
    double c_max(0), c_mean(0);
    unsigned int n(0);
    for(std::vector<Robot*>::iterator it = objects.begin() ; it!=objects.end() ; ++it){
        Robot *obj = *it;
        ObjectRob *oobj = obj->getObjectRob();
        bool ok=true;
        if(!oobj) ok = false;
        if(ok){
            ok = oobj->hasVerticalityConstraint();
        }
        if(ok){
            Eigen::Vector3d v = oobj->keepVerticalAxis();
            v.normalize();
            Eigen::Affine3d tf = m3dGeometry::p3dToEigen(*obj->getJoint(0)->getAbsPos());
            Eigen::Vector3d v2 = tf.linear() * v;
            assert(v2[2] <=1);
            assert(v2[2] >=-1);
            double  zpz = v2[2];
            double a = acos(zpz); // in [0, PI]
            double a_norm = (a - oobj->maxVerticalTolerance() ) / 0.2;
            double c = getCost(a_norm);
            c_max = std::max(c,c_max);
            c_mean += c;
            n++;
        }
        //p3d_matrix4 pos = obj->baseJnt->abs_pos;
        // Eigen::Affine3d tf = m3dGeometry::p3dToEigen(obj->joints[1]->abs_pos);
        // EigEigen::Vector3d z = Eigen::Vector3d::UnitZ();
        // EigEigen::Vector3d z_p = tf * z;
        // double a = std::acos(z_p.z());
        // double zpz = obj->getJoint(0)->getAbsPos()[2][2] ;
        // double a = acos(zpz) ;
        // c = getCostSigmoid(a);
    }
    if(n) {
        c_mean = c_mean / n;
    }

    return getCostMaxMean(c_max,c_mean);

}

std::vector<Robot *> KeepManipulatedVertical::getObjects() {
    std::vector<Robot*> v;
    p3d_rob *r = this->commonCF()->getRobot()->getRobotStruct();
    grasp_struct *gp=r->grasped_objects;
    while (gp){
        v.push_back(global_Project->getActiveScene()->getRobot(gp->objectPt));
        gp=gp->next;
    }
    return v;
}


} // namespace CF
} // namespace HRICS

