#include "HRICS_CF_PathLengthCost.h"
#include "HRICS_costfunction.h"

using namespace std;
namespace HRICS {
namespace CF {

PathLengthCost::PathLengthCost(CostFunctionCommon *commonCF, double min, double max, double low, double up, double maxToMeanRatio, FunctionShape shape):
    VirtualCostFunction(commonCF,min,max,low,up,maxToMeanRatio,shape)
{

}

PathLengthCost::~PathLengthCost()
{

}

double PathLengthCost::getCost(std::vector<LocalPath *> &pathes)
{
    API::Trajectory path(commonCF()->getRobot());
    for(std::vector<LocalPath*>::iterator it = pathes.begin(); it!=pathes.end(); ++it){
        path.push_back(std::tr1::shared_ptr<LocalPath>(new LocalPath(**it)));
    }
    return getCost(path);

}

double PathLengthCost::getCost(API::Trajectory &path){
    double len= getTotalLength(path);
    return getCost(len);
}

double PathLengthCost::getTotalLength(API::Trajectory &path)
{
    const std::map<int,double> &jointWeights =commonCF()->jointWeightsCurrentCost(commonCF()->getRobot()->getHriAgent()->type);
    std::vector<int> joints;
    for(std::map<int,double>::const_iterator it = jointWeights.begin(); it != jointWeights.end(); ++it){
        if(it->second >0){
            joints.push_back(it->first);
        }
    }

    return getLength(path,joints);

}

double PathLengthCost::getLength(API::Trajectory &path, std::vector<int> &jointIndices)
{
    double pathLength = path.getRangeMax();
    double stepSize = min(p3d_get_env_dmax(), pathLength);
    double param = stepSize;
    double distance(0);
    if(pathLength==0){return 0;}
    Robot *r = commonCF()->getRobot();
    Configuration initConf = *r->getCurrentPos();
    confPtr_t previousConf,currentConf;
    std::vector<Eigen::Vector3d> previousPos,currentPos;
    previousConf = path.getBegin();
    r->setAndUpdate(*previousConf);
    for(std::vector<int>::iterator it=jointIndices.begin() ; it!=jointIndices.end(); ++it){
        previousPos.push_back(r->getJoint(*it)->getVectorPos());
    }

    while (param <= pathLength)
    {
        currentConf = path.configAtParam(param);
        r->setAndUpdate(*currentConf);
        currentPos.clear();
        for(unsigned int i=0;i<jointIndices.size(); i++){
            currentPos.push_back(r->getJoint(jointIndices[i])->getVectorPos());
            distance += (currentPos[i] - previousPos[i]).norm();
        }

        previousPos.swap(currentPos);
        previousConf = currentConf;
        if (param == pathLength)
            param = pathLength + 1.;
        else if (pathLength - param > stepSize + 1e-3)
            param += stepSize;
        else {
            stepSize = pathLength - param;
            param = pathLength;
        }
    }

    r->setAndUpdate(initConf);
    return distance;
}

void PathLengthCost::setMaxLength(API::Trajectory &path)
{
    setUpperBound(getTotalLength(path));
}





} // namespace CF
} // namespace HRICS

