#include "HRICS_CF_TorsoHeight.h"
#include "HRICS_costfunction.h"

using namespace std;

namespace HRICS {
namespace CF {

INIT_MOVE3D_STATIC_LOGGER(TorsoHeight,"planners.hrics.cf.torsoheight");

TorsoHeight::TorsoHeight(CostFunctionCommon *commonCF, double min, double max, double low, double up, double maxToMeanRatio, FunctionShape shape):
    VirtualCostFunction(commonCF,min,max,low,up,maxToMeanRatio,shape)
{
    _keepItLow= new KeepItLow(commonCF,min,max,low,up,maxToMeanRatio,shape);
    this->onHumanMove();
}

TorsoHeight::~TorsoHeight()
{

}

double TorsoHeight::getCost()
{
    return _keepItLow->getCost();
}

void TorsoHeight::findJoints()
{
    _shoulders.clear();
    _pelvises.clear();
    for (std::vector<Robot*>::const_iterator it=this->commonCF()->getHumans().begin();it!=commonCF()->getHumans().end();it++){
        Robot *r=*it;
        HRI_MANIP *manip = r->getHriAgent()->manip;
        for (int i=0;i<manip->shoulder_joint_ids_no;i++){
            _shoulders.push_back(r->getJoint(manip->shoulder_joint_ids[i]));
        }
        M3D_DEBUG("pelvis?: "<<r->getJoint(0)->getName());
        _pelvises.push_back(r->getJoint(0));
    }
}

void TorsoHeight::computeLawParam()
{
    double highest_pelvis(-std::numeric_limits<double>::infinity());
    double lowest_shoulder(std::numeric_limits<double>::infinity());
    for (std::vector<Robot*>::const_iterator it=this->commonCF()->getHumans().begin();it!=commonCF()->getHumans().end();it++){
        Robot *r=*it;
        HRI_MANIP *manip = r->getHriAgent()->manip;
        for (int i=0;i<manip->shoulder_joint_ids_no;i++){
            double h = r->getJoint(manip->shoulder_joint_ids[i])->getVectorPos()[2];
            lowest_shoulder = std::min(lowest_shoulder,h);
            M3D_DEBUG("pelvis?: "<<r->getJoint(0)->getName());

            h = r->getJoint(0)->getVectorPos()[2];
            highest_pelvis = std::max(highest_pelvis,h);
        }
    }

    setUpperBound(lowest_shoulder);
    setLowerBound(highest_pelvis);
    M3D_DEBUG("lowest_shoulder = " << lowest_shoulder<<" cost="<<_keepItLow->getCost(lowest_shoulder));
    M3D_DEBUG("highest_pelvis = " << highest_pelvis<<" cost="<<_keepItLow->getCost(highest_pelvis));
    M3D_DEBUG("middle = " << (lowest_shoulder+highest_pelvis)/2<<" cost="<<_keepItLow->getCost((highest_pelvis+lowest_shoulder)/2));

}

void TorsoHeight::onHumanMove()
{
    computeLawParam();
}

void TorsoHeight::setLowerBound(double lowerBound)
{
    _keepItLow->setLowerBound(lowerBound);
}

double TorsoHeight::loweBound() const
{
    return _keepItLow->lowerBound();
}

void TorsoHeight::setUpperBound(double upperBound)
{
    _keepItLow->setUpperBound(upperBound);
}

double TorsoHeight::upperBound() const
{
    return _keepItLow->upperBound();
}

string TorsoHeight::getName()
{
    return "TorsoHeight";
}

} // namespace CF
} // namespace HRICS

