#include "socketplot.h"

INIT_MOVE3D_STATIC_LOGGER(SocketPlot,"planners.logger.socketplot");

std::string SocketPlot::__name_prefix = std::string("planners.plot.");

SocketPlot::SocketPlot(std::string name):
    _index(0),_bufferMaxSize(10)
{
    std::string n = __name_prefix;
    n=n.append(name);
    _logger=logm3d::getLogger(n);
    this->sendEmpty();
}

SocketPlot::~SocketPlot()
{

}

void SocketPlot::addPoint(double x, double y)
{
    std::vector<double> v;
    v.push_back(x);
    v.push_back(y);
    addPoint(v);
}

void SocketPlot::addPoint(std::vector<double> p)
{
    _buffer.push_back(p);
    if(_buffer.size() >= bufferMaxSize()){
        this->send();
    }
}

void SocketPlot::addIndexedValue(double val)
{
    std::vector<double> v(2);
    v[0]=_index;
    v[1]=val;
    _index++;
    this->addPoint(v);
}

std::string SocketPlot::vectorToString(std::vector<double> v){
    std::ostringstream ss;
    ss<<'[';
    for (unsigned int i=0;i<v.size()-1;i++){
        ss<<v[i]<<",";
    }
    ss<<v.back();
    ss<<']';
    return ss.str();
}

void SocketPlot::send()
{
    while(_buffer.size()){
        std::vector<double> v = _buffer.front();
        _buffer.pop_front();

        LOGM3D_TRACE(_logger,"['"<<this->_logger->getName()<<"',"<<vectorToString(v)<<"];");
    }
}

void SocketPlot::sendEmpty()
{
    LOGM3D_TRACE(_logger,"['"<<this->_logger->getName()<<"',[]];;");
}
unsigned int SocketPlot::bufferMaxSize() const
{
    return _bufferMaxSize;
}

void SocketPlot::setBufferMaxSize(unsigned int bufferMaxSize)
{
    _bufferMaxSize = bufferMaxSize;
}

