#include "Sem_PlanParts.hpp"

#include "Sem_AgentType.hpp"
#include "LightPlanner-pkg.h"

#include "API/planningAPI.hpp"
#include <fstream>



namespace Sem {
using namespace std;

INIT_MOVE3D_STATIC_LOGGER(PlanParts,"planners.semantics.planparts");
INIT_MOVE3D_STATIC_LOGGER(PlanningPart,"planners.semantics.jointsset");

PlanParts::PlanParts(std::string global_json_path, bool proactive):
    _proactive(proactive),_global_json_path(global_json_path)
{
    setProactive(proactive);


}

PlanParts::~PlanParts()
{
    for (std::map<AgentType,std::map<PlanningType,std::vector<PlanningPart*> > >::iterator it = _planParts.begin(); it!=_planParts.end();++it){\
        for (std::map<PlanningType,std::vector<PlanningPart*> >::iterator it2=it->second.begin(); it2!=it->second.end();it2++){
            for (std::vector<PlanningPart*>::iterator it3=it2->second.begin(); it3!=it2->second.end();it3++){
                if (*it3)
                    delete *it3;
            }
            it2->second.clear();
        }
        it->second.clear();
    }
    _planParts.clear();

}

bool PlanParts::fetchData()
{
    bool ok=true;
    unsigned int rob_n=global_Project->getActiveScene()->getNumberOfRobots();
    std::set<AgentType> types;
    for (unsigned int i=0;i<rob_n;i++){
        Robot *r = global_Project->getActiveScene()->getRobot(i);
        if(r->getHriAgent()){
            types.insert(r->getHriAgent()->type);
        }
    }

    for (std::set<AgentType>::iterator it=types.begin();it!=types.end();++it){
        //erase and update/read all joints sets for this agent type
        clearData(*it);
        ok = ok && readJsonAgent(*it);
    }
    return ok;
}

void PlanParts::clearData(AgentType t)
{
    std::map<AgentType,std::map<PlanningType, std::vector<PlanningPart*> > >::iterator it=_planParts.find(t);
    if (it!=_planParts.end()){
        for (std::map<PlanningType,std::vector<PlanningPart*> >::iterator it2=it->second.begin(); it2!=it->second.end();it2++){
            for (std::vector<PlanningPart*>::iterator it3=it2->second.begin(); it3!=it2->second.end();it3++){
                if (*it3)
                    delete *it3;
            }
            it2->second.clear();
        }
        it->second.clear();
    }
}

bool PlanParts::importJsonFile(string path)
{
    Json::Value root;
    return getJsonRoot(root,path) && readJsonAgent(root,path);
}

bool PlanParts::createFromActiveJoints(PlanningType plan_type)
{
    bool ok(false);
    PlanningPart *jset =PlanningPart::createFromActiveJoints(plan_type);
    if (jset){
        ok=true;
        this->addPlanPart(jset);
    }
    return ok;

}

bool PlanParts::getJsonRoot(Json::Value &root,std::string &file_path){
    bool ok=false;
    std::ifstream ifs;
    AgentType hri_agn_type;
    if(file_path.size()){
        ifs.open(file_path.c_str());
        ok=(ifs);
    }else{
        ok=false;
    }
    if(ok){
        try{
        ifs>>root;
        }catch (std::runtime_error e){
            M3D_ERROR("error while parsing file "<<file_path<<": "<<e.what());
            ok=false;
        }catch (std::exception e){
            M3D_ERROR("error while parsing file "<<file_path<<": "<<e.what());
            ok=false;
        }
        ifs.close();
    }
    if(ok){
        std::string type = root["agent_type"].asString();
        hri_agn_type=AgentType(type);
        if (!hri_agn_type.isDefined()){
            M3D_WARN("unknown agent type found in "<<file_path<<" : "<<type);
            ok=false;
        }
        //if(ok && hri_agn_type != t){
        //    M3D_WARN("in file "<<file_path<<", agent_type (\""<<type<<"\") differs from the one in filename");
        //    ok=false;
        //}
    }
    if(!ok){
        root.clear();
    }

    return ok;
}


bool PlanParts::readJsonAgent(AgentType t)
{
    bool ok=false;
    Json::Value root;
    std::string file=getPath(t);
    ok=getJsonRoot(root,file);
    if (ok){
        ok=readJsonAgent(root,file);
    }
    return ok;
}

bool PlanParts::readJsonAgent(Json::Value &root,const std::string &file){
    bool ok =true;
    Json::Value parts_dict = root["planning_parts"];
    AgentType t;
    t = AgentType(root["agent_type"].asString());
    ok=t.isDefined();
    if(ok){
        for(Json::Value::iterator it= parts_dict.begin(); it != parts_dict.end() ; ++it){
            std::string plantype_s=it.key().asString();
            for(Json::Value::iterator it2 = (*it).begin(); it2 != (*it).end(); it2++){
                if(!parsePlanPart(t,plantype_s,*it2)){
                    M3D_WARN("error occured during parsing of file "<<file);
                    ok=false;
                }
            }
        }
    }
    return ok;

}

bool PlanParts::readJsonPlanType(AgentType agn_type, PlanningType plan_type)
{
    bool ok=false;
    Json::Value root;
    std::string file=getPath(agn_type);
    ok=getJsonRoot(root,file);
    if(ok){
        ok=parsePlanPart(agn_type,plan_type,root["planning_parts"][plan_type.toString()]);
    }
    return ok;

}

bool PlanParts::parsePlanPart(AgentType agn_type, const string &plan_type_str, const Json::Value &joint_set_json)
{
    PlanningType plantype(plan_type_str);
    //plantype=stringToPlanningType(plan_type_str);
    //bool ok = (plantype!=SEM_PLANNIG_TYPE_SIZE);
    bool ok(true);
    if(!ok){
        M3D_WARN("unknown plan type \""<<plan_type_str<<"\"");
    }else{
        ok = parsePlanPart(agn_type,plantype,joint_set_json);
    }
    return ok;
}

bool PlanParts::parsePlanPart(AgentType agn_type, PlanningType plantype, const Json::Value &joint_set_json)
{
    std::vector<int> arm_ids, joint_ids;
    bool ok=true;
    //full arms
    if (joint_set_json.isMember("arm_id")){
        arm_ids.push_back(joint_set_json["arm_id"].asInt());
    }else if (joint_set_json.isMember("arm_ids")){
        for (Json::Value::iterator it2= joint_set_json["arm_ids"].begin(); it2!=joint_set_json["arm_ids"].end();++it2){
            arm_ids.push_back((*it2).asInt());
        }
    }

    //joints
    if(joint_set_json.isMember("joints") && joint_set_json["joints"].isArray()){
        for (Json::Value::iterator itj=joint_set_json["joints"].begin(); itj!=joint_set_json["joints"].end();++itj){
            joint_ids.push_back((*itj).asInt());
        }
    }
    if(ok)
        this->addPlanPart(agn_type,plantype,joint_ids,arm_ids);

    return ok;

}

void PlanParts::save()
{
    for(std::map<AgentType,std::map<PlanningType,std::vector<PlanningPart*> > >::const_iterator it = _planParts.begin();it!=_planParts.end();++it){
        M3D_TRACE("saving for agent type "<<it->first.toString());
        writeJson(it->first);
    }
}

void PlanParts::writeJson(AgentType t)
{
    using namespace Json;
    Value root;
    std::ofstream ofs;
    bool ok=true;
    std::map<AgentType,std::map<PlanningType,std::vector<PlanningPart*> > >::iterator it;

    //first of all, fetch the current data in the file, we don't want to overwrite it.
    if(!this->readJsonAgent(t)){
        std::ifstream ifstest(this->getPath(t).c_str());
        if(ifstest){
            ifstest.close();
            M3D_ERROR("couldn't parse file before writting to it. Aborting the saving and leave the file untouched");
            ok=false;
        }else{
            //file do not exist, so create it!
            M3D_INFO("creating file to receive Planning Parts at "<<getPath(t));
        }
    }

    if(ok){
        it=_planParts.find(t);
        ok = (it!=_planParts.end());
    }
    if(ok){
        if(! t.isDefined()){
            M3D_ERROR("no string for agent type value: "<<t.toHriAgentType());
            ok=false;
        }
    }
    if(ok){
        root["agent_type"]=t.toString();

        Value json_plan_parts;
        for(std::map<PlanningType,std::vector<PlanningPart*> >::iterator itm=(*it).second.begin();
            itm!=(*it).second.end(); ++itm)
        {
            for(std::vector<PlanningPart*>::iterator itjs=(*itm).second.begin();
                itjs!=(*itm).second.end();itjs++)
            {
                appendJointsSetToJson(json_plan_parts,*itjs);
            }
        }
        root["planning_parts"] = json_plan_parts;
    }

    if(ok){
        ofs.open(getPath(t).c_str());
        ok=(ofs);
        if (!ok){
            M3D_ERROR("could not open file "<<getPath(t));
        }
    }
    if(ok){
        ofs<<root;
        ofs.close();
    }

}

void PlanParts::appendJointsSetToJson(Json::Value &planning_parts_dict, const PlanningPart *jointsset)
{
    using namespace Json;
    Value json_part;
    Value json_arm_ids,json_joints_ids;
    std::vector<int> arm_ids=jointsset->armIds();
    std::vector<int> joints_ids=jointsset->jointsIndices();
    for(std::vector<int>::const_iterator it=arm_ids.begin();it!=arm_ids.end();it++){
        json_arm_ids.append(Value(*it));
    }
    for(std::vector<int>::const_iterator it=joints_ids.begin();it!=joints_ids.end();it++){
        json_joints_ids.append(Value(*it));
    }
    if(jointsset->armIds().size()){
        json_part["arm_ids"]=json_arm_ids;
    }
    if(jointsset->jointsIndices().size()){
        json_part["joints"]=json_joints_ids;
    }

    planning_parts_dict[jointsset->planType().toString()].append(json_part);
}

string PlanParts::getPath(AgentType t)
{
    std::string file=_global_json_path;
    file.append("/PlanningParts/");
    bool ok = t.isDefined();
    if (ok){
        file.append(t.toString());
        file.append(".json");
    }else{
        file="";
    }
    return file;

}

void PlanParts::addPlanPart(AgentType agent_type, PlanningType plan_type, const std::vector<int> &joints, const std::vector<int> &arm_ids)
{
    PlanningPart*  jset = new PlanningPart(agent_type,plan_type,joints,arm_ids);
    addPlanPart(jset);

}

struct functorJointsSetPtrEqual : std::unary_function<PlanningPart*,bool>{
    explicit functorJointsSetPtrEqual(PlanningPart *j):j1(j){}
    bool operator()(PlanningPart * j2){ return (*j1==*j2);}
    PlanningPart *j1;
};
void PlanParts::addPlanPart(PlanningPart *jointsset)
{
    std::map<PlanningType, std::vector<PlanningPart*> > & m= _planParts[jointsset->agentType()];
    std::vector<PlanningPart*> const & v = m[jointsset->planType()];

    std::vector<PlanningPart*>::const_iterator it = std::find_if(v.begin(),v.end(),functorJointsSetPtrEqual(jointsset));
    if(it==v.end()){
        _planParts[jointsset->agentType()][jointsset->planType()].push_back(jointsset);
    }else{
        M3D_INFO("Planning part already exist");
    }
}

std::vector<PlanningPart *> PlanParts::getPlanningParts(AgentType agent_type, PlanningType plan_type)
{
    std::map<AgentType, std::map<PlanningType, std::vector<PlanningPart*> > >::const_iterator it = _planParts.find(agent_type);
    std::map<PlanningType, std::vector<PlanningPart*> >::const_iterator it2;
    if ( it== _planParts.end()){
        if (!proactive()){
            M3D_INFO("no plan part for agent type "<<agent_type.toString()<<", fetching data");
            if(readJsonAgent(agent_type)){
                return getPlanningParts(agent_type,plan_type);
            }
        }
    }
    it = _planParts.find(agent_type);
    if ( it!= _planParts.end()){
        it2=it->second.find(plan_type);
        if(it2 != it->second.end()){
            return it2->second;
        }
    }else{
    }
    return std::vector<PlanningPart *>(0);
}

//std::map<PlanningType,std::string> createPlanningTypeStrings(){
//    std::map<PlanningType,std::string> m;
//    m[SEM_PLAN_BASE]="plan_base";
//    //m[PLAN_ARM_LEFT]="plan_arm_left";
//    //m[PLAN_ARM_RIGHT]="plan_arm_right";
//    //m[PLAN_ARM_OTHER]="plan_arm_other";
//    m[SEM_PLAN_MANIP]="plan_manip";
//    m[SEM_PLAN_ALL]="plan_all";
//    m[SEM_PLAN_MOBILE_MANIP]="plan_mobile_manip";
//    return m;
//}


//std::map<PlanningType,std::string> PlanParts::_planningTypeStrings = createPlanningTypeStrings();
//string PlanParts::planningTypeToString(PlanningType t)
//{
//    std::string s;
//    if(_planningTypeStrings.find(t) != _planningTypeStrings.end()){
//        s=_planningTypeStrings[t];
//    }else{
//        s="";
//    }
//    return s;
//}

//PlanningType PlanParts::stringToPlanningType(const string &s)
//{
//    PlanningType t(SEM_PLANNIG_TYPE_SIZE);
//    for (std::map<PlanningType,std::string>::const_iterator it=_planningTypeStrings.begin();it!=_planningTypeStrings.end();it++){
//        if(it->second == s){
//            t=it->first;
//            break;
//        }
//    }
//    return t;
//}
bool PlanParts::proactive() const
{
    return _proactive;
}

void PlanParts::setProactive(bool proactive)
{
    _proactive = proactive;
    if(_proactive){
        this->fetchData();
    }
}



PlanningPart::PlanningPart():
    _jointsIndices(0),
    _armIds(0)
{
    setPlanType(PlanningType());
}

PlanningPart::PlanningPart(AgentType agent_type, PlanningType plan_type, const std::vector<int> &joint_ids, const std::vector<int> &arm_ids)
{
    this->setAgentType(agent_type);
    this->setPlanType(plan_type);
    this->setJointsIndices(joint_ids);
    this->setArmIds(arm_ids);
}

PlanningPart::PlanningPart(std::vector<PlanningPart *> planning_parts):
    _jointsIndices(0),
    _armIds(0)
{
    this->setPlanType(PlanningType());//planning type is not defined

    if(planning_parts.size()){
        std::set <int> jnts;
        std::set<int> arms;
        AgentType agn_t = planning_parts.front()->agentType();
        for(std::vector<PlanningPart*>::const_iterator it = planning_parts.begin(); it!= planning_parts.end();it++){
            assert((*it)->agentType() == agn_t);
            if((*it)->agentType() != agn_t){
                throw (std::runtime_error("agent type differs in given planning_parts"));
            }
            jnts.insert((*it)->jointsIndices().begin(),(*it)->jointsIndices().end());
            arms.insert((*it)->armIds().begin(),(*it)->armIds().end());
        }
        _jointsIndices.assign(jnts.begin(),jnts.end());
        _armIds.assign(arms.begin(),arms.end());
        setAgentType(agn_t);
    }
}

PlanningPart::~PlanningPart()
{

}

void PlanningPart::concat(const PlanningPart &p)
{
    std::set <int> jnts;
    std::set<int> arms;
    assert(p.agentType() == agentType());
    if(p.agentType() != agentType()){
        throw (std::runtime_error("agent type differs in given planning_parts"));
    }
    jnts.insert(_jointsIndices.begin(),_jointsIndices.end());
    arms.insert(_armIds.begin(),_armIds.end());
    jnts.insert(p.jointsIndices().begin(),p.jointsIndices().end());
    arms.insert(p.armIds().begin(),p.armIds().end());

    _jointsIndices.assign(jnts.begin(),jnts.end());
    _armIds.assign(arms.begin(),arms.end());
}


std::vector<int> PlanningPart::getAllJointIds() const
{
    std::vector<int> v,aux;
    v=jointsIndices();
    for (std::vector<int>::const_iterator it=armIds().begin(); it!=armIds().end(); it++){
        aux=getJointsOfArmId(*it);
        v.insert(v.end(),aux.begin(),aux.end());
    }
    return v;

}

std::vector<int> PlanningPart::getJointsOfArmId(int armId) const
{
    Robot* r=getRobotOfType(agentType());
    std::vector<int> v;
    if(!r){
        M3D_WARN("no robot with type "<<agentType().toString());
    }else{
        assert(armId >=0 && static_cast<unsigned int>(armId) < r->getRobotStruct()->armManipulationData->size());
        ArmManipulationData const &mData=(*r->getRobotStruct()->armManipulationData)[armId];
        p3d_cntrt* ct = mData.getCcCntrt();
        M3D_TRACE("joints for arm no "<<armId<< " for robot type "<<agentType().toString()<<" (specimen "<<r->getName()<<", cntrt name: "<<ct->namecntrt<<")");
        for(int i = 0; i < ct->npasjnts; i++){
            p3d_jnt * joint = ct->pasjnts[i];
            if(joint->type != P3D_FIXED){
                v.push_back(joint->num-1);
                M3D_TRACE("\t joint "<<joint->name<<" ("<<joint->num<<")");
            }
        }
        if((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik"))){
            p3d_jnt * joint = r->getRobotStruct()->joints[ct->argu_i[0]];
            v.push_back(joint->num-1);
            M3D_TRACE("\t joint "<<joint->name<<" ("<<joint->num<<")");
        }
        std::sort(v.begin(),v.end());
    }
    return v;

}

PlanningPart *PlanningPart::createFromActiveJoints(PlanningType plan_type)
{
    Robot *r=global_Project->getActiveScene()->getActiveRobot();
    AgentType type;
    PlanningPart *jset= 0;
    bool ok= r && r->getHriAgent();
    if (!ok){
        if(r){
            M3D_WARN("Robot "<<r->getName()<<" has no HRI_AGENT set");
        }
        if(!r){
            M3D_WARN("No robot set as active");
        }
    }
    if (ok){
        std::vector<int> joints_id, arm_ids;
        std::map<int,bool*> arm_jnts_all;
        std::vector<std::map<int,bool*> > arm_jnts_by_arm;
        type = r->getHriAgent()->type;
        jset = new PlanningPart(type,plan_type,std::vector<int>(0),std::vector<int>(0));
        unsigned int nb_arm = r->getRobotStruct()->armManipulationData->size();
        unsigned int numberOfJoints = r->getNumberOfJoints();


        //create arm maps to detect if full arms are active
        for(unsigned int i=0;i<nb_arm;i++){
            std::vector<int> buff_arm_jnt=jset->getJointsOfArmId(i);
            std::map<int,bool*> arm_map;
            for (unsigned int j=0;j<buff_arm_jnt.size();j++){
                arm_map[buff_arm_jnt[j]]=new bool(false);
                arm_jnts_all[buff_arm_jnt[j]] = arm_map[buff_arm_jnt[j]];
            }
            arm_jnts_by_arm.push_back(arm_map);
        }

        for (unsigned int i=0;i<numberOfJoints;i++){
            bool active=false;
            for(unsigned int j=0;j<r->getJoint(i)->getNumberOfDof();j++){
                //active = active || r->getJoint(i)->isJointDofUser(j);
                if(r->getJoint(i)->getJointStruct()->dof_data[j].is_active_for_planner>0){
                    active=true;
                    M3D_TRACE("active DoF: joint "<<r->getJoint(i)->getName()<<" dof "<<j);
                }
            }
            if (active){
                int id = r->getJoint(i)->getId();
                M3D_TRACE("joint "<<r->getJoint(i)->getName() << " ("<<id<<") is active");
                if(arm_jnts_all.find(i)!=arm_jnts_all.end()){
                    //is arm joint
                    *arm_jnts_all[id] = true;
                }else{
                    //otherwise add as joint
                    //arm joints will be added either as being part of an arm, or as a joint if arm not complete, but latter
                    joints_id.push_back(id);
                }
            }
        }

        //find if some full arm are active:
        for (unsigned int i=0;i<arm_jnts_by_arm.size();i++){
            bool complete=true;
            for (std::map<int,bool*>::const_iterator it=arm_jnts_by_arm[i].begin();it!=arm_jnts_by_arm[i].end();++it){
                complete = complete && (*it->second == true);
                if (!complete) break;
            }
            if (complete){
                //add the arm
                arm_ids.push_back(i);
            }else{
                //add active joints of the arm, if any
                for (std::map<int,bool*>::const_iterator it=arm_jnts_by_arm[i].begin();it!=arm_jnts_by_arm[i].end();++it){
                    if( *it->second ==true){
                        joints_id.push_back(it->first);
                    }
                }
            }
        }

        std::sort(joints_id.begin(),joints_id.end());
        jset->setArmIds(arm_ids);
        jset->setJointsIndices(joints_id);

        //clean:
        for(std::map<int,bool*>::iterator it=arm_jnts_all.begin();it!=arm_jnts_all.end();++it){
            delete it->second;
        }
        arm_jnts_all.clear();
        arm_jnts_by_arm.clear();
    }

    return jset;

}

PlanningType PlanningPart::planType() const
{
    return _planType;
}

void PlanningPart::setPlanType(const PlanningType &planType)
{
    _planType = planType;
}
const std::vector<int> &PlanningPart::jointsIndices() const
{
    return _jointsIndices;
}

void PlanningPart::setJointsIndices(const std::vector<int> &jointsIndices)
{
    _jointsIndices = jointsIndices;
}
AgentType PlanningPart::agentType() const
{
    return _agentType;
}

void PlanningPart::setAgentType(const AgentType &agentType)
{
    _agentType = agentType;
}
const std::vector<int> & PlanningPart::armIds() const
{
    return _armIds;
}

void PlanningPart::setArmIds(const std::vector<int> &armIds)
{
    _armIds = armIds;
}

Robot *PlanningPart::getRobotOfType(AgentType t)
{
    Robot* rob=0;
    for(unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots();i++){
        Robot* r=global_Project->getActiveScene()->getRobot(i);
        HRI_AGENT * h=r->getHriAgent();
        if (h && AgentType(h->type) == t){
            rob=r;
            break;
        }
    }
    return rob;
}

bool PlanningPart::operator==(PlanningPart const &other) const{
    bool equal=true;
    if(this->agentType() != other.agentType())
        equal=false;
    if(equal){
        if (this->planType() != other.planType())
            equal=false;
    }
    if(equal){
        std::vector<int> vthis = this->getAllJointIds();
        std::vector<int> vother = other.getAllJointIds();
        std::sort(vthis.begin(),vthis.end());
        std::sort(vother.begin(),vother.end());
        equal=(vthis==vother);
    }
    return equal;
}

bool PlanningPart::operator!=(PlanningPart const &other) const{
    return !(*this == other);
}

std::vector<std::string> createDefaultPlanningTypes(){
    std::vector<std::string> v;
    v.push_back("manipulation");
    v.push_back("navigation");
    v.push_back("mobile_manipulation");
    v.push_back("all");
    v.push_back("gaze");
    return v;
}

std::vector<std::string> PlanningType::__strs = createDefaultPlanningTypes();

PlanningType::PlanningType()
{
    _value =-1;
}

PlanningType::PlanningType(std::string s)
{
    std::vector<std::string>::const_iterator it=std::find(PlanningType::__strs.begin(),__strs.end(),s);
    if (it!=PlanningType::__strs.end()){
        this->_value=it-__strs.begin();
    }else{
        __strs.push_back(s);
        this->_value=__strs.size()-1;
    }
}

PlanningType::~PlanningType()
{

}

std::string PlanningType::toString() const
{
    if(isDefined()){
        return __strs[_value];
    }else{
        return "undef!";
    }
}

bool operator==(PlanningType const &p1, PlanningType const &p2)
{
    assert (p1._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p1._value) < PlanningType::__strs.size()) );
    assert (p2._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p2._value) < PlanningType::__strs.size()) );
    assert ( p1.isDefined() == (p1._value >= 0));
    assert ( p2.isDefined() == (p2._value >= 0));
    return p1._value == p2._value;
}
bool operator!=(PlanningType const &p1, PlanningType const &p2)
{
    assert (p1._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p1._value) < PlanningType::__strs.size()) );
    assert (p2._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p2._value) < PlanningType::__strs.size()) );
    assert ( p1.isDefined() == (p1._value >= 0));
    assert ( p2.isDefined() == (p2._value >= 0));
    return p1._value != p2._value;
}

bool operator<(PlanningType const &p1, PlanningType const &p2)
{
    assert (p1._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p1._value) < PlanningType::__strs.size()) );
    assert (p2._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p2._value) < PlanningType::__strs.size()) );
    assert ( p1.isDefined() == (p1._value >= 0));
    assert ( p2.isDefined() == (p2._value >= 0));
    return p1._value<p2._value;
}
bool operator>(PlanningType const &p1, PlanningType const &p2)
{
    assert (p1._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p1._value) < PlanningType::__strs.size()) );
    assert (p2._value == -1 || ( p1._value>=0 && static_cast<unsigned int>(p2._value) < PlanningType::__strs.size()) );
    assert ( p1.isDefined() == (p1._value >= 0));
    assert ( p2.isDefined() == (p2._value >= 0));
    return p1._value>p2._value;
}

std::vector<PlanningType> PlanningType::currentlyDefinedPlanningTypes()
{
    std::vector<PlanningType> v;
    for(std::vector<std::string>::const_iterator it=__strs.begin();it!=__strs.end();++it){
        v.push_back(PlanningType(*it));
    }
    return v;
}


} // namespace Sem

