#include "Sem_SemanticsManager.hpp"

#include <stdlib.h>

namespace Sem {

INIT_MOVE3D_STATIC_LOGGER(SemanticsManager,"planners.semantics.manager");

SemanticsManager *SemanticsManager::__mainSemanticsManager = 0;

SemanticsManager::SemanticsManager()
{
    setProactive(0);
    std::string path(getenv("HOME_MOVE3D"));
    path+="/Semantics";
    setGlobalJsonPath(path);
    createModules();
}

SemanticsManager::~SemanticsManager()
{
    if(_planParts){
        delete _planParts;
    }
    if(_actions){
        delete _actions;
    }
}

SemanticsManager *SemanticsManager::mainSemanticsManager()
{
    if(__mainSemanticsManager ==0){
        M3D_INFO("creating new main semantic Manager");
        __mainSemanticsManager=new SemanticsManager();
    }
    return __mainSemanticsManager;
}

void SemanticsManager::createModules()
{
    M3D_DEBUG("creating modules");
    setPlanParts(new PlanParts(globalJsonPath(),proactive()));
    setActions(new Actions(globalJsonPath(),proactive()));
    setActionsManager(new ActionsManager(actions(),planParts()));
}
ActionsManager *SemanticsManager::actionsManager() const
{
    assert(_actionsManager->actions() == actions());
    assert(_actionsManager->planParts() == planParts());
    return _actionsManager;
}

void SemanticsManager::setActionsManager(ActionsManager *actionsManager)
{
    _actionsManager = actionsManager;
}


Actions *SemanticsManager::actions() const
{
    return _actions;
}

void SemanticsManager::setActions(Actions *actions)
{
    _actions = actions;
}

bool SemanticsManager::proactive() const
{
    return _proactive;
}

void SemanticsManager::setProactive(bool proactive)
{
    _proactive = proactive;
}

std::string SemanticsManager::globalJsonPath() const
{
    return _globalJsonPath;
}

void SemanticsManager::setGlobalJsonPath(const std::string &globalJsonPath)
{
    _globalJsonPath = globalJsonPath;
}

PlanParts *SemanticsManager::planParts() const
{
    return _planParts;
}

void SemanticsManager::setPlanParts(PlanParts *planParts)
{
    _planParts = planParts;
}


} // namespace Sem

