/*
 *  AStarReplanner.cpp
 *  libmove3D-planners 
 *
 *  Created by Benjamin Vadant on 01/02/13.
 *  Based on Jim Mainprice replanningAlgorithms.hpp/cpp
 *  Copyright 2013 LAAS/CNRS. All rights reserved.
 *
 */

#include "AStarReplanner.hpp"
#include "ReplanningSimulator.hpp"

#include "API/project.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/plannerFunctions.hpp"
#include "planner/TrajectoryOptim/Stomp/stompOptimizer.hpp"
#include "planner/TrajectoryOptim/trajectoryOptim.hpp"
#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"
#include "planner/cost_space.hpp"
#include "planner/Diffusion/Variants/RRT-Star.hpp"

#include "HRICS_costspace.hpp"

#include "LightPlanner-pkg.h"
#include "Graphic-pkg.h"
#include "Collision-pkg.h"

#include "move3d-headless.h"

#include <boost/thread/thread.hpp>
#include <sys/time.h>

using namespace std;
using namespace tr1;

AStarReplanner::AStarReplanner(Robot* r) : SimpleReplanner(r)
{

}

AStarReplanner::~AStarReplanner()
{
    
}

bool AStarReplanner::init()
{
    SimpleReplanner::init();
    
    navigation = new HRICS::Navigation( robot );
    
    return true;
}

void AStarReplanner::run()
{
    isPlanning = true;
    cout << "AStarReplanner :: Start replanning step : " << __func__ << endl;
    
    planningSucceeded = false;
    
    idRun++;
    
    //Time of Execution
    timeval tim;
    gettimeofday(&tim, NULL);
    double tInit = tim.tv_sec+(tim.tv_usec/1000000.0);
    double timeExecution = 0.0;
    //Time of Execution
    
    
    API::Trajectory newPortion = currentTrajectory.extractSubTrajectory( positionSwitch, currentTrajectory.getRangeMax(), false);
    cout << "newPortion.getNbOfPaths() : " << newPortion.getNbOfPaths() << endl;
    cout << "newPortion.getRangeMax() : " << newPortion.getRangeMax() << endl;
    
    navigation->reset();
    API::Trajectory* path = navigation->computeRobotTrajectory( newPortion.getBegin(), newPortion.getEnd() );
    
    //Time of Execution
    gettimeofday(&tim, NULL); 
    timeExecution = tim.tv_sec+(tim.tv_usec/1000000.0) - tInit;
    //Time of Execution
    
    // Set the time left for smoothing
    PlanEnv->setBool( PlanParam::trajWithTimeLimit, true );
    PlanEnv->setDouble( PlanParam::timeLimitSmoothing, timeLimitRePlanning-0.3-timeExecution );
    
    if( path != NULL && 
            !PlanEnv->getBool(PlanParam::stopPlanner) && 
            PlanEnv->getBool(PlanParam::withSmoothing) )
    {
        p3d_smoothing_function( robot->getRobotStruct(), path->replaceP3dTraj(NULL), 100, -1 );             //100??
    }
    
    if( path ) 
    {
        // Cut the new trajectory in lp average piecese
        API::Trajectory finalNewPortion = robot->getCurrentTraj();
        finalNewPortion.cutTrajInSmallLPSimple( newPortion.getRangeMax() / localPathAverageLength );
        std::pair<bool,API::Trajectory> concatenatedTrajectory = concatenateToCurrentTrajectory( finalNewPortion );
        planningSucceeded = concatenatedTrajectory.first;
        
        globalRePlanningEnv->storeTrajectoryToDraw( *path, 0 );                                                                   //simulator
        delete path;
        
        if( planningSucceeded ) {
            currentTrajectory = concatenatedTrajectory.second;
        }
        else {
            cout << "Could not concatenate trajectory" << endl;
        }
    }
    
    if( !planningSucceeded ) {
        cout << "AStarReplanner :: Replanning failed" << endl;
    }
    
    cout << "AStarReplanner :: End replanning step : " << __func__ << endl;
    isPlanning = false;
}
