/*
 *  RRTReplanner.cpp
 *  libmove3D-planners 
 *
 *  Created by Benjamin Vadant on 01/02/13.
 *  Based on Jim Mainprice replanningAlgorithms.hpp/cpp
 *  Copyright 2013 LAAS/CNRS. All rights reserved.
 *
 */

#include "RRTReplanner.hpp"
#include "ReplanningSimulator.hpp"

#include "API/project.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/plannerFunctions.hpp"
#include "planner/TrajectoryOptim/Stomp/stompOptimizer.hpp"
#include "planner/TrajectoryOptim/trajectoryOptim.hpp"
#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"
#include "planner/cost_space.hpp"
#include "planner/Diffusion/Variants/RRT-Star.hpp"

#include "HRICS_costspace.hpp"

#include "LightPlanner-pkg.h"
#include "Graphic-pkg.h"
#include "Collision-pkg.h"

#include "move3d-headless.h"

#include <boost/thread/thread.hpp>
#include <sys/time.h>

using namespace std;
using namespace tr1;

RRTReplanner::RRTReplanner(Robot* r) : SimpleReplanner(r)
{
    
}

RRTReplanner::~RRTReplanner()
{
    
}

bool RRTReplanner::init()
{
    SimpleReplanner::init();
    
    ENV.setBool(Env::isCostSpace,false);
    ENV.setBool(Env::useTRRT,false);
    return true;
}

void RRTReplanner::run()
{
    isPlanning = true;
    cout << "RRTReplanner :: Start replanning step : " << __func__ << endl;
    
    planningSucceeded = false;
    
    PlanEnv->setBool(   PlanParam::planWithTimeLimit, true );
    PlanEnv->setDouble( PlanParam::timeLimitPlanning, timeLimitRePlanning-0.3 );// Set 0.3 seconds to do the concatenation of trajectories
    
    idRun++;
    
    API::Trajectory newPortion = currentTrajectory.extractSubTrajectory( positionSwitch, currentTrajectory.getRangeMax(), false);
    cout << "newPortion.getNbOfPaths() : " << newPortion.getNbOfPaths() << endl;
    cout << "newPortion.getRangeMax() : " << newPortion.getRangeMax() << endl;

    //Time of Execution
    timeval tim;
    gettimeofday(&tim, NULL);
    double tInit = tim.tv_sec+(tim.tv_usec/1000000.0);
    double timeExecution = 0.0;
    //Time of Execution
    
    p3d_traj* path = p3d_planner_function( robot->getRobotStruct(),
                                           newPortion.getBegin()->getConfigStruct(), 
                                           newPortion.getEnd()->getConfigStruct() );
    
    //Time of Execution
    gettimeofday(&tim, NULL); 
    timeExecution = tim.tv_sec+(tim.tv_usec/1000000.0) - tInit;
    //Time of Execution
    
    // Set the time left for smoothing
    PlanEnv->setBool( PlanParam::trajWithTimeLimit, true );
    PlanEnv->setDouble( PlanParam::timeLimitSmoothing, timeLimitRePlanning-0.3-timeExecution );
  
    if( path != NULL && 
            !PlanEnv->getBool(PlanParam::stopPlanner) && 
            PlanEnv->getBool(PlanParam::withSmoothing) )
    {
        p3d_smoothing_function( robot->getRobotStruct(), path, 100, -1 );
    }
    
    if( path ) {
        // Cut the new trajectory in lp average piecese
        API::Trajectory finalNewPortion = robot->getCurrentTraj();
        finalNewPortion.cutTrajInSmallLPSimple( newPortion.getRangeMax() / localPathAverageLength );
        std::pair<bool,API::Trajectory> concatenatedTrajectory = concatenateToCurrentTrajectory( finalNewPortion );
        planningSucceeded = concatenatedTrajectory.first;
            
        if( planningSucceeded ) {
            currentTrajectory = concatenatedTrajectory.second;
        }
        else {
            cout << "RRTReplanner :: Could not concatenate trajectory" << endl;
        }
    }
    
    if( !planningSucceeded ) {
        cout << "RRTReplanner :: Replanning failed" << endl;
    }
    
    cout << "RRTReplanner :: End replanning step : " << __func__ << endl;
    isPlanning = false;
}
