#include "TrajEval.h"
#include <boost/bind.hpp>
#include "cost_space.hpp"
#include "HRICS_costspace.hpp"

using namespace std;

INIT_MOVE3D_STATIC_LOGGER(TrajEval,"planners.cost.trajeval");
TrajEval::TrajEval()
{

}

double TrajEval::maximum(API::Trajectory& t,std::vector<boost::function<double (API::Trajectory &)> > &funcs){
    double v=-numeric_limits<double>::infinity();
    for(vector<boost::function<double(API::Trajectory &)> >::iterator it=funcs.begin(); it!=funcs.end(); ++it){
        v=max(v,(*it)(t));
    }
    return v;
}
double TrajEval::sum(API::Trajectory& t,std::vector<boost::function<double (API::Trajectory &)> > &funcs){
    double v=0;
    for(vector<boost::function<double(API::Trajectory &)> >::iterator it=funcs.begin(); it!=funcs.end(); ++it){
        v+=(*it)(t);
    }
    return v;
}

double TrajEval::alpha(API::Trajectory &t,double a,boost::function<double(API::Trajectory &)> &f){
    return a * f(t);
}

boost::function<double(API::Trajectory&)> TrajEval::getTrajFunction(std::string name){
    return boost::bind(TrajEval::traj_cost,_1,name);
}


double TrajEval::traj_cost(API::Trajectory &t, std::string &fx_name){
    std::string prev_cost=global_costSpace->getSelectedCostName();
    global_costSpace->setCost(fx_name);
    double eval = t.costRecomputed();
    global_costSpace->setCost(prev_cost);
    return eval;
}

double TrajEval::example(API::Trajectory &t){
    std::vector<boost::function<double(API::Trajectory&)> > max_f;
    std::vector<boost::function<double(API::Trajectory&)> > sum_f;
    max_f.push_back(boost::bind(TrajEval::traj_cost,_1,string("NoCost")));
    //max_f.push_back(boost::bind(TrajEval::traj_cost,_1,string("costClearance")));
    max_f.push_back(boost::bind(TrajEval::traj_cost,_1,string("costHumanGrids")));

    sum_f.push_back(boost::bind(TrajEval::maximum,_1,max_f));
    //sum_f.push_back(boost::bind(TrajEval::traj_cost,_1,"NoCost"));
    sum_f.push_back(boost::bind(TrajEval::alpha,_1,0.1,TrajEval::getTrajFunction("NoCost")));
    return TrajEval::sum(t,sum_f);
}

void TrajEval::run_example(){
    M3D_INFO("running TrajEval example");
    API::Trajectory t=global_Project->getActiveScene()->getActiveRobot()->getCurrentTraj();
    double c =example(t);
    M3D_INFO("cost is "<<c);
}
