#include "graphAlgorithms.h"
#include <map>
#include <algorithm>
#include <iostream>
#include <limits>


using namespace std;
namespace mho{
vector<pair<cost_t,vector<NodeInterface*> > > enumerate_paths(GraphInterface *g,
                                                              NodeInterface *x,NodeInterface *destination){
    return enumerate_paths(g,x,destination,numeric_limits<cost_t>::quiet_NaN());
}

vector<pair<cost_t,vector<NodeInterface*> > > enumerate_paths(GraphInterface *g,
                                                              NodeInterface *x,NodeInterface *destination,
                                                              cost_t max_cost){
    vector<NodeInterface*> pile;
    vector<cost_t> cost_pile;
    //vector<NodeInterface*> D, T;
    map<NodeInterface*,vector<EdgeInterface*> >to_visit;
    vector<pair<cost_t,vector<NodeInterface*> > > paths;
    NodeInterface* top,*current;
    cost_t current_cost=0;
    unsigned int count=0;


    /*D=g->getNodes();
    vector<NodeInterface*>::iterator f=find(D.begin(),D.end(),x);
    if(f!=D.end()){
        //remove x from D
        *f=D.back();
        D.pop_back();
    }else{
        //error
    }*/

    to_visit[x]=g->sons(x);
    if(!to_visit[x].empty()){
        pile.push_back(x);
        cost_pile.push_back(0);
    }


    while(!pile.empty()){
        count++;
        top=pile.back();
        current_cost=cost_pile.back();//current_cost=cost to reach top
        if(!to_visit[top].empty()){
            //get an element to visit after 'top' and remove it
            current=to_visit[top].back()->oppositeSideOf(top);
            current_cost+=to_visit[top].back()->getCost();
            to_visit[top].pop_back();
            if(max_cost!=0 && current_cost>=max_cost){
                //stop exploration here.
            }else if(current==destination){
                paths.push_back(pair<cost_t,vector<NodeInterface*> >(current_cost,pile));
                paths.back().second.push_back(current);
            }else if(find(pile.begin(),pile.end(), current)==pile.end()){ //current is not in pile
                /*vector<NodeInterface*>::iterator f=find(D.begin(),D.end(),current);
                if(f!=D.end()){
                    //remove current from D
                    *f=D.back();
                    D.pop_back();
                }*/

                to_visit[current]=g->sons(current);
                pile.push_back(current);
                cost_pile.push_back(current_cost);
                if(destination==0){
                    paths.push_back(pair<cost_t,vector<NodeInterface*> >(current_cost,pile));
                }

            }else{
                //TODO add circuit
            }

        }else{
            //T.push_back(top);
            pile.pop_back();
            cost_pile.pop_back();
        }
    }
    cout << "enumerate_paths() done in "<< count << " iterations."<<endl;
    return paths;
}

pair<cost_t,vector<NodeInterface*> > random_path(GraphInterface *g, NodeInterface *x,NodeInterface *destination){
    return random_path(g,x,destination,numeric_limits<cost_t>::quiet_NaN());
}

pair<cost_t,vector<NodeInterface*> > random_path(GraphInterface *g, NodeInterface *x,NodeInterface *destination,cost_t max_cost){
    vector<NodeInterface*> path;
    vector<cost_t> cost_pile;
    vector<vector<EdgeInterface*> > all_possibles;//stack of unexplored edges for each node in the explored path
    path.push_back(x);
    cost_pile.push_back(0);
    all_possibles.push_back(g->sons(x));//adds the sons of the first node
    cost_t current_cost;
    while(path.size()!=0 && path.back()!=destination){
        current_cost=cost_pile.back();
        unsigned int ap_i=all_possibles.size()-1;//back
        NodeInterface * next=0;
        while(all_possibles.back().size()!=0 && next==0){
            unsigned int index=rand()%all_possibles[ap_i].size();
            next=all_possibles[ap_i][index]->oppositeSideOf(path.back());
            if(current_cost+all_possibles[ap_i][index]->getCost()>=max_cost){
                //don't explore this branch
            }else if (find(path.begin(),path.end(),next)==path.end()){//next is not in path
                cost_pile.push_back(current_cost+all_possibles[ap_i][index]->getCost());
                all_possibles.push_back(g->sons(next));
                path.push_back(next);
            }
            swap(all_possibles[ap_i].at(index),all_possibles[ap_i].back());
            all_possibles[ap_i].pop_back();
        }
        if(next==0){
            path.pop_back();
            all_possibles.pop_back();
            cost_pile.pop_back();
        }
    }
    return pair<cost_t,vector<NodeInterface*> >(current_cost,path);
}


std::pair<cost_t,std::deque<NodeInterface*> > dijkstra(GraphInterface *g,NodeInterface* n1,NodeInterface* n2)
{
    std::map<NodeInterface*,cost_t> dist;
    std::map<NodeInterface*,NodeInterface*> pred;
    std::vector<pair<cost_t,NodeInterface*> > pq;
    std::vector<NodeInterface*> nodes=g->getNodes();

    for(std::vector<NodeInterface*>::iterator it=nodes.begin();it!=nodes.end();++it){
        dist.insert(pair<NodeInterface*,cost_t>(*it,numeric_limits<cost_t>::infinity() ));
        pred.insert(pair<NodeInterface*,NodeInterface*>(*it,0));
    }
    dist[n1]=0;
    //use a vector and arrange it as a heap
    for (std::vector<NodeInterface*>::iterator it=nodes.begin();it!=nodes.end();++it){
        pq.push_back(pair<cost_t,NodeInterface*>(dist[*it],*it));
    }
    std::make_heap(pq.begin(),pq.end(),std::greater<pair<cost_t,NodeInterface*> >());

    while(!pq.empty()){
        NodeInterface* u=pq.front().second;
        std::pop_heap(pq.begin(),pq.end(),std::greater<pair<cost_t,NodeInterface*> >());
        pq.pop_back();
        for (unsigned int i=0;i<g->sons(u).size();++i){
            NodeInterface* v=g->sons(u).at(i)->oppositeSideOf(u);
            cost_t c=g->sons(u).at(i)->getCost();
            if(dist[v]>dist[u]+c){
                //decrease key of v
                unsigned int search_i=0;
                while(pq[search_i].second != v){search_i++;}
                //pq[search_i] is v
                pq[search_i].first=dist[u]+c;//update cost
                while(search_i!=0 && pq[search_i].first<pq[(search_i-1)/2].first){
                    //swap father and son
                    std::swap(pq[search_i],pq[(search_i-1)/2]);
                    search_i=(search_i-1)/2;
                }

                //TODO CHECK

                dist[v]=dist[u]+c;
                pred[v]=u;
            }
        }
    }

    deque<NodeInterface*> path;
    NodeInterface* nc=n2;
    if (dist[n2]<numeric_limits<cost_t>::infinity()){
        while(nc!=n1 && nc!=0){
            path.push_front(nc);
            nc=pred[nc];
        }
        path.push_front(n1);
    }
    return pair<cost_t,std::deque<NodeInterface*> >(dist[n2], path);
}
std::pair<cost_t,std::vector<NodeInterface*> > dijkstra(GraphInterface *g, NodeInterface* n1, NodeInterface* n2, bool vect){
    std::vector<NodeInterface*> ret;
    std::pair<cost_t,std::deque<NodeInterface*> > d_ret=dijkstra(g,n1,n2);
    ret.assign(d_ret.second.begin(),d_ret.second.end());
    return std::pair<cost_t,std::vector<NodeInterface*> >(d_ret.first,ret);
}

ReverseDisjkstra::ReverseDisjkstra(GraphInterface *g){
    setGraph(g);
}
void ReverseDisjkstra::setGraph(GraphInterface *g){
    this->g=g;
}
void ReverseDisjkstra::setTarget(NodeInterface *target){
    this->target=target;
}

void ReverseDisjkstra::compute(NodeInterface *target){
    this->setTarget(target);
    this->compute();
}

void ReverseDisjkstra::compute(){
    std::vector<pair<cost_t,NodeInterface*> > pq;
    std::vector<NodeInterface*> nodes=g->getNodes();

    dist.clear();
    pred.clear();
    for(std::vector<NodeInterface*>::iterator it=nodes.begin();it!=nodes.end();++it){
        dist.insert(pair<NodeInterface*,cost_t>(*it,numeric_limits<cost_t>::infinity() ));
        pred.insert(pair<NodeInterface*,NodeInterface*>(*it,0));
    }
    dist[target]=0;
    //use a vector and arrange it as a heap
    for (std::vector<NodeInterface*>::iterator it=nodes.begin();it!=nodes.end();++it){
        pq.push_back(pair<cost_t,NodeInterface*>(dist[*it],*it));
    }
    std::make_heap(pq.begin(),pq.end(),std::greater<pair<cost_t,NodeInterface*> >());

    while(!pq.empty()){
        NodeInterface* u=pq.front().second;
        std::pop_heap(pq.begin(),pq.end(),std::greater<pair<cost_t,NodeInterface*> >());
        pq.pop_back();
        for (unsigned int i=0;i<g->sons(u).size();++i){
            NodeInterface* v=g->sons(u).at(i)->oppositeSideOf(u);
            cost_t c=g->sons(u).at(i)->getCost();
            if(dist[v]>dist[u]+c){
                //decrease key of v
                unsigned int search_i=0;
                while(pq[search_i].second != v){search_i++;}
                //pq[search_i] is v
                pq[search_i].first=dist[u]+c;//update cost
                while(search_i!=0 && pq[search_i].first<pq[(search_i-1)/2].first){
                    //swap father and son
                    std::swap(pq[search_i],pq[(search_i-1)/2]);
                    search_i=(search_i-1)/2;
                }

                //TODO CHECK

                dist[v]=dist[u]+c;
                pred[v]=u;
            }
        }
    }

}

std::pair<cost_t,std::deque<NodeInterface*> >  ReverseDisjkstra::getPathDeque(NodeInterface *start){
    deque<NodeInterface*> path;
    NodeInterface* nc=start;
    if (dist[start]<numeric_limits<cost_t>::infinity()){
        while(nc!=target && nc!=0){
            path.push_back(nc);
            nc=pred[nc];
        }
        path.push_back(target);
    }
    cost_t cost=dist[start];
    return make_pair(cost,path);
}

std::pair<cost_t,std::vector<NodeInterface*> > ReverseDisjkstra::getPathVector(NodeInterface *start){
    vector<NodeInterface*> path;
    NodeInterface* nc=start;
    if (dist[start]<numeric_limits<cost_t>::infinity()){
        while(nc!=target && nc!=0){
            path.push_back(nc);
            nc=pred[nc];
        }
        path.push_back(target);
    }
    cost_t cost=dist[start];
    return make_pair(cost,path);
}

}

